#ifndef CLASS_L4dqCheck_ICC
#define CLASS_L4dqCheck_ICC


// implementation of the member functions for L4dqCheck class library
// ===================================================================
// Author: Martin Braeuer 
//
// NOTE: 
//
// Modification history:
// MBr  Nov 99  : Initial developmant as part of ALIGN5
// MBr  29.11.99: 1. porting to ARTE
// MBr  30.11.99: bad-chip-side mechanism to this Class
//                    => aviation1-0
// MBr  03.12.99: bad-chip-side mechanism tested and improved
// ==================================================================

// includes :
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>
#include  <malloc.h>

// STL stuff:
#include "map.h"
#include "multimap.h"

#include  "grr.h"
#include  "iol.h"
#include  "hbook.h"
#include  "marvin.h"
#include  "vdsleda.h"

// VDS specific:
#include "Struct_Gede.h"   // use Gede_t structs
#include "Struct_Hitb.h"   // use Hitb_t structs
#include "Struct_Track.h"  // use Track_t structs

#include "Class_GEDE.h"    // geometry information
#include "Class_HOLMES.h"  // holmes track finding

// include the class's header file:
#include "alca/AC_SVD_L4dqCheck.h"


// a private maintain/check routine:
void L4dqCheck_class::dump() {
  Gede_t *hgedeP;
  for (int j=0;j<MAXHELIX+1;++j) {
    printf("%2d: ",j);
    for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
      hgedeP = it->second;
      int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);

      //for (int i=0;i<MAXUNSIN;++i)
      switch (helixstat[usid][j]) {
        case H_LOW:     printf("-");break;
        case H_WORKING: printf("0");break;
        case H_HIGH:    printf("+");break;
        default:        printf("?");
      }
    }
    printf("\n");
  }
}

/* ==================================================================
 * == L4dqCheck_class constructor : std constructor                ==
 * == needs : ---                                                  ==
 * == Result : Failure message because of missing HOLMES pointer   ==
 * ==                                                              ==
 * ================================================================== 
 */
L4dqCheck_class::L4dqCheck_class() {
  fError_out("Default constructor call failed!");
  return;
}


/* ==================================================================
 * == L4dqCheck_class constr. : Prepare structures for L4dqCheck   ==
 * == Parameter: HOLMES Pointer                                    ==
 * ==                                                              ==
 * ================================================================== 
 */
L4dqCheck_class::L4dqCheck_class(HOLMES_class *inHOLMESp, 
			   int (*p1)(int, int, int) ) {
  int hstat,usid;
  Gede_t *hgedeP;

  prepare = 0; // This 'makes' a new class -> NOTHING can be prepared!
  accumulate = 0;
  evaluate = 0;
  Pstore = NULL;

  // ensure that HOLMES is ok:
  if (inHOLMESp == NULL) {
    fError_out("HOLMES pointer is invalid!");
    return;
  }
  HOLMESp = inHOLMESp;
  GEDEp   = HOLMESp->gedeP();  
  if (GEDEp == NULL) {
    fError_out("GEDE pointer is invalid!");
    return;
  }
  // fine all info resurces are availible
  for (int i=0;i<MAXUNSIN;++i) 
    for (int j=0;j<MAXHELIX+1;++j) { 
      helixstat[i][j] = H_LOW;
      o_helixstat[i][j] = H_LOW;    
      ocche[i][j] = 0;    
    } // loop over all helix/occ structs

  // set the occupancy store pointer
  Pstore = p1;
  if (Pstore == NULL) { // we ARE offline..
    fprintf(stderr,"AVIATION: L4dqCeck prepare: storing helix occupancy internally\n");
  }

  cEVCUT      = EVCUT;  
  cMINCUT     = MINCUT;  
  cMAXCUT_N   = MAXCUT_N;  
  cMAXCUT_P   = MAXCUT_P;  
  cHELIXCUT   = HELIXCUT;  
  cMINMEANHIT = MINMEANHIT; 
  globalmeanhit = 0.;
  int nhelix = 8;
  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
    hgedeP = it->second;
    usid = fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
    if ((hgedeP->lay == LAY_N1) || (hgedeP->lay == LAY_N2))
      nhelix = 10;
    else 
      nhelix =  8;
    if (hgedeP->status != 0) hstat = H_WORKING; // set working according to gede
     else hstat = H_LOW;
    for (int j=0;j<nhelix;++j)   
      helixstat[usid][j] = hstat;
    helixstat[usid][MAXHELIX] = hstat;
  } // of gede loop 
  prepare = 1;
  return;
}


/* ==================================================================
 * == L4dqCheck_class destructor : Delete structures for align5    ==
 * ==                                                              ==
 * ================================================================== 
 */
L4dqCheck_class::~L4dqCheck_class() {
 fClear();
}


/* ==================================================================
 * == fClear : Clear/unbook all allocated structures               ==
 * ==                                                              ==
 * ================================================================== 
 */
void L4dqCheck_class::fClear() {
  if (prepare>0) { // only if prepared crrectly..
    //    if (type     !=NULL ) { free((char*) type     );type       = NULL;}
  }   
  prepare    = 0;
  accumulate = 0;
  evaluate   = 0;
  ntrack =0;

  return;
}


void L4dqCheck_class::fError_out(const char *s) {
  /*====================================
     Give output to the various sources
   =====================================
   */

  fprintf(stderr,"AVIATION: Class_L4dqCheck: Fatal Error : '%s'!!\n",s);
  
  return;
} // =============== END of fOut member func. ========================================

/* ==================================================================
 * == fPrepare : Prepare structures for L4dqCheck class            ==
 * == Parameter: Pointer to HOLMES structure                       ==
 * ==            A function taking usid, chip and value to store   ==
 * ==            occupancies in rhp objects                        ==
 * == needs : ----                                                 ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fPrepare( ) {
  prepare    = 1;
  accumulate = 0;
  evaluate   = 0;
  // Do this always during init 
  for (int i=0;i<MAXUNSIN;++i)
    for (int j=0;j<MAXHELIX;++j)
      ocche[i][j] = 0;    
  return(1);
}


/* ==================================================================
 * == fSetHelixPar : Set Cuts for helixes                          ==
 * == needs  : --                                                  ==
 * == pars   : evcut     cluster < events/EVCUT => dead!           ==
 * ==          mincut    cluster < MINCUT       => dead!           ==
 * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
 * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
 * ==          minchips  >MINCHIPS number of chips need to work    ==
 * ==          minmeanhit > <hits all sides>/minmeanhit must       ==
 * ==                       be on a plane                          ==
 * == return : 1 : OK, 0: not initialised                          ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fSetPar(int evcut, int mincut, float maxn, float maxp,
			     int minchips, float minmeanhit) {
  if (prepare != 1) return(0); // prepare before
  cEVCUT    = evcut;   
  cMINCUT   = mincut;  
  cMAXCUT_N = maxn;    
  cMAXCUT_P = maxp;    
  cHELIXCUT = minchips;
  cMINMEANHIT= minmeanhit;
  return(1);
} // =============== END of fSetHelixPar member func. ========================================


/* ==================================================================
 * == fGetHelixPar : Get Cuts for helixes                          ==
 * == needs  : --                                                  ==
 * == pars   : evcut     cluster < events/EVCUT => dead!           ==
 * ==          mincut    cluster < MINCUT       => dead!           ==
 * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
 * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
 * ==          minchips  >MINCHIPS number of chips need to work    ==
 * ==          minmeanhit > <hits all sides>/minmeanhit must       ==
 * ==                       be on a plane                          ==
 * == return : 1 if set correctly                                  ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fGetPar(int *evcut, int *mincut, float *maxn, float *maxp,
			     int *minchips, float *minmeanhit) {
  if (prepare != 1) return(0); // prepare before
  *evcut      = cEVCUT;    
  *mincut     = cMINCUT;   
  *maxn       = cMAXCUT_N; 
  *maxp       = cMAXCUT_P; 
  *minchips   = cHELIXCUT;
  *minmeanhit = cMINMEANHIT;
  return(1);
} // =============== END of fGetHelixPar member func. ========================================
       

/* =====================================================================
 * == fAccuTrack: Insert a whole HOLMES track into the int. occ. maps ==
 * == needs     : Class prepared                                      ==
 * == pars      : side=side OR USID of hit, cog cog of hit            ==
 * == return    : 1 if OK, 0 if not prepared or invalid side-code     ==
 * ==                                                                 ==
 * ===================================================================== 
 */
int L4dqCheck_class::fAccuTrack(int traid) {
  if (prepare != 1) return(0); // prepare before
  evaluate = 0; // we have new hits!

  Track_t track=*(HOLMESp->trackP(traid)); // You have to project to 00 !! TGl!
  for (int j=0;j<track.nhit;++j) {
    Hitb_t  *hit = HOLMESp->hitbP(track.hit[j]); // get hitb Pointer for hit-index
    if (Pstore == NULL)
      ocche[fSideUnsin(hit->gede->side)][fHelixid(hit->gede->side,hit->cog)]++;
    else
      Pstore(fSideUnsin(hit->gede->side),fHelixid(hit->gede->side,hit->cog),1);
    ++accumulate;
  }
  return(1);
} // =============== END of fAccuTrack member func. ========================================


/* ====================================================================
 * == fAccuHit : Insert a hit into the int. occ. maps                ==
 * == needs    : Class prepared                                      ==
 * == pars     : side=side OR USID of hit, cog cog of hit            ==
 * == return   : 1 if OK, 0 if not prepared or invalid side-code     ==
 * == Note     : This works with usid and side id !!!                ==
 * == Note     : Use THIS always to store. A NULL in the contructor  ==
 * ==             will lead to an internal store, a function pointer ==
 * ==             to external storage. (YOU have to copy back then!) ==
 * ==                                                                ==
 * ==================================================================== 
 */
int L4dqCheck_class::fAccuHit(int side,float cog) {
  if (prepare != 1) return(0); // prepare before
  evaluate = 0; // we have a new hit!

  int gedechan = -1;
  if (side > 200)  
    gedechan = fSideUnsin(side);
  else 
    gedechan = side;
  if ((gedechan < 0) || (gedechan > 127)) return(0); // accept only 0..127 !!
  //printf("  gc%d  %8.3f %d |",gedechan,cog,fHelixid(side,cog));fflush(stdout);

  if (Pstore == NULL) {
    ocche[gedechan][fHelixid(side,cog)]++;
    ++accumulate;            // only in case of internal store !!!
  }
  else
    Pstore(gedechan,fHelixid(side,cog),1);
  return(1);
} // =============== END of fAccuHit (side,usid) member func. =====================





/* ==================================================================
 * == fAcceptTrack: check that track does not touch bad chips      ==
 * == needs    : preppare has to run before                        ==
 * == pars     : track no in HOLMES struct                         ==
 * == return   : 1 if track does not contain a suspecious hit      ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fAcceptTrack(int traid) {
  int usid;
  if (prepare != 1) return(0); // prepare before
  Track_t track=*(HOLMESp->trackP(traid,0.0)); // You have to project to 00 !! TGl!
  for (int j=0;j<track.nhit;++j) {
    Hitb_t  *hit = HOLMESp->hitbP(track.hit[j]); // get hitb Pointer for hit-index
    usid = fUnsin(hit->gede->sl,hit->gede->sec,hit->gede->lay);
    if (helixstat[usid][fHelixid(hit->gede->side,hit->cog)] != H_WORKING) return(0);
  } 
  return(1);
} // =============== END of fAcceptTrack member func. ========================================

/* ==================================================================
 * == fAcceptHit: check that a hit does not contain bad chips      ==
 * == needs    : preppare has to run before                        ==
 * == pars     : side ID and cog of hit                            ==
 * == return   : 1 hit OK, 0 if not prepared or hit bad            ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fAcceptHit(int side,float cog) {
  if (prepare != 1) return(0); // prepare before
  if (helixstat[fSideUnsin(side)][fHelixid(side,cog)] != H_WORKING) return(0);
                                                               else return(1);
} // =============== END of fAcceptHit member func. ========================================


/* ==================================================================
 * == fEvaluate: set up a new helixstate matrix                    ==
 * == needs    : fAccuHit at least MIN_ACCU times                  ==
 * == pars     : evdone=number of processed events                 ==
 * == return   : 1: no change, 0: chips changed, -1: error         ==
 * ==                                                              ==
 * ================================================================== 
 */

int L4dqCheck_class::fEvaluate1(int evdone) {
  Gede_t *hgedeP;

  //printf("AVIATION: L4dqCheck: eval nev=%d hi=%d\n",evdone,accumulate);
  if (accumulate < MIN_ACCU) return(-1);
  int nhelix=0;
  // clear the helixstate array:
  for (int i=0;i<MAXUNSIN;++i) 
    for (int j=0;j<MAXHELIX+1;++j) {
      o_helixstat[i][j] = helixstat[i][j];
    }

  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
    hgedeP = it->second;
    int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
    for (int j=0;j<MAXHELIX+1;++j) {
      helixstat[usid][j] = H_WORKING; // set all existing to WORING
    }
  }
  // evaluate helixes:, loop over all gedes existing
  //printf("  %d\n",helixstat[0][0]);
  //dump();printf("\nd1\n");
  int minok = int(evdone/cEVCUT);
  if (minok < cMINCUT) minok=cMINCUT;
  //printf("  %d\n",helixstat[0][0]);

  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
    hgedeP = it->second;
    int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);

    float meanhit=0.;
    if (minok < cMINCUT) minok=cMINCUT;
    if (((hgedeP->lay)==LAY_N1) || ((hgedeP->lay)==LAY_N2)) 
      nhelix=10;
    else 
      nhelix=8;

    // the exclude-loops:
    // find to few events:
    int hcount =0;
    // calc underrun, calc mean
    for (int i=0;i<nhelix;++i) {
      if (ocche[usid][i] < minok) 
	helixstat[usid][i] = H_LOW; // to few cluster
      else {
	++hcount;
	helixstat[usid][i] = H_WORKING;
      }
      meanhit += float(ocche[usid][i]);
    }
    if (nhelix == 8) {
      helixstat[usid][8] = H_LOW;
      helixstat[usid][9] = H_LOW;
    }

    int hcount_o=hcount;
    //printf("%d  hc=%2d\n",hgedeP->side,hcount);
    while (1) {
      //calc act. mean value:
      meanhit=0.;
      for (int i=0;i<nhelix;++i) 
	if (helixstat[usid][i] == H_WORKING) meanhit+=float(ocche[usid][i]);
      meanhit/=hcount;

      // flag to death with overrruns:
      hcount = 0;
      for (int i=0;i<nhelix;++i) {
	if (helixstat[usid][i] == H_WORKING)
	  if ((hgedeP->lay == LAY_N1) ||(hgedeP->lay == LAY_N2))
	    if (ocche[usid][i] > (cMAXCUT_N*meanhit)) 
	      helixstat[usid][i] = H_HIGH; // to much cluster
	    else ++hcount;
	  else 
	    if (ocche[usid][i] > (cMAXCUT_P*meanhit)) 
	      helixstat[usid][i] = H_HIGH; // to much cluster
	    else ++hcount;
      }
      if (hcount_o == hcount) break; // did not find anydody to remove!
      hcount_o=hcount;
    }
    //    printf("   %d  hc=%2d\n",hgedeP->side,hcount);
    

    //total plane to kill:
    if (hcount < cHELIXCUT) {
      //printf("%d  ",hgedeP->side);
      for (int i=0;i<MAXHELIX+1;++i) {
	helixstat[usid][MAXHELIX] = H_LOW;  // dead for align5
	//	printf("%3d  %4d   ",i,ocche[usid][i]);
      }
      hgedeP->status = 0;                 // dead for tracking
      //printf("\n");
    }
    //printf("  %d\n",helixstat[0][0]);

  } // of gede loop 1
  //dump();printf("\nd2\n");
  //now we build the global mean hit no's:
  globalmeanhit=0.;
  int planecount = 0;
  int totalhelix = 0;
  int hitweight[MAXUNSIN]; // number of participating helixes
  int hitsplane[MAXUNSIN];
  for (int i=0;i<MAXUNSIN;++i) {
    hitsplane[i]=0;
    hitweight[i]=0;
  }
  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
    hgedeP = it->second;

    int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
    int hitsonplane = 0;
    if (helixstat[usid][MAXHELIX] == H_WORKING) ++planecount; // only workings
                                           else continue; 
    for (int i=0;i<nhelix;++i)
      if (helixstat[usid][i] == H_WORKING) {
	hitsonplane+=ocche[usid][i]; // take the hits of each working helix
	++hitweight[usid];
	++totalhelix;
      }
    // printf("%3d hw=%2d\n",usid,hitweight[usid]);
    //hitsonplane = int(hitsonplane*10/hitweight[usid]);
    hitsplane[usid] = hitsonplane;
    globalmeanhit += hitsonplane;
  } // of gede loop 2

  //dump();

  globalmeanhit /= planecount;

  for (int i=0;i<MAXUNSIN;++i) {
    //if (hitsplane[i] >0) printf("  %3d : g=%5.0f  act=%5d nh=%2d\n",i,globalmeanhit,hitsplane[i],hitweight[i]);
    if ((hitsplane[i] >0)&&(hitsplane[i] < globalmeanhit/cMINMEANHIT)) { // kill a plane again:
      for (int j=0;j<MAXHELIX+1;++j) 
	//printf("  %d: %4d %4d",j,ocche[i][j],helixstat[i][j]);
      for (int j=0;j<MAXHELIX+1;++j)
	helixstat[i][j] = H_LOW;
      // hgedeP->status = 0;                 // dead for tracking
      //printf("\n");
    }
  }

  have_acc    = accumulate;
  have_events = evdone;
  accumulate = 0; // reset the accumulation flag, to start a new one
  evaluate = 1;   // now we have succesfully evaluated

  // check if plane had identical problems before!
  int retk = 1;
  for (int i=0;i<MAXUNSIN;++i) {
    if ((  helixstat[i][MAXHELIX] != H_WORKING) &&
	(o_helixstat[i][MAXHELIX] != H_WORKING)    ) continue;
    for (int j=0;j<MAXHELIX+1;++j) 
      if (o_helixstat[i][j] != helixstat[i][j]) {
	retk=0;
	//printf("c hange in (%3d,%2d)  o=%d n=%d\n",i,j,o_helixstat[i][j],helixstat[i][j]);
      }
  }
  //printf("return %d\n",retk);
  //dump();
  //int i;
  //printf("enter1: ");scanf("%d",&i);
  return(retk);
}  // =============== END of fEvaluate1 member func. ========================================

/* ==================================================================
 * == fResult  : Textoutput of result                              ==
 * == needs    : fEvaluate1 has to run before                      ==
 * == pars     :                                                   ==
 * == return   : 1 if OK, 0 if not evaluated before                ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fResult() {
 Gede_t *hgedeP;
 int nc,nhelix,usid,i;
 if (evaluate != 1) return(0); // evaluate before
 
 printf("\nList of planes with suspicious properties:\n");
 printf(" used cuts (for %d accumulated hits in %d events) : \n",have_acc,have_events);
 printf("  EVCUT      (no of clus/chip < events/EVCUT)                 : %3d\n",cEVCUT);
 printf("  MINCUT     (no of clus/chip < MINCUT)                       : %3d\n",cMINCUT);
 printf("  MAXCUT_N   (no of clus/chip > MAXCUT_N*meanhits_on_n_plane) : %6.2f\n",cMAXCUT_N);
 printf("  MAXCUT_P   (no of clus/chip > MAXCUT_P*meanhits_on_p_plane) : %6.2f\n",cMAXCUT_P);
 printf("  MINMEANHIT (each plane must have > meanhit/minmeanhit clus) : %6.2f\n",cMINMEANHIT);
 printf("  HELIXCUT   ( > HELIXCUT chips need to work )                : %d\n\n",cHELIXCUT);

 // loop over gede's:
 for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
   hgedeP = it->second;
   usid = fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
   if (((hgedeP->lay)==LAY_N1) || ((hgedeP->lay)==LAY_N2)) 
     nhelix=10;
   else 
     nhelix=8;
   if (helixstat[usid][MAXHELIX] == H_LOW) {
     printf("%3d %d : Entire Plane dead! \n",usid,hgedeP->side);
   } else {
     // check for changes:
     nc = 0;
     for (i=0;i<nhelix;++i)
       if (helixstat[usid][i] != H_WORKING) {
	 nc=1;break;
       }
     if (nc == 1) {
       printf("%3d %d : Helixes dead     : ",usid,hgedeP->side);
       for (i=0;i<nhelix;++i) {
	 printf("  %d",i);
	 if (helixstat[usid][i] == H_LOW)     printf("--");
	 if (helixstat[usid][i] == H_WORKING) printf("  ");
	 if (helixstat[usid][i] == H_HIGH)    printf("++");
       }
       printf("\n");
     }
   }
 }
 return(1);
} // =============== END of fResult member func. ========================================

/* ==================================================================
 * == fGetWorking : returns the number of working helixes          ==
 * == needs    : fEvaluate1 has to run before                      ==
 * == pars     :                                                   ==
 * == return   : -1: if error, number otherwise                    ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fGetWorking() {
 Gede_t *hgedeP;
 int nhelix;
 
 if (evaluate != 1) return(-1); // evaluate before

 int working = 0;
 for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
   hgedeP = it->second;
   int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
   if (((hgedeP->lay)==LAY_N1) || ((hgedeP->lay)==LAY_N2)) 
     nhelix=10;
   else 
     nhelix=8;
   for (int j=0;j<nhelix;++j)
     if (helixstat[usid][j] == H_WORKING) ++working; 
 }
 return(working);
} // =============== END of fGetWorking member func. ========================================

/* ==================================================================
 * == fGetNonWorking : returns the number of not working helixes   ==
 * == needs    : fEvaluate1 has to run before                      ==
 * == pars     :                                                   ==
 * == return   : -1: if error, number otherwise                    ==
 * ==                                                              ==
 * ================================================================== 
 */
int L4dqCheck_class::fGetNotWorking() {
 Gede_t *hgedeP;
 int nhelix;
 
 if (evaluate != 1) return(-1); // evaluate before

 int nworking = 0;
 for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
   hgedeP = it->second;
   int usid=fUnsin(hgedeP->sl,hgedeP->sec,hgedeP->lay);
   if (((hgedeP->lay)==LAY_N1) || ((hgedeP->lay)==LAY_N2)) 
     nhelix=10;
   else 
     nhelix=8;
   for (int j=0;j<nhelix;++j)
     if (helixstat[usid][j] != H_WORKING) ++nworking; 
 }
 return(nworking);
} // =============== END of fGetNotWorking member func. ========================================



#endif
// ============ EOF ====== EOF ====== EOF ====== EOF ====== EOF ====== EOF
