// ********************************************************************
//
//  Routine:      acReadRICH
//  Author:       D. Dujmic,  Dec 1, 1999 
//
//  Description:  - update RITH table with Cherenkov radius,
//                - update internal buffer of Riter with alignment constants for mirrors
//                - if key not assigned, use run number as major number
//                - returns -1 error
//                           0 no update needed
//                           1 update done using key
//                           2 update done using run number (RITH) and maxrev (MIRAL)
//                
//
//  Modification:
//     D.D., Jan 26, 2000, check for initialization added 
//     D.D., Apr 03, 2000, rad0 changed to radians 
//     D.D., May 25, 2000, keytable access for hot/cold channels
//                         moved to rich initialization (RICHDB.C)
//     d.d., Nov 28, 2000, return+arte_error replaces with arte_exit
//     5 Apr 2002 (RP,MS)  table MIRAL (mirror alignment constants) added,
//                         reorganisation of code (separate routines for RITH and MIRAL,
//                         called from acReadRICH)
// ********************************************************************

extern "C" {
#include <stdlib.h>
#include <stdio.h>
}

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"

extern "C" {
  void riter_put_miral_(int *len,int *itype, int *id, float *sx, float *sy, 
		   float *ex, float *ey,float *ex1, float *ey1 );
}
int acReadRichRITH( int KTstatus);
int acReadRichMiral( int KTstatus);


int acReadRICH( int KTstatus) {
  
  int ier=acReadRichRITH( KTstatus);
  acReadRichMiral( KTstatus);
  return ier;
}


int acReadRichRITH( int KTstatus) {

  static int  old_run=0;         // rich database
  static char dbsite[256];       // 
  char   *dbname=0;              //
  Tables x=0;                    //
  int    t=0;                    //
  int    ret=0;                  //
  
  static int RICHserver = -1;    // key table
  static int RICHtable  = -1;    //
  
  static int ac_rich_init=0;     // initialization
      
  while (1) {

    switch (KTstatus) {
      
    case 2: 
      ////////////////////// initialization //////////////////////
      
      // if key not assigned just find RICH_RECODB
      // (RICH_RECODB must be defined)
      
      dbname=getenv("RICH_RECODB");
      if (!dbname) {
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: RICH_RECODB not defined" , -1);
	//return -1;
	ARTE_exit(GRR_SERROR, " $$$ acReadRICH: RICH_RECODB not defined", -1);
      }
      strcpy(&dbsite[0],&dbname[0]);
      ac_rich_init=1;
      RICHserver = KeyFindServer(dbsite);
      if (RICHserver==-1) {
	ARTE_error( GRR_SWARNING," $$$ acReadRICH: KeyTable server not found", -1); 
	KTstatus = -2;
	continue;
      }
      RICHtable=KeyFindTable(RICHserver, "RICH_RADIUS");
      if (RICHtable==-1) {
	ARTE_error( GRR_SWARNING," $$$ acReadRICH: KeyTable not found", -1);
	KTstatus = -2;
	continue;
      }
      
    case 1: 
      /////////////////////// key assigned /////////////////////
      
      // Database entry with assigned key
      // take the keytable revision and run
      
      if (!ac_rich_init) {
	KTstatus=2; // initialize first
	continue;
      }
      if (RICHserver==-1 || RICHtable==-1) {
      	KTstatus = -2; // key not assigned for RICH
	continue;
      }
      ARTE_error( GRR_SDEBUG," $$$ acReadRICH: Key assigned", 0);
      x=OpenTables(KeyServerName(RICHserver),TABLEreadonly);
      if (!x) {
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Can not open database", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Can not open database", -1);
      }
      if (!KeyTableNew(RICHtable)) return 0;
      t = BookTable(x, KeyTableName(RICHtable), 0, 0);
      if (!t) {
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Cannot find a table with RICH radius", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Cannot find a table with RICH radius", -1);
      }
      KeyTableStamp(x, KeyServerName(RICHserver), KeyTableName(RICHtable));    
      if (!FetchTable(x,t,0)) { 
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Cannot fatch a table for assigned key", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Cannot fatch a table for assigned key", -1);
      }
      ret=1;
      
      break;
      
    case -2: 
      //////////////////// key NOT assigned ////////////////////
      
      // Database entry without assigned key:
      // take the highest revision of nearest run
      
      if (!ac_rich_init) {
	KTstatus=2; // initialize first
	continue;
      }
      if (old_run==ArtePointer<EVHD>(1)->run) return 0; // no update
      
      ARTE_error( GRR_SDEBUG," $$$ acReadRICH: Key NOT assigned", 0);
      x=OpenTables(dbsite,0);
      if (!x) {
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Cannot open database", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Cannot open database", -1);
      }
      t=BookTable(x,"RICH_RADIUS",0,0);
      if (!t) {
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Cannot book table RICH_RADIUS", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Cannot book table RICH_RADIUS", -1);
      }
      int run_no=ArtePointer<EVHD>(1)->run;
      TableStamp(x,run_no,TABLEmaxrev);
      if (!FetchTable(x,t,TABLElt)) { 
	//ARTE_error( GRR_SERROR," $$$ acReadRICH: Cannot fetch a table for run", -1);
	//return -1;
	ARTE_exit( GRR_SERROR," $$$ acReadRICH: Cannot fetch a table for run", -1);
      }
      ret=2;
    }
    break;
  }
  
  //////////////////// update RITH /////////////////////
  SkipTableField(x,1);
  int   major=TableMajor(x);
  int   minor=TableMinor(x);
  float parf[7];
  int   pari[2];
  GetTableField(x,TABLEfloat,10000,&parf[0]);
  GetTableField(x,TABLEfloat,10000,&parf[1]);
  GetTableField(x,TABLEfloat,10000,&parf[2]); // radius
  GetTableField(x,TABLEfloat,10000,&parf[3]); // time
  GetTableField(x,TABLEfloat,10000,&parf[4]); // min search radius
  GetTableField(x,TABLEfloat,10000,&parf[5]); // max search radius
  GetTableField(x,TABLEfloat,10000,&parf[6]); // search step
  GetTableField(x,  TABLEint,10000,&pari[0]); // occupancy cut
  GetTableField(x,  TABLEint,10000,&pari[1]); // threshold
  
  ArteTable<RITH>::clear();
  RITH rith;
  rith.rad0=parf[2]*.001;
  rith.run =major;
  rith.ver =minor;
  rith.time=int(parf[3]);
  memcpy(&rith.par[0], &parf[4], sizeof(float)*3);
  rith.par[3]=float(pari[0]);
  rith.par[4]=float(pari[1]);
  ArteTable<RITH>::push_back(rith);
  ARTE_error( GRR_SDEBUG," $$$ acReadRICH: ArteTable<RITH> refreshed", 0);
  
  old_run=ArtePointer<EVHD>(1)->run;
  CloseTables(x);
  
  return ret;
}


int acReadRichMiral( int KTstatus) {
  
  static int  old_run=0;         // rich database
  static char dbsite[256];       // 
  char   *dbname=0;              //
  Tables x=0;                    //
  int    t=0;                    //
  int    ret=0;                  //
  
  static int RICHserver = -1;    // key table
  static int RICHtable  = -1;    //
  static int ac_rich_init=0;     // initialization
  
  
  if (!ac_rich_init) {
    dbname=getenv("RICH_INSTALLDB");
    if (!dbname) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: var. RICH_INSTALLDB not defined", 0);
      return -1;
    }
    strcpy(&dbsite[0],&dbname[0]);
    ac_rich_init=1;
  }  

  switch (KTstatus) {
    
  case 2: 
    
    RICHserver = KeyFindServer(dbsite);
    if (RICHserver==-1) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: server not found", 0);
      return -1;
    }
    RICHtable=KeyFindTable(RICHserver, "MIRAL");
    if (RICHtable==-1) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: table not found", 0);
      return -1;
    }
    
  case 1: 
    
    if (RICHserver==-1 || RICHtable==-1) {
      return -1;
    }
    x=OpenTables(KeyServerName(RICHserver),TABLEreadonly);
    if (!x) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot open tables", 0);
      return -1;
    }
    if (!KeyTableNew(RICHtable)) return 0;
    t = BookTable(x, KeyTableName(RICHtable), 0, 0);
    if (!t) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot book table", 0);
      return -1;
    }
    KeyTableStamp(x, KeyServerName(RICHserver), KeyTableName(RICHtable));    
    if (!FetchTable(x,t,0)) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot fetch table", 0);
      return -1;
    }
    ret=1;
    break;
      
  case -2: 
    
    if (old_run==ArtePointer<EVHD>(1)->run) return 0; // no update
    
    x=OpenTables(dbsite,0);
    if (!x) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot open database", 0);
      return -1;
    }
    t=BookTable(x,"MIRAL",0,0);
    if (!t) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot book table", 0);
      return -1;
    }
    TableStamp(x,TABLEmaxrev,TABLEmaxrev); /* goes to the last revision of the table*/
    if (!FetchTable(x,t,TABLElt)) {
      ARTE_error(GRR_SWARNING, "acReadRichMiral: cannot fetch table", 0);
      return -1;
    }
    ret=2;
  }

  SkipTableField(x,1);

  int nl=256;
  int itype[nl], id[nl];
  float sx[nl], sy[nl];
  float ex[nl], ey[nl];
  float ex1[nl], ey1[nl];
  int nb[8],minnl;
  
  nb[0]=GetTableField(x,TABLEint  ,  nl,itype); 
  nb[1]=GetTableField(x,TABLEint  ,  nl,id); 
  nb[2]=GetTableField(x,TABLEfloat,  nl,sx); 
  nb[3]=GetTableField(x,TABLEfloat,  nl,sy); 
  nb[4]=GetTableField(x,TABLEfloat,  nl,ex); 
  nb[5]=GetTableField(x,TABLEfloat,  nl,ey); 
  nb[6]=GetTableField(x,TABLEfloat,  nl,ex1); 
  nb[7]=GetTableField(x,TABLEfloat,  nl,ey1);

  CloseTables(x);

  minnl=nl;
  for (int i=0;i<8;i++)  {
    if (nb[i]<minnl) minnl=nb[i] ;
  }
  riter_put_miral_(&minnl,itype,id,sx,sy,ex,ey,ex1,ey1);

  old_run=ArtePointer<EVHD>(1)->run;

  return ret;
}


