
// *********************************************************************
//
//  Routine:      acread_
//  Author:       T.Oest
//
//  Description:  general steering routine to apply internal and
//                
//
//  Modification:
//  acGlobAlChange() switched off (causes a segmentation fault
//  if the database is not defined)                   T.Oest  2/8/99
//  VDS and global alignment constants read in        T.Oest 16/9/99
//  keytable code by M. Hissmann added                T.Oest 20/9/99
//  call to acReadTARGET added                        T.Oest 24/9/99
//  call of acReadDQ added                            T.Oest 29/10/99
//  alignment is read in only for DST if reco on or
//  KeyBook was called in the kumac                   T.Oest 02/12/99
//  set alca flag, add routine to return alcaFlag     
//  call acReadMuon and ITR_arte_updateGEDE           T.Oest 10/02/00
//  acSetDateTime() included, GTAR update moved after global
//  acReadGlobal                                      T.Oest 08/03/00 
//  of only HITB preparation is requested: read in 
//  only those constants which are used               T.Oest 27/03/00 
//  acReadECAL included                               T.Oest 31/03/00
//  acReadTARGET called for every event               T.Oest 11/04/00
//  use ARTE_mark() to save position                  AG     14/04/00
//  more ARTE_mark()                                  AG     25/04/00
//  use faComp() to check components                  AG     11/05/00
//  more ARTE_mark()                                  AG      2/06/00
//  polishing                                         AG      6/06/00
//  call of new acReadITR routine added               TZ     13/10/00
//  polished code                                     AG     29/11/00
//  printout old key                                  AG      6/12/00
//  added timing                                      AG      6/12/00
//  apply alignment if TAHD exists                    DG     12/12/00
//  add memory usage monitoring                       JH     13/11/01
//  adapt acread to be called also for MC             JH     13/11/01
//  do not read CnA constants for miniDST unless
//  forced in kumac                                   JH     16/11/01
//  printout for MC OTR only for first event          JH     28/11/01
//  check L4HitMask                                   JH     28/01/02
//  reset 2 lower bytes of alcaFlag per event         JH     29/01/02 
//  Use faReco() and faHit() to get masks             JH     09/03/02
//  remove check if component bit in component mask
//  and reco mask is ON since it is done in RC_init   JH     19/03/02
//  acReadMuon also used for MC                       JH     24/07/02
//
// ********************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "database/keytable.h"
#include "component.h"
#include "reco/rcflags.hh"
#include "itrarte/ITR_arte.h"
#include "Otr/OtrMc.h"
#include "Otr/OtrI.h"
#include "database/table.h"

// defines
#define COMP_M_GLOBAL 0x00000001

// external
extern int FATIME;
extern int FAMEM;
extern int FAON;
extern int KeyBookWasCalled;
extern int acFlagReadVDS ;
extern int acFlagReadOTR ;
extern int acFlagReadITR ;
extern int acFlagReadECAL;
extern int acFlagReadMUON;
extern int acFlagReadRICH;
extern int acFlagReadCnA;
extern int acFlagReadGLOBAL;
extern int CnAkey;
extern int CnArevision;
extern int get_CnAKey(int runnumber); 

// prototypes
int acReadTARGET(int);
int acReadVDS(int);
int acReadITR(int);
int acReadOTR(int);
int acReadRICH(int);
int acReadECAL(int);
extern "C" int acReadMuon(int);
int acReadGlobal(int);
int acReadDQ(int);
void acGlobalAlignmentForGEDE(unsigned int);
void acSetDateTime();

// globals
static unsigned int alcaFlag  = 0;


extern "C" void
acread_()
{
  // set mark
  ARTE_mark( "acread_ 0", 0 );

  // Do not read CnA constants if Farm Online ARTE and both RecoMask and HitMask are 0
  if (!faReco(0xFFFF) && !faHit(0xFFFF)) return;

  // Do not read CnA constants if reconstruccion is OFF unless reading is forced in kumac
  if (recoFlag == 0 && acFlagReadCnA == 0 ) return; 

  // Do not read CnA constants if miniDST event unless reading is forced in kumac. This condition
  // should be redundant since the miniDST the reconstruction should be OFF.
  if (ArteTable<MCIN>::size() == 0 && ArteTable<DEVT>::size() == 0 && acFlagReadCnA == 0 ) return; 


  // memory usage monitoring
  int in[2], out[2];
  static int mem_vds = 0;
  static int mem_otr = 0;
  static int mem_itr = 0;
  static int mem_ecal = 0;
  static int mem_muon = 0;
  static int mem_rich = 0;
  static int mem_global = 0;
  static int mem_target = 0;

  //
  // start timing
  //
  static Chrono *cron_acread = 0;
  if( !cron_acread ){
    ChronoManager CM;
    cron_acread = CM.createChrono( "cron_acread" );
  }
  cron_acread->start();

  //
  // set date and time in event header
  //
  ARTE_mark( "acSetDateTime", 0 );
  acSetDateTime();
  

  //=========================================================================
  // check for the first call if and how the CnA constants should be read in:
  //   - reading without the CnA keytable if KEYBOOK command not in kumac
  //     The acReadXXX functions should handle this case (KTstatus = -2)
  //   - no reading  if all subdetector reconstructions are OFF
  //   - no reading  for MINIs (ArteTable<DEVT>::size() == 0)
  //   - reading (all subd) if forced in kumac (/CNA/READ CNA ON)
  //   - reading of any subdetector CnA constants can be switched 
  //     off in kumac individually (e.g. /CNA/READ OTR OFF)
  //=========================================================================
  
  static int firstCall = 0;
  int KTstatus = 0;

  if (firstCall == 0){
    firstCall = 1;
    if (KeyBookWasCalled != 1) { 
      if (FAON) {  // FARM mode online
	ARTE_exit( GRR_SFATAL,
		   "CnA keytable was not initialized (missing KEYBOOK command in kumac. Exiting!", 0);
      } else {     // offline, allow loading of CnA constants without keytable 
	ARTE_error( GRR_SWARNING, 
		    "CnA constants will not be read using the keytable!. KEYBOOK missing in kumac", 0 );
	KTstatus = -2; // allow loading of CnA constants once (default versions in acReadXXX functions)	
      }
    }
  }

  
  //==================================================================
  // check for a new alignment/calibration key (DKEY table)
  //      KTstatus will contain status of KeyTable. The acReadXXX
  //      functions should react accordingly
  //               -2  no KUIP keytable initalisation (reading of
  //                   CnA constants via keytable off)
  //               -1  failed to find keytable 
  //                0  No new keytable
  //                1  new keytable (same structure)
  //                2  new structure (e.g. the 1st time)
  //==================================================================
  

  int key;
  static int oldKey = 0;
  char message[256];

  // set CnA key for DB access
  //
  if (ArteTable<MCIN>::size() == 0 ) {  // data (read CnAkey from event header)
    key = faEVHDdbmajor();              // key = -1 if no EVHD table (for L2ARTEfmtr)
  } else {                              // Monte Carlo (use key specified in kumac ) 
    key = CnAkey;                       // key = 0 if RUN2CNAKEY missing in kumac for MC
  }
 
  ArtePointer<EVHD> evhd = *ArteTable<EVHD>::begin();    

  // load CnA table if new and valid key 
  //
  if (KeyBookWasCalled == 1) { // Keytable was initialized (KEYBOOK command in kumac)
    if (key > 0) KTstatus = KeyLoadTables( key );
    if (KTstatus==-1) {        // the keytable could not be found
      sprintf( message, "keytable %d for KEYBOOK=%d not found in database. Check that the KEYBOOK "
	       "revision %d you have specified in the kumac exists for the run %d. !!! EXITING !!!!", 
	       faEVHDdbmajor(), CnArevision, CnArevision, evhd->run );
      ARTE_exit( GRR_SFATAL, message, 0 );
    }
  }

  // write CnA key and revision in event header EVHD
  //
  evhd->dbmajor = key;
  evhd->dbminor = CnArevision;


  //
  // For backwards compatibility, if MC reconstruction with no acread 
  // (RUN2CNAKEY missing in kumac), then only call acReadOTR loading the keytable first
  //
  if (ArteTable<MCIN>::size() != 0 && CnAkey==0) {  // old MC reconstruction for OTR
    int key=0, status=0;
    static int otrfirst=1;
    char *otrRunNumber=getenv("OtrRunnumber"); 
    if (otrRunNumber) {
      if (otrfirst) {
	otrfirst = 0;
	key = get_CnAKey(atoi(otrRunNumber));
	if (key > 0) {
	  sprintf( message, "Loading CnA keytable %d corresponding to run given by "
		   "setenv OtrRunnumber %d in kumac", key, atoi(otrRunNumber));
	  ARTE_error( GRR_SINFO, message, 0 );
	  status = KeyLoadTables(key);
	} else {
	  sprintf( message, "Could not find CnA keytable %d corresponding to run given by "
		   "setenv OtrRunnumber %d in kumac", key,  atoi(otrRunNumber));
	  ARTE_error( GRR_SFATAL, message, 0 );
	}
      }
      acReadOTR(status);
    }
    cron_acread->stop();
    return;
  }



  //==================================================================
  // set alignment constants, transforms NEDE to GEDE tables
  //  - no action if no reconstruction will be done
  //  - alignment sensitive parameters should be taken from 
  //    the subdetector database (no off values with respect to NEDE)
  //  - constant values (pitch) should be taken from NEDE
  //  - return flag: > 0   constants were changed
  //                 = 0   no change of constants
  //  - return code  < 0   error, constants not found or read error
  //                 = 0     constants uptodate
  //                 = 1     constants read in by key mechanism
  //                 = 2     constants read in using event time or run number
  //==================================================================

  //
  // call subdetector acReadXXX functions
  //
  int flagTARGET = 0;
  int flagVDS    = 0;
  int flagITR    = 0;
  int flagOTR    = 0;
  int flagMUON   = 0;
  //int flagTRD    = 0;
  int flagRICH   = 0;
  int flagECAL   = 0;
  //int flagHPT    = 0;
  int flagGLOBAL = 0;
  int flagDQ     = 0;

  if (KTstatus!=0) {

    // info message
    sprintf( message, "acread: New CnA key. run=%d old key=%d new key=%d",
	     arteflag_.numrun, oldKey, key );
    oldKey = key; 
    ARTE_error( GRR_SINFO, message, GRR_OK );
      

    // DQ. Dummy for the moment
    ARTE_mark( "acReadDQ", 0 );
    flagDQ = acReadDQ(KTstatus);    

    // VDS
    if ((recoVDS != 0 || acFlagReadCnA != 0) && acFlagReadVDS == 1){
      ARTE_mark( "acReadVDS", 0 );
      vm_size(in);
      flagVDS = acReadVDS(KTstatus);
      vm_size(out); mem_vds += (out[0]-in[0]);
    }

    // ITR
    if ((recoITR != 0 || acFlagReadCnA != 0) && acFlagReadITR == 1){
      ARTE_mark( "acReadITR", 0 );
      vm_size(in);
      flagITR = acReadITR(KTstatus);
      vm_size(out); mem_itr += (out[0]-in[0]);
    }

    // OTR
    if ((recoOTR != 0 || acFlagReadCnA != 0) && acFlagReadOTR == 1){
      ARTE_mark( "acReadOTR", 0 );
      vm_size(in);
      flagOTR = acReadOTR(KTstatus);
      vm_size(out); mem_otr += (out[0]-in[0]);
    }

    // RICH
    if ((recoRICH != 0 || acFlagReadCnA != 0) && acFlagReadRICH == 1){
      ARTE_mark( "acReadRICH", 0 );
      vm_size(in);
      flagRICH = acReadRICH(KTstatus);
      vm_size(out); mem_rich += (out[0]-in[0]);
    }

    // ECAL
    if ((recoECAL != 0 || acFlagReadCnA != 0) && acFlagReadECAL == 1){
      ARTE_mark( "acReadECAL", 0 );
      vm_size(in);
      flagECAL = acReadECAL(KTstatus);
      vm_size(out); mem_ecal += (out[0]-in[0]);
    }

    // MUON
    if ((recoMUON != 0 || acFlagReadCnA != 0) && acFlagReadMUON == 1 ){
	ARTE_mark( "acReadMuon", 0 );
	vm_size(in);
	flagMUON = acReadMuon(KTstatus);
	vm_size(out); mem_muon += (out[0]-in[0]);
    }

    // GLOBAL. 
    if (acFlagReadGLOBAL) {
      ARTE_mark( "acReadGlobal", 0 );
      vm_size(in);
      flagGLOBAL = acReadGlobal(KTstatus);
      vm_size(out); mem_global += (out[0]-in[0]);
    }
  }


  // TARGET. Copy the target position to GTAR and apply alignment corrections
  // DTAR-->GTAR with alignment corrections should be done outside acReadTARGET 
  // and here only the retrieving of new global constants. 
  // For the moment no action for MC in acReadTARGET
  //
  ARTE_mark( "acReadTARGET", 0 );
  vm_size(in);
  flagTARGET = acReadTARGET(KTstatus);
  vm_size(out); mem_target += (out[0]-in[0]);

  
  //==================================================================
  // Set alca flag as the result of the acReadXXX functions
  // The function acRead() can be used to retrieve the alcaFlag. Some
  // packages could need to update internal buffers when new CnA constants
  // were loaded. They should check the lower 2 bytes.
  // The subdetectors are identified by the DAQ indentifier (component.h)
  //          COMP_M_TARG
  //          COMP_M_SVD
  //          COMP_M_ITR
  //          COMP_M_OTR
  //          COMP_M_HIPT
  //          COMP_M_MUON
  //          COMP_M_RICH
  //          COMP_M_ECAL
  //          COMP_M_TRD
  //          0x00000001  -> global alignment
  //          
  //     the following bits are set in alcaFlag
  //         COMP_M_xxx        if new constants were read in for the event
  //         COMP_M_xxx << 16  keep history of errors reading the constants
  //     
  //==================================================================
  
  unsigned int alcaUpdate = 0;

  // reset 2 lower bytes of alcaFlag, those used by other packages to check
  // whether there was an update of the contants. The 2 upper bytes are used to 
  // keep the history whether there was an error reading the constants

  alcaFlag &= 0xFFFF0000;      

  if (flagTARGET >= 1) alcaUpdate |=  COMP_M_TARG;
  if (flagVDS    >= 1) alcaUpdate |=  COMP_M_SVD;
  if (flagITR    >= 1) alcaUpdate |=  COMP_M_ITR;
  if (flagOTR    >= 1) alcaUpdate |=  COMP_M_OTR;
  if (flagRICH   >= 1) alcaUpdate |=  COMP_M_RICH;
  if (flagECAL   >= 1) alcaUpdate |=  COMP_M_ECAL;
  if (flagMUON   >= 1) alcaUpdate |=  COMP_M_MUON;
  if (flagGLOBAL >= 1) alcaUpdate |=  COMP_M_GLOBAL;

  alcaFlag   |= alcaUpdate;

  // successful update of CnA constants. Set subdetector bit
  if (flagTARGET >= 1) alcaFlag   |=   COMP_M_TARG   << 16;
  if (flagVDS    >= 1) alcaFlag   |=   COMP_M_SVD    << 16;
  if (flagITR    >= 1) alcaFlag   |=   COMP_M_ITR    << 16;
  if (flagOTR    >= 1) alcaFlag   |=   COMP_M_OTR    << 16;
  if (flagRICH   >= 1) alcaFlag   |=   COMP_M_RICH   << 16;
  if (flagECAL   >= 1) alcaFlag   |=   COMP_M_ECAL   << 16;
  if (flagMUON   >= 1) alcaFlag   |=   COMP_M_MUON   << 16;
  if (flagGLOBAL >= 1) alcaFlag   |=   COMP_M_GLOBAL << 16;

  // error reading CnA constants. Unset subdetector bit
  if (flagTARGET < 0) alcaFlag   &= ~(COMP_M_TARG   << 16);
  if (flagVDS    < 0) alcaFlag   &= ~(COMP_M_SVD    << 16);
  if (flagITR    < 0) alcaFlag   &= ~(COMP_M_ITR    << 16);
  if (flagOTR    < 0) alcaFlag   &= ~(COMP_M_OTR    << 16);
  if (flagRICH   < 0) alcaFlag   &= ~(COMP_M_RICH   << 16);
  if (flagECAL   < 0) alcaFlag   &= ~(COMP_M_ECAL   << 16);
  if (flagMUON   < 0) alcaFlag   &= ~(COMP_M_MUON   << 16);
  if (flagGLOBAL < 0) alcaFlag   &= ~(COMP_M_GLOBAL << 16);

  
  //==================================================================
  // update GEDE tables with global alignment if there was any internal 
  // alignment update
  //==================================================================

  if (alcaUpdate != 0 && ArteTable<MCIN>::size() == 0){
    ARTE_mark( "acGlobalAlignmentForGEDE", 0 );
    acGlobalAlignmentForGEDE(alcaUpdate);
  }



  // stop timing
  //
  cron_acread->stop();
  float cputime = cron_acread->lastTime();
  if( FATIME ) printf( "\033[1m TIME: ac=%4.2f \033[0m \n", cputime );

  // memory usage monitoring printout
  //
  if (FAMEM) printf("\033[31m acread memory usage (KB) : "
		    "VDS:%d, OTR:%d, ITR:%d, ECAL:%d, MUON:%d, RICH:%d, "
		    "GLOBAL:%d, TARGET:%d\033[0m\n", 
		    mem_vds/1024, mem_otr/1024, mem_itr/1024, mem_ecal/1024,
		    mem_muon/1024, mem_rich/1024, mem_global/1024, mem_target/1024); 

  ARTE_mark( "acread_ 1", 0 );
  return;
}


//*****************************************************************
// function to return alcaFlag
//*****************************************************************
unsigned int
acFlag()
{  
  return alcaFlag; 
}
