#ifndef ANL_EVSELECT 
#define ANL_EVSELECT

// *****************************************************************************
// 
//  Author:  Michael Schmelling                Created:  27-Oct-00
//           MPI-Heidelberg                    Modified:  1-Nov-00
//  Email:   Michael.Schmelling@mpi-hd.mpg.de  HERA-B Collaboration
//
// anl_evselect.hh
// header file for anl_evselect.C
//
// changes:
// 30 Oct 2000 (MS) redefined flags and extended anl_evselect_info
// 31 Oct 2000 (MS) bit-masks for target wire selection
//  1 Nov 2000 (MS) add DOC++ documentation
//  
// *****************************************************************************

#include "../CVStag.evselect"

// target wire definitions


/** @name Bitmask definitions for target wires
    Sets of wires can be specified by ORing 
    the masks for the individual wires. Some 
    groups are already defined.  
    @memo 
 */ 
//@{ 
/// station 1 above   
#define EVSEL_ABOVE1      1
/// station 1 below   
#define EVSEL_BELOW1      2
/// station 1 inner   
#define EVSEL_INNER1      4
/// station 1 outer   
#define EVSEL_OUTER1      8
/// station 2 above
#define EVSEL_ABOVE2     16
/// station 2 below   
#define EVSEL_BELOW2     32
/// station 2 inner
#define EVSEL_INNER2     64
/// station 2 outer
#define EVSEL_OUTER2    128
/// all wires of station 1
#define EVSEL_STATION1   15
/// all wires of station 2 
#define EVSEL_STATION2  240
/// all wires from both stations
#define EVSEL_ALL       255
//@}

// selection definitions
 
/** @name Bitmask definitions for return codes
    The actual return code in general will be 
    an OR of the bit-masks defined for the 
    specific conditions. 
    @memo 
 */
//@{
/// BX corresponds to an empty bunch
#define EVSEL_WRONGBX     1
/// too few VDS segments 
#define EVSEL_VDSTRACKS   2
/// no candidate primary vertex found
#define EVSEL_NOVERTEX    4
/// event is out of time
#define EVSEL_OUTOFTIME   8
//@}

/** Data structure for information exchange between user and 
    event selection function. Selection parameter are in 
    upper case, current values of the respective variables
    in lower case. In addition some statistical information
    is stored. For a detailed discussion please consult
    the functional description.
    @memo 
*/

struct anl_evselect_info {
  /// bx-number of current event (0-219)
  int   bx;             
  /// event selection flag
  int   evselectflag;   
  /// number of VDS track segments found
  int   ntrvds;         
  /// maximum transverse distance between tracks in candidate vertex
  float dvertex;        
  /// dvertex*dvertex
  float d2vertex;       
  /// x-coordinate of candidate primary vertex
  float xver;           
  /// y-coordinate of candidate primary vertex
  float yver;           
  /// z-coordinate of candidate primary vertex
  float zver;           
  /// average outer tracker  drifttime
  float ottmean;       
  /// rms outer tracker drifttime 
  float ottwidth;      
  /// total number of events processed
  int   nevtot;        
  /// coasting beam index
  float cb;            
  /// uncertainty of cb 
  float dcb;           
  /// fraction of accepted events
  float facc;           
  /// fraction of rejected events
  float freject;        
  /// fraction of events with wrong BX
  float fwrongbx;       
  /// fraction of events with too few VDS track segments
  float fvdstracks;     
  /// fraction of events without primary vertex candidate
  float fnovertex;      
  /// fraction of out-of-time events 
  float foutoftime;     
  /// offset between BX count and bunch pattern
  int   BXOFFSET;       
  /// requested mimimum number of hits per VDS track segment 
  int   MINHITS;        
  /// requested minumum number of VDS track segments
  int   NTRACKMIN;      
  /// maximum allowed transverse distance between tracks in a candidate primary vertex
  float DVERTEX;        
  /// DVERTEX*DVERTEX
  float D2VERTEX;       
  /// minimum accepted value for ottmean
  float OTTMEAN_MIN;    
  /// maximum accepted value for ottmean
  float OTTMEAN_MAX;   
  /// maximum accepted value for ottwidth
  float OTTWIDTH_MAX;  
};

typedef struct anl_evselect_info  anl_evselect_info;


/** The {\tt target} bitmask defines which wires are to be 
    cosidered in the primary vertex search. For details of 
    the algorithm please refer to the functional description.
  
    @param target bitmask specifying target wires 
    @param info   information exchange with calling program
    @return event selection flag 
    @author Michael Schmelling
    @memo 
 */

int anl_evselect(int target, anl_evselect_info & info);

#endif






