//----------------------------------------------------------------------------//
//
//   File:        EcalPretrigger.h
//
//   Description: simulation of the ECAL pretrigger
//
//   Author:      Michael Noerenberg
//                HERA-B Collaboration
//                DESY Hamburg
//                Notkestrasse 85
//                D-22603 Hamburg
//                Germany
//
//   Last Mod.:   13.1.2000
//
//   e-mail:      mino@mail.desy.de
//
//----------------------------------------------------------------------------//

#ifndef _EcalPretrigger_h
#define _EcalPretrigger_h

//----------------------------------------------------------------------------//

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <math.h>

//----------------------------------------------------------------------------//
class EcalPretrigger
{ 
private:
  EcalPretrigger();
  EcalPretrigger(const EcalPretrigger&);

  friend void every_class_needs_a_friend(); // to avoid compiler warnings ;-)

struct rectangle
  {
    float x_min;
    float y_min;
    float x_max;
    float y_max;

    rectangle(float x1, float y1, float x2, float y2):
      x_min(x1 < x2 ? x1 : x2),
      y_min(y1 < y2 ? y1 : y2),
      x_max(x1 < x2 ? x2 : x1),
      y_max(y1 < y2 ? y2 : y1)
      {};

    int containsPoint(float x, float y)
      { return x_min <= x && x <= x_max && y_min <= y && y <= y_max ? 1 : 0; }
  };
 
public:

  struct Cluster_t
  {
    float esum;
    float x;
    float y;
    float z;
    int   jpsiee;
    int   fdid;
 
    float ePx;
    float ePy;
    float ePz;
 
    Cluster_t(float ESUM, float X, float Y, float Z = 1350.0,
              int Jpsiee = 0, int Fdid = 0,
              float EPx = 0.0, float EPy = 0.0, float EPz = 0.0):
 
      esum(ESUM),
      x(X), y(Y), z(Z),
      jpsiee(Jpsiee), fdid(Fdid),
      ePx(EPx), ePy(EPy), ePz(EPz)
 
      {
      };
    
    bool operator == (const Cluster_t& rhs) const
      {
        float epsilon = 0.001;
        return (fabs(esum - rhs.esum) < epsilon * fabs(esum + rhs.esum)) &&
          (fabs(x - rhs.x) < epsilon * fabs(x + rhs.x)) &&
          (fabs(y - rhs.y) < epsilon * fabs(y + rhs.y)) &&
          (fabs(z - rhs.z) < epsilon * fabs(z + rhs.z)) &&
          (jpsiee == rhs.jpsiee);
      }
  }; 

  /**
   * some constants concerning the ECAL geometry
   */

  static float outerCellPitch;    // outer  cell pitch //
  static float middleCellPitch;   // middle cell pitch //
  static float innerCellPitch;    // inner  cell pitch //

  static float cellPitch[3];

  static const int   numInnerCellsX  = 70;
  static const int   numInnerCellsY  = 40;
  static const int   numMiddleCellsX = 80;
  static const int   numMiddleCellsY = 48;
  static const int   numOuterCellsX  = 56;
  static const int   numOuterCellsY  = 42;

  static int numCellsX[3];
  static int numCellsY[3];

  static float z_ECAL; // z-position of the ECAL //
 
  /**
   * internal state
   */

private:
  static int               m_MagnetOn;            // default: magnet off     //
  static float             m_Cut;                 // default: pt-cut 1.0 GeV //

  static int               m_GeoMode;             // default: accept all     //
  static std::vector<rectangle> m_ActiveRects;
  static std::vector<int>       m_ActiveBoards;        // list of installed pretrigger boards //
  static std::vector<Cluster_t> m_voClusters;

  static float m_Energie [3][80+2][48+2];
  static float m_Cluster [3][80+2][48+2];
  static int   m_Jpsiee  [3][80+2][48+2];
  static float m_EPx     [3][80+2][48+2];
  static float m_EPy     [3][80+2][48+2];
  static float m_EPz     [3][80+2][48+2];
  static int   m_Fdid    [3][80+2][48+2];

  static float m_highest_pt;
  static float m_2highest_pt;

public:
  static std::vector<Cluster_t>::iterator clusterIteratorBegin()
    {
      return m_voClusters.begin();
    }

  static std::vector<Cluster_t>::iterator clusterIteratorEnd()
    {
      return m_voClusters.end();
    }
 
  static float getHighestPt()
    {
      return m_highest_pt;
    };
 
  static float get2HighestPt()
    {
    return m_2highest_pt;
  };     

  static void setMagnetOn (int   magnetOn)
    {
       // cout << "*** EcalPretrigger::setMagnetOn(" << magnetOn << ") ***" << endl; 
       m_MagnetOn = magnetOn;
    };

  static void setCut      (float cut     )
    { 
       // cout << "*** EcalPretrigger::setCut(" << cut << ") ***" << endl; 
       m_Cut      = cut;
    };

  static void setGeoMode  (int   geoMode ) { m_GeoMode  = geoMode;  };
  
  static void addRectangle(float x1, float y1, float x2, float y2)
  { m_ActiveRects.push_back(rectangle(x1, y1, x2, y2)); };

  static void addBoard(int fdid)
  { m_ActiveBoards.push_back(fdid); };

  /** int readHITC()
   *
   * reads ARTE table HITC into internal cell memory
   * returns number of entries in HITC
   */

  static int readHITC();

  /** void dumpRects()
   *
   */

  static void dumpRects();

  /** void setup()
   *
   */

  static void setup();

  /** void reset()
   *
   * set energie of all cells to 0.0
   */

  static void reset();
   
  /** void dumpCells(int region, std::ostream& os = std::cout)
   *
   */

  static void dumpCells(int region,
                        std::ostream& os = std::cout);

  /** void addHit(float energie, float x_pos, float y_pos)
   *
   */

  static void addHit(float energie,
                     float x_pos,
                     float y_pos,
		     int   jpsiee = 0,
                     int   fdid   = 0,
                     float epx    = 0.0,
                     float epy    = 0.0,
                     float epz    = 0.0); 

  /** int searchTowers()
   *
   * searches towers in energie table and writes them in the cluster table
   * returns the number of found towers
   */

  static int searchTowers();

  /** int processTowers()
   *
   * reads HITC bank, searches towers in energie table and writes them in the cluster table
   * returns the number of found towers
   */

  static int processTowers();

  /** int searchClusters()
   *
   * searches clusters in energie table and writes them in the cluster table
   * returns the number of found clusters
   */

  static int searchClusters();

  /** int processClusters()
   *
   * searches clusters in energie table and writes them in the cluster table
   * returns the number of found clusters
   */

  static int processClusters();

  /** int getRegion(float x, float y)
   *
   * tests to which ECAL region (inner, middle or outer) a hit given by x and y belongs
   * output: 0 -> inner, 1 -> middle, 2 -> outer
   */

  static int getRegion(float x_pos,
                       float y_pos);

  /** void getCell(float x_pos, float y_pos, int& region, int& x_cell, int& y_cell)
   *
   * returns (via reference) the region and cell of a given position
   */

  static void getCell(float x_pos,
                      float y_pos,
                      int&  region,
                      int&  x_cell,
                      int&  y_cell);

  /** void getPosition(int region, int x_cell, int y_cell, float& x_pos, float& y_pos)
   *
   * returns (via reference) the position of a cell in a given region
   */

  static void getPosition(int    region,
                          int    x_cell,
                          int    y_cell,
                          float& x_pos,
                          float& y_pos);

  /** int testGeo(float x_pos, float y_pos)
   *
   */

  static int testGeo(float x_pos,
                     float y_pos);

  /** int testCut(float energie, float x_pos, float y_pos)
   *
   * depending on the value of m_MagnetOn apply Ktrig or pt cut
   */

  static int testCut(float energie,
                     float x_pos,
                     float y_pos);

  /** int testCut(float energie, int region, int x_cell, int y_cell)
   *
   * depending on the value of m_MagnetOn apply Ktrig or pt cut
   */

  static int testCut(float energie,
                     int   region,
                     int   x_cell,
                     int   y_cell);

};

//----------------------------------------------------------------------------//

#endif   // _EcalPretrigger_h //

//----------------------------------------------------------------------------//








