// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for a BEE event
//
// changes:
// 27 Jan 1997 creation by C. Borgmeier
// 20 Nov 2000 (TG) modified #include statements,
//                  added using namespace CLUE
// 07 Dec 2000 (TG) added #include "ClueRelation.hh"
// 12 Dec 2000 (TG) improved filling of TrackClone <-> RecoTrack relation
// 14 Dec 2000 (TG) code cleanup
// 22 Dec 2000 (TG) commented some printout
// 02 Jan 2001 (TG) moved code to class FillTable
// 02 Jan 2001 (TG) end() optimization, updateTOPP() in global namespace now
// 30 Mar 2001 (TG) added filling of Tcode, Ir, Bg
// 14 Aug 2001 (TG) bug fix in case of MC data
//
// ********************************************************************
#include <algorithm>
#include <farm/farm.hh> // for faDDAQevent_time();

#include "clue/cluearte.hh"
#include "clue/RtEvent.hh"
#include "clue/Topp.hh"

void updateTOPP(); // defined in PdtEntry.C

#ifdef NAMESPACE
using namespace CLUE;
#endif

//==============================================================================
// RtEvent::FillTopp
//==============================================================================
bool RtEvent::FillTopp() {

  //----------------------------------------------------------------------------
  // Fill Topp table
  //----------------------------------------------------------------------------
  // if Topp is not initialized yet, do it now
  if( toppLink==0 ) toppLink = new GClonesArray<Topp*>( 100 );
  if( toppLink->size() != ArteTable<TOPP>::size() ) {
#ifdef USE_ROOT
    toppLink->Delete();
#else
#warning missing code (C.B. 8/31/99)
#endif

    const ArteTableIterator<TOPP> end = ArteTable<TOPP>::end();
    for( ArteTableIterator<TOPP> topp = ArteTable<TOPP>::begin();
	 topp != end; ++topp ) {

      ArteTable<Topp>::push_back( Topp(*topp) );
    }
    updateTOPP(); // defined in PdtEntry.C

#ifdef ClueDEBUG
    cout << "Print Topp table: " << endl;
    const ArteTableIterator<Topp> tend = ArteTable<Topp>::end();
    for( ArteTable<Topp>::iterator tp = ArteTable<Topp>::begin();
	 tp != tend; ++tp) {
      cout << **tp << endl;
    }
#endif
  }

  return true;
} // RtEvent::FillTopp


//==============================================================================
// RtEvent::FillTables()
//==============================================================================
bool RtEvent::FillHeader() {

  //----------------------------------------------------------------------------
  // fill Event header from EVHD & DIHD
  //----------------------------------------------------------------------------
  if( ArteTable<EVHD>::size()==1 ) {
    ArtePointer<EVHD> evhd = *ArteTable<EVHD>::begin();

    Evt      = evhd->evt;
    Run      = evhd->run;
    Exp      = evhd->exp;
    Vfld     = evhd->vfld;
    Flagdst  = evhd->flagdst;

    Dbmajor  = evhd->dbmajor;
    Dbminor  = evhd->dbminor;

    Flag     = evhd->flag;
    Unixtime = faDDAQevent_time();
  } else {
    Evt = Run = Exp = 0;
    Vfld = 0;
    Flagdst = Dbmajor = Dbminor = Flag = Unixtime = 0;
  }
  
  if( ArteTable<DIHD>::size()==1 ) {
    Bx      = (*ArteTable<DIHD>::begin())->bx;
  } else Bx = 0;

  Tcode = faDDAQFCS_tcode();
  Ir    = faDTARia();
  Bg    = faDTARbg();

  // Cherenkov radius for beta = 1
  if( ArteTable<RITH>::size() > 0 ) {
    Rad0    = (*ArteTable<RITH>::begin())->rad0;
  } else Rad0 = 0;

  // Fill Hits in detector components
  if( ArteTable<EVHI>::size() != 9 ) {
    if( ArteTable<MCIN>::size() == 0 ) {
      // indeed real data
      cerr << "RtEvent::FillHeader error: ArteTable EVHI size is "
	   << ArteTable<EVHI>::size() << " instead of 9" << endl;
    }
  } else {
    Hvds = ArtePointer<EVHI>(1)->hit;
    Hwch = ArtePointer<EVHI>(2)->hit;
    Hmsg = ArtePointer<EVHI>(3)->hit;
    Hptc = ArtePointer<EVHI>(4)->hit;
    Hric = ArtePointer<EVHI>(5)->hit;
    Htrd = ArtePointer<EVHI>(6)->hit;
    Hcal = ArtePointer<EVHI>(7)->hit;
    Hmuo = ArtePointer<EVHI>(8)->hit;
    Hmux = ArtePointer<EVHI>(9)->hit;
  }

  return true;
} // RtEvent::FillHeader
