// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   16.07.2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Particle ID routine for RTRAS.
//
// changes:
// 16 Jul 2001 (TG) creation from Michael Schmellings code
// 28 Sep 2001 (TG) added PID_init() function
// 10 Jan 2001 (TG) set initial likelihoods to 0.2 to avoid division by zero
//                  problems. The meaning doesn't change: all hypothesises are
//                  of equal likelihood.
//
// **********************************************************************
#include "clue/cluearte.hh"
#include "smatrix/Functions.hh"
#include "clue/PID.hh"
#include "cluearte/checkPID.hh"
#include "hbook.h"

static const int ID_RICH     = 7000;
static bool  PID_monitoring  = false;

/** @name HBOOK initalitaztion

    Initialize HBOOK histos for PID monitoring. This function should be called in
    the ARTE usinit() function.

    @author T. Glebe
 */
//==============================================================================
// PID_init
//==============================================================================
void PID_init() {

  PID_monitoring = true;

  hbook1(ID_RICH+ 1,"msq  - all"      ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 2,"msq  - electrons",1200,-0.1,1.1);
  hbook1(ID_RICH+ 3,"msq  - muons"    ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 4,"msq  - pions"    ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 5,"msq  - kaons"    ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 6,"msq  - protons"  ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 7,"msq  - none"     ,1200,-0.1,1.1);
  hbook1(ID_RICH+ 8,"PID",20,-0.5,19.5);
  hbook2(ID_RICH+ 9,"theta**2 vs 1/p**2 - all ",100,0.,0.1,70,0.,0.0035);
  hbook2(ID_RICH+10,"theta**2 vs 1/p**2 - none",100,0.,0.1,70,0.,0.0035);

}

/** @name Particle Identification algorithm

    Strategy: Basic PID-device is the RICH detector. If
    information from RISE is present, it will be used.
    Otherwise likelihoods from RITER are used. In case
    neither leads to a positive identification despite the
    fact that a ring is associated to the track, some
    private code is used. (The code in many cases gives
    results similar to the output of RISE.) Rare light
    particles (e,mu) need a RICH likelihood above 0.99
    to be identified. Pions are identified (essentially by
    default) if the likelihood exceeds 0.3. Kaons are
    identified if the likelihood is larger than 0.6
    and protons are required to have a likelihood larger
    than 0.3. If hits in the muon system are present,
    then a particle with muon-likelihood above 0.0 is
    identified as muon. Also, if no RICH information is
    present, a track with hits in the muon system is
    identified as a muon.
    
    @author Michael Schmelling 15-Mar-2001
 */
//==============================================================================
// checkPID
//==============================================================================
const PID_data& checkPID(const ArtePointer<RTRA>& rtra) {

  const ArtePointer<RTRI> rtri = getPTR<RTRI,RTRA>(rtra);

  static const float THO2    = square(0.0521);  // cerenkov angle squared for beta=1 particles
  static const float DTHCER2 = square(0.00035); // assumed uncertainty of THO

  static PID_data pid;
  int PID = 0;
  float Lelectron = 0.2, Lmuon = 0.2, Lpion = 0.2, Lkaon = 0.2, Lproton = 0.2;

  if(rtri->flt > 0) { // if RITER is present
    if(rtri->lte  > 0.99) PID = PID::electron;
    if(rtri->ltmu > 0.99) PID = PID::muon;
    if(rtri->ltpi > 0.30) PID = PID::pion;
    if(rtri->ltk  > 0.60) PID = PID::kaon;
    if(rtri->ltp  > 0.30) PID = PID::proton;
    Lelectron = rtri->lte;
    Lmuon     = rtri->ltmu;
    Lpion     = rtri->ltpi;
    Lkaon     = rtri->ltk;
    Lproton   = rtri->ltp;
  } 
  else {
    if(rtri->fls > 0) { // if RISE is present
      if(rtri->lse  > 0.99) pid.PID = PID::electron;
      if(rtri->lsmu > 0.99) pid.PID = PID::muon;
      if(rtri->lspi > 0.30) pid.PID = PID::pion;
      if(rtri->lsk  > 0.60) pid.PID = PID::kaon;
      if(rtri->lsp  > 0.30) pid.PID = PID::proton;
      Lelectron = rtri->lse;
      Lmuon     = rtri->lsmu;
      Lpion     = rtri->lspi;
      Lkaon     = rtri->lsk;
      Lproton   = rtri->lsp;
    }
  } // if flt > 0

  //  cout << "PID from RISE/RITER: " << PID << endl;

  const float Cer2 = square(rtra->cer);
  const float P2   = square(1./rtra->pf);

  // if neither RISE nor RITER present:

  if(Cer2 > 0 && PID==0) {
    const float Melectron2 = square(0.000511);
    const float Mmuon2     = square(0.105658);
    const float Mpion2     = square(0.1396);
    const float Mkaon2     = square(0.494);
    const float Mproton2   = square(0.938);
    const float dpsq       = rtra->cvf[14];  //1/dp^2
    const float Msq        = P2*(THO2 - Cer2);
    const float vMsq       = 4.*P2*(Msq*Msq*dpsq + P2*Cer2*DTHCER2);

    Lelectron = square(Msq - Melectron2)/vMsq;
    Lmuon     = square(Msq - Mmuon2)/vMsq;
    Lpion     = square(Msq - Mpion2)/vMsq;
    Lkaon     = square(Msq - Mkaon2)/vMsq;
    Lproton   = square(Msq - Mproton2)/vMsq;
    if(Lelectron < 3) Lelectron = exp(-0.5*Lelectron); else Lelectron = 0.;
    if(Lmuon     < 3) Lmuon     = exp(-0.5*Lmuon);     else Lmuon     = 0.;
    if(Lpion     < 3) Lpion     = exp(-0.5*Lpion);     else Lpion     = 0.;
    if(Lkaon     < 3) Lkaon     = exp(-0.5*Lkaon);     else Lkaon     = 0.;
    if(Lproton   < 3) Lproton   = exp(-0.5*Lproton);   else Lproton   = 0.;
    const float Lsum = Lelectron + Lmuon + Lpion + Lkaon + Lproton;
    if(Lsum > 0) {
      Lelectron /= Lsum;
      Lmuon     /= Lsum;
      Lpion     /= Lsum;
      Lkaon     /= Lsum;
      Lproton   /= Lsum;
    }
    if(Lelectron > 0.99) PID = PID::electron;
    if(Lmuon     > 0.99) PID = PID::muon;
    if(Lpion     > 0.30) PID = PID::pion;
    if(Lkaon     > 0.60) PID = PID::kaon;
    if(Lproton   > 0.30) PID = PID::proton;
  } // Cer2>0 && PID==0

  // if no RICH information is available - believe the muon system
  if(PID != PID::kaon && PID != PID::proton && rtra->hmu>0) PID = PID::muon;

  //  cout << "FINAL PID: " << PID << endl;

  // save PID information
  pid.clear();
  pid.PID       = PID;
  pid.Lelectron = Lelectron;
  pid.Lmuon     = Lmuon;
  pid.Lpion     = Lpion;
  pid.Lkaon     = Lkaon;
  pid.Lproton   = Lproton;

  // MONITORING
  if(PID_monitoring) {
    hf1(ID_RICH+8,PID,1.);
    if(Cer2 > 0) {
      const float Msq = P2*(THO2 - Cer2);
      hf1(ID_RICH+1, Msq, 1.);
      hf2(ID_RICH+9, 1./P2, Cer2, 1.);
      if(PID==PID::electron) hf1(ID_RICH+2,Msq,1.);
      if(PID==PID::muon    ) hf1(ID_RICH+3,Msq,1.);
      if(PID==PID::pion    ) hf1(ID_RICH+4,Msq,1.);
      if(PID==PID::kaon    ) hf1(ID_RICH+5,Msq,1.);
      if(PID==PID::proton  ) hf1(ID_RICH+6,Msq,1.);
      if(PID==0) {
	hf1(ID_RICH+7,Msq,1.);
	hf2(ID_RICH+10, 1./P2, Cer2, 1.);
      } // if PID==0
    } // if Cer2 > 0
  } // if PID_monitoring

  return pid;
}
