// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   16. Jul 1999
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: clueWriteEvent() writes the current ARTE event into a BEE file.
//              The filename must be given by the kuip command
//
//              CLUE/SETFILENAME xxx
//
//              before the first invocation.
//
// changes:
// 16 Jul 1999 creation by C. Borgmeier
// 22 Nov 2000 (TG) added using namespace CLUE
// 14 Dec 2000 (TG) replaced clueBranch() call by ClueTTree() class
// 18 Dec 2000 (TG) changed ttree type to ClueTTree*
// 02 Jan 2001 (TG) removed obsolete include files
// 02 Jan 2001 (TG) added TableFillBase parameter to clueWriteEvent()
// 03 Jan 2001 (TG) rewritten clueWriteEvent(), clueGetEvent()
// 04 Jan 2001 (TG) replaced clueEvent->Clear() by clueEvent->Delete()
// 03 Dec 2001 (TG) introduced IgnoreTObjectStreamer() calls to reduce file size
// 04 Dec 2001 (TG) improvement in ClueGetEvent()
//
// ********************************************************************

#include <string>

#ifdef USE_ROOT
// ROOT includes
#include <TROOT.h>
#include <TFile.h>
//#include <TTree.h>
#include "clue/ClueTTree.hh"
#include <TBranch.h>
#endif

#include "clue/cluearte.hh"
#include "clue/RtEvent.hh"
#include "clue/StreamObject.hh"
#include "clue/RecoTrack.hh"
#include "clue/RecoVertex.hh"
#include "clue/MC_Track.hh"
#include "clue/MC_Vertex.hh"
#include "clue/Topp.hh"
#include "clue/Wire.hh"
#include "clue/EcalHit.hh"
#include "clue/TrackClones.hh"
#include "cluearte/TableFillBase.hh"

#ifdef NAMESPACE
using namespace CLUE;
#endif

extern "C" void kugets_( char*, int *, int );
extern "C" void kugeti_( int& );
extern "C" void clkuip_();
void atregkuip( void (*function)(), int priority );


#ifdef USE_ROOT
static TFile* cluefile = 0;


//
// this object writes the CLUE file on program termination
//
static class ClueCloser
{
public:
  ClueCloser()
    :fROOT( "recon", "recon" )
    {
      //if( !TROOT::Initialized() ) {
      //fROOT = new TROOT( "recon", "recon" );
      //}
    }
  ~ClueCloser()
    { if( cluefile!=0 ) cluefile->Write(); }
  
  TROOT fROOT;
} _clueCloser;
#endif /* USE_ROOT */


//
// this object initializes the CLUE kuip commands
//
static class ClueIniter
{
public:
  ClueIniter()
    {
      atregkuip( clkuip_, 10 );
    } 
} _clueIniter;


//
//
//


static std::string beefilename;

//
// cluesetfilename
//
extern "C" void cluesetfilename_()
{
  int len;
  int len1 = 512;
  char input[512];
  kugets_(input, &len, len1);
  
  beefilename = std::string( input, len );
}


//
//
//
static bool storeRSEG_Flag    = false;
extern "C" void cluestorersegflag_()
{
  int flag;
  kugeti_( flag );

  storeRSEG_Flag = flag;
}


RtEvent* clueEvent          = 0;
#ifdef USE_ROOT
static ClueTTree* ttree     = 0;
static TTree* toppTree      = 0;
static int initialCompLevel = 2;


//
//
//
static int clueCreateFile()
{
  if( !beefilename.empty() ) {
    
    // avoid writing fBits and fUniqueID
    // possible in ROOT 3.0x
    RtEvent::Class()->IgnoreTObjectStreamer();
    StreamObject::Class()->IgnoreTObjectStreamer();
    RecoTrack::Class()->IgnoreTObjectStreamer();
    RecoVertex::Class()->IgnoreTObjectStreamer();
    MC_Track::Class()->IgnoreTObjectStreamer();
    MC_Vertex::Class()->IgnoreTObjectStreamer();
    Topp::Class()->IgnoreTObjectStreamer();
    Wire::Class()->IgnoreTObjectStreamer();
    EcalHit::Class()->IgnoreTObjectStreamer();
    TrackClones::Class()->IgnoreTObjectStreamer();
    
    if( clueEvent==0 ) clueEvent = new RtEvent();
    
    cluefile = new TFile( beefilename.c_str(),"RECREATE",
			  "Clue ntuple");
    cluefile->SetCompressionLevel( initialCompLevel );
    
    ttree = new ClueTTree("Clue","CLUE tree");

    //    ttree = new TTree("Clue","CLUE tree");
    //ttree->SetAutoSave(1000000);  // autosave when 1 Mbyte written
    Int_t split = 1;
    Int_t bufsize = 65536;
    if (split)  bufsize /= 4;

    //    clueBranch( ttree, "event", "RtEvent", &clueEvent, bufsize, split);
    ttree->Branch("event","RtEvent", &clueEvent, bufsize, split);

    cluefile->cd();
    toppTree = new TTree( "constants", "constants" );
    
    toppTree->Branch( "toppTable", &RtEvent::toppLink, bufsize, 1 );

    return 0;
  }
  return -1;
}
#endif


//
// cluesetbranchstatus
//
extern "C" void cluesetbranchstatus_()
{
  int len;
  int len1 = 511;
  char input[512];
  kugets_( input, &len, len1 );
  input[len] = 0;

  int flag;
  kugeti_( flag );

#ifdef USE_ROOT
  if( ttree==0 ) {
    if( clueCreateFile()!=0 || ttree==0 ) return;
  }
  
  ttree->SetBranchStatus( input, flag );
#endif
}


extern "C" void cluesetcompressionlevel_()
{
  int level;
  kugeti_( level );
  if( level < 0 || level > 9 ) {
    
  } else {
#ifdef USE_ROOT
    if( cluefile!=0 ) {
      cluefile->SetCompressionLevel( level );
    } else {
      initialCompLevel = level;
    }
#endif
  }
}


//
// write the current ARTE event
//
int clueWriteEvent( const TableFillBase& tf ) {

  if( !beefilename.empty() ) {
#ifdef USE_ROOT
    static bool firstcall = true;

    if( firstcall ) {
      if( clueCreateFile()!=0 ) return -1;

      firstcall = false;
      clueEvent->FillTopp();
      toppTree->Fill();       // fill Topp
    }

    clueEvent->Delete();
    if( tf.EventSelect() ) {
      //      clueEvent->Clear();
      clueEvent->FillHeader(); // fill event header
      tf.Fill();               // fill ArteTables
      if(storeRSEG_Flag == true) {
	tf.FillRSEG();
      }
      ttree->Fill();

#ifdef XXX
      if( firstcall ) {
	firstcall = false;
	clueEvent->FillTopp();
	toppTree->Fill();       // fill Topp
      }
#endif    
    }
#endif
  }
  return 0;
}


//
// get the current ARTE event as RtEvent object
//
RtEvent* clueGetEvent( const TableFillBase& tf ) {

  if( clueEvent==0 ) {
    clueEvent = new RtEvent();
    clueEvent->FillTopp();
  }

  clueEvent->Delete();
  clueEvent->FillHeader(); // fill event header
  tf.Fill();               // fill ArteTables
  if(storeRSEG_Flag == true) {
    tf.FillRSEG();
  }

  return clueEvent;
}
