// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   01.10.2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Code to select proper photon candidates
//
// changes:
// 01 Okt 2001 (TG) creation from Michael Schmellings code
//
// **********************************************************************
#include <map>
#include "clue/cluearte.hh"
#include "clue/EcalHit.hh"
#include "clue/PID.hh"
#include "clue/RecoTrack.hh"
#include "hbook.h"

static const int ID_ECAL      = 8000;
static bool  EHIT_monitoring  = false;

const float WIDTHmax  =  9.;     // ECAL cluster width cut
const float ESUMmin   =  1.;     // ECAL cluster energy cut
const float RECAL     =  3.;     // radius for track-associated energy

// RCCL <-> EcalHit map
static map<ArtePointer<RCCL>,ArtePointer<EcalHit> > rcclmap;

//==============================================================================
// fTransport: return (x,y) of a track transported to z (behind magnet)
//==============================================================================
void  fTransport(const ArtePointer<RTRA>& rtra, float z, float &x, float &y) {

  double qp = rtra->pf;
  double xf = rtra->xf;
  double yf = rtra->yf;
  double zf = rtra->zf;
  double tx = rtra->txf;
  double ty = rtra->tyf;

  // extrapolate track to the kick-plane

  double zk = 506.683 + qp*16.6829;
  double dz = zk - zf;

  y = yf + dz*ty;
  x = xf + dz*tx;

  // determine the field integrals Bx and By

  double Bx = -0.005035 - qp*0.00084925;
  double By = -0.671728 - qp*0.00974943;
  Bx += (-0.0000011697+qp*0.000041778)*x + (0.0000896607-qp*0.00036773)*y;
  By += ( 0.0000999526+qp*0.000665649)*x + (0.0000274325-qp*0.00004491)*y;
  Bx += 0.00000083922*x*x + 0.0000176894*x*y - 0.0000011782*y*y;
  By += 0.00000821807*x*x - 0.0000019324*x*y - 0.0000145821*y*y;

  // calculate the track direction after the kink

  double B   = sqrt(Bx*Bx+By*By);
  double bx  = Bx/B;
  double by  = By/B;
  double arg = B*qp;
  double c   = cos(arg);
  double s   = sin(arg);
  double cx = (1.-c)*bx;
  double cy = (1.-c)*by;

  double n0 = (cx*bx+c)*tx +  cx*by   *ty - s*by;
  double n1 =  cy*bx   *tx + (cy*by+c)*ty + s*bx;
  double n2 =   s*by   *tx -  s*bx    *ty + c;

  // and extrapolate to the end point

  dz = z - zk;
  x += dz*n0/n2;
  y += dz*n1/n2;

  return;
}

//==============================================================================
// EHIT_init
//==============================================================================
void EHIT_init() {

  EHIT_monitoring = true;

  hbook1(ID_ECAL+1,"cluster width"  , 20, 0.5, 20.5);
  hbook1(ID_ECAL+2,"ECAL hits - all",120, 0.5,120.5);
  hbook1(ID_ECAL+3,"ECAL hits - acc",120, 0.5,120.5);
  hbook2(ID_ECAL+4,"dy vs dx  wrt track impact",100,-25.,25.,100,-25.,25.);
  hbook1(ID_ECAL+5,"1/r dn/dr wrt track impact",100,  0.,25.);
  hbook2(ID_ECAL+6,"dy vs dx  wrt track-extrap",100,-25.,25.,100,-25.,25.);
  hbook1(ID_ECAL+7,"1/r dn/dr wrt track-extrap",100,  0.,25.);

  for(int i=1; i<=WIDTHmax; ++i) {
    hbook1(ID_ECAL+10+i,"lg(E)(width(id))",100,-2.,3.);
    hbook1(ID_ECAL+30+i,"asymmetry(width(id))",100,0.,10.);
  }

}

//==============================================================================
// selectRCCL: select electron/photon candidates 
//==============================================================================
void selectRCCL() {

  static const float ASYMa     =  1.0;    // offset for ECAL asymmetry cut
  static const float ASYMb     =  0.3333; // width-slope for ECAL asymmetry cut

  int  nrccl = 0;  // total number of clusters
  int  negs  = 0;  // number e/gamma showers

  const ArteTableIterator<RCCL> eend = ArteTable<RCCL>::end();
  for(ArteTable<RCCL>::iterator itrccl = ArteTable<RCCL>::begin();
      itrccl != eend; ++itrccl) {

    itrccl->flag = 0;
    nrccl       += 1;

    int   width = int(itrccl->wid);
    float esum  = itrccl->esum;
    float asym  = itrccl->asym;

    if(EHIT_monitoring) { hf1(ID_ECAL+1,width,1.); }
    if(width> WIDTHmax         ) continue;
    if(EHIT_monitoring) { hf1(ID_ECAL+10+width,log10(esum),1.); }
    if(esum < ESUMmin          ) continue;
    if(EHIT_monitoring) { hf1(ID_ECAL+30+width,asym,1.); }
    if(asym > ASYMa+width*ASYMb) continue;

    itrccl->flag = 1;
    negs        += 1;
  }

  if(EHIT_monitoring) {
    hf1(ID_ECAL+2,nrccl,1.);
    hf1(ID_ECAL+3,negs, 1.);
  }

  return;
}


//==============================================================================
// assignClusters
//==============================================================================
void assignClusters(const ArtePointer<RecoTrack>& rtrack ) {

  const ArtePointer<RTRA> rtra = rtrack->rtra();
  float xecal = 0;
  float yecal = 0;
  fTransport(rtra, 1344., xecal, yecal);
  
  float esum = 0.;
  const ArteTableIterator<RCCL> eend = ArteTable<RCCL>::end();
  for(ArteTable<RCCL>::iterator itrccl = ArteTable<RCCL>::begin();
      itrccl != eend; ++itrccl) {
    if(itrccl->flag == 0) continue;
    float dx = xecal - itrccl->x;
    float dy = yecal - itrccl->y;
    float dr = sqrt(dx*dx+dy*dy);
    if(EHIT_monitoring) {
      hf2(ID_ECAL+4,dx,dy,1.);
      hf1(ID_ECAL+5,dr,1./dr);
    }
    if(dr<RECAL) {
      esum += itrccl->esum;
      itrccl->flag += 10;
    }
  }

  // misuse mass for eoverp
  rtrack->set_mass(esum/rtrack->p());
  
  return;
}

//==============================================================================
// findBremsstrahlung
//==============================================================================
void findBremsstrahlung() {

  const ArteTableIterator<RecoTrack> rend = ArteTable<RecoTrack>::end();
  for( ArteTableIterator<RecoTrack> rit = ArteTable<RecoTrack>::begin();
       rit != rend; ++rit ) {
    const ArtePointer<RecoTrack> rtra = *rit;

    float ebrems = 0.;

    const ArteTableIterator<RCCL> eend = ArteTable<RCCL>::end();
    for(ArteTable<RCCL>::iterator itrccl = ArteTable<RCCL>::begin();
        itrccl != eend; ++itrccl) {
      if(itrccl->flag == 0) continue;

      // straigth line extrapolation behind magnet => find bremsstrahlung
      // which was emitted before magnet
      float xecal = rtra->x() + (1344.- rtra->z())*rtra->tx();
      float yecal = rtra->y() + (1344.- rtra->z())*rtra->ty();

      float dx = (xecal - itrccl->x)*rtra->charge();
      float dy =  yecal - itrccl->y;
      float dr = sqrt(dx*dx+dy*dy);

      // misused mass: eoverp
      if(EHIT_monitoring && rtra->mass() > 0.5 && itrccl->flag==1) {
        hf2(ID_ECAL+6,dx,dy,1.);
        hf1(ID_ECAL+7,dr,1./dr);
      }
      if(rtra->mass() > 0.5 && dr<RECAL) {
        ebrems += itrccl->esum;
        itrccl->flag += 100;
      }
    } // for rccl

    rtra->set_ebrems(ebrems);
  } // for RecoTrack

  return;
}

//==============================================================================
// fillEcalHitTable
//==============================================================================
void fillEcalHitTable() {

  rcclmap.clear();

  const ArteTableIterator<RCCL> eend = ArteTable<RCCL>::end();
  for(ArteTable<RCCL>::iterator itrccl = ArteTable<RCCL>::begin();
      itrccl != eend; ++itrccl) {

    if(itrccl->flag != 1) continue;
    const ArtePointer<RCCL> rccl = *itrccl;

    const ArtePointer<EcalHit> ehit = 
      ArteTable<EcalHit>::push_back( EcalHit(rccl) );
    rcclmap[rccl] = ehit;
  }

  const map<ArtePointer<RCCL>,ArtePointer<EcalHit> >::iterator mend =
    rcclmap.end();

  // set EcalHit relation of RecoTrack's
  const ArteTableIterator<RecoTrack> rend = ArteTable<RecoTrack>::end();
  for(ArteTableIterator<RecoTrack> rit = ArteTable<RecoTrack>::begin();
      rit != rend; ++rit) {
    const ArtePointer<RecoTrack> rtrack = *rit;
    const ArtePointer<RTRA> rtra = rtrack->rtra();

    // set relation to Ecal cluster
    const ArteRelationIterator<RCCL> rcend = rtra->rccl.end();
    for(ArteRelationIterator<RCCL> rcit = rtra->rccl.begin(); rcit != rcend;
	++rcit) {
      const ArtePointer<RCCL> rccl = *rcit;

      if(rccl->flag == 1 && rcclmap.find(rccl) != mend) {
	createRelation(rtrack, rcclmap[rccl]);
      }
    } // for RCCL

    // perform final PID
    // =================
    // misuse mass() as eoverp
    // =================
    // muons must not have associated ECAL energy
    // electrons with RICH must have  e/p > 0.5
    // electrons without RICH require e/p > 0.8
    // for identified electrons do bremsstrahlung recovery
    if(rtrack->mass()>0.0 && rtrack->pid() == PID::muon)
      rtrack->set_pid(PID::pion);
    if(rtrack->mass()>0.5 && rtrack->pid() == PID::electron)
      rtrack->set_pid(PID::pion);
    if(rtrack->mass()>0.8 && rtrack->pid() == 0)
      rtrack->set_pid(PID::electron);

    if(rtrack->pid() == PID::electron) {
      rtrack->ebrems_recovery();
    }

    // finally, set mass back to 0!
    rtrack->set_mass(0.);
  } // for RecoTrack

  return;
}
