// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for a BEE event
//
// changes:
// 27 Jan 1997 creation by C. Borgmeier
// 20 Nov 2000 (TG) modified #include statements,
//                  added using namespace CLUE
// 07 Dez 2000 (TG) added #include "ClueRelation.hh"
// 12 Dez 2000 (TG) improved filling of TrackClone <-> RecoTrack relation
//
// ********************************************************************
#include <algorithm>
#include <set>
#include <multimap.h>

#include "clue/cluearte.hh"
#include "clue/StreamObject.hh"  // needed for RtEvent.hh
#include "gate/ObjArray.hh"      // needed for RtEvent.hh
#include "gate/ClonesArray.hh"   // needed for RtEvent.hh
#include "geometry/Point3D.hh"   // needed for ClueUtil.hh
#include "clue/RtEvent.hh"

#include "clue/Wire.hh"
#include "clue/EcalHit.hh"
#ifdef USE_GATE
#include "clue/RecoTrack.hh"     // needed by MC_Track.hh
#endif
#include "geometry/ClLorentzVector.hh"
#include "clue/MC_Track.hh"
#include "clue/MC_Vertex.hh"
#include "clue/RecoVertex.hh"
#include "clue/TrackClones.hh"
#include "clue/CustomTrack.hh"
#include "clue/Topp.hh"
#include "clue/SmartPointer.hh"  // for Pattern.hh
#include "clue/Pattern.hh"       // for PdtEntry.hh
#include "clue/PdtEntry.hh"
#include "grelation/ClueRelation.hh"

#ifdef NAMESPACE
using namespace CLUE;
#endif

#include <farm/farm.hh> // for faDDAQevent_time();


//==============================================================================
// trait to configure fillConeMap & createCloneRelation
//==============================================================================
template <class T> 
struct R_trait {
  typedef const int                    First_t;
  typedef const ArtePointer<T>         Second_t;
  typedef multimap<First_t, Second_t>  Cl_map;
  typedef Cl_map::value_type           value_type;
  typedef Cl_map::iterator             Cl_mapit;
  typedef Cl_map::const_iterator       Cl_mapcit;
};


//==============================================================================
// fillCloneMap
//==============================================================================
template <class T>
void fillCloneMap(typename R_trait<T>::Cl_map& clonemap,
		  typename R_trait<T>::First_t key,
		  typename R_trait<T>::Second_t ptr) {
  typedef typename R_trait<T>::value_type value_type;

  clonemap.insert(value_type(key,ptr)); 
}


//==============================================================================
// createCloneRelation
//==============================================================================
template <class T> 
void createCloneRelation(typename R_trait<T>::Cl_map& clonemap,
			 unsigned int id) {
  typedef typename R_trait<T>::First_t   First_t;
  typedef typename R_trait<T>::Cl_mapcit Cl_mapcit;

  for(Cl_mapcit it = clonemap.begin(); it != clonemap.end();) {
    const First_t key = it->first;
    //    cout << " key: " << key << endl;

    if(clonemap.count(key) > 1) {
      //      cout << " ===========> count: " << clonemap.count(key) << endl;

      const ArtePointer<TrackClones> clones
	= ArteTable<TrackClones>::push_back( TrackClones( id ));
      
      it = clonemap.upper_bound(key);

      // loop over clones
      for(Cl_mapcit it2 = clonemap.lower_bound(key);
	  it2 != it; ++it2){
	createRelation(clones, it2->second);
      }
      //      --it;
    } else {
      ++it;
    }
  } // for
} // end of createCloneRelation


//==============================================================================
// RtEvent::init
//==============================================================================
int RtEvent::init() {
  Clear();

  //----------------------------------------------------------------------------
  // TOPP
  //----------------------------------------------------------------------------
  // if Topp is not initialized yet, do it now
  if( toppLink==0 ) toppLink = new GClonesArray<Topp*>( 100 );
  if( toppLink->size() != ArteTable<TOPP>::size() ) {
#ifdef USE_ROOT
    toppLink->Delete();
#else
#warning missing code (C.B. 8/31/99)
#endif

    for( ArteTable<TOPP>::iterator topp = ArteTable<TOPP>::begin();
	 topp != ArteTable<TOPP>::end(); ++topp ) {

      //      cout << " Topp: " << Topp(*topp) << endl;
      ArteTable<Topp>::push_back( Topp(*topp) );
    }
    ClPdtEntry::updateTOPP();

#ifdef ClueDEBUG
#endif
    cout << "Print Topp table: " << endl;
    for( ArteTable<Topp>::iterator tp = ArteTable<Topp>::begin();
	 tp != ArteTable<Topp>::end(); ++tp) {
      cout << **tp << endl;
    }

  }
  
  //----------------------------------------------------------------------------
  // EVHD
  //----------------------------------------------------------------------------
  if( ArteTable<EVHD>::size()==1 ) {
    ArtePointer<EVHD> evhd = *ArteTable<EVHD>::begin();

    Evt      = evhd->evt;
    Run      = evhd->run;
    Exp      = evhd->exp;
    Vfld     = evhd->vfld;
    Flagdst  = evhd->flagdst;

    Dbmajor  = evhd->dbmajor;
    Dbminor  = evhd->dbminor;

    Flag     = evhd->flag;
    Unixtime = faDDAQevent_time();
  } else {
    Evt = Run = Exp = 0;
    Vfld = 0;
    Flagdst = Dbmajor = Dbminor = Flag = Unixtime = 0;
  }
  
  if( ArteTable<DIHD>::size()==1 ) {
    Bx      = (*ArteTable<DIHD>::begin())->bx;
  } else Bx = 0;
  
 
  //----------------------------------------------------------------------------
  // GTAR
  //----------------------------------------------------------------------------
  for( ArteTableIterator<GTAR> gtar = ArteTable<GTAR>::begin();
       gtar != ArteTable<GTAR>::end(); ++gtar ) {

    // select inserted wires
    if( fabs(gtar->x) < 1.2 && fabs(gtar->y) < 1.2) {
      ArteTable<Wire>::push_back( Wire(*gtar) );
    }
    //    Wire w( *gtar );
    //    Wires->push_back( &w );
  }

  cout << "Wires bank:" << endl;
  for(ArteTable<Wire>::iterator it = ArteTable<Wire>::begin(); 
      it != ArteTable<Wire>::end(); ++it) {
    cout << **it << endl;
  }
#ifdef XXX
  for(GClonesArray<Wire*>::const_iterator it=Wires->begin(); it!=Wires->end();
      ++it) {
    cout << **it << endl;
  }
#endif

  //----------------------------------------------------------------------------
  // RCCL
  //----------------------------------------------------------------------------
  for( ArteTableIterator<RCCL> rccl = ArteTable<RCCL>::begin();
       rccl != ArteTable<RCCL>::end(); ++rccl ) {

    ArteTable<EcalHit>::push_back( EcalHit(*rccl) );
  }


  //----------------------------------------------------------------------------
  // RTRA
  //----------------------------------------------------------------------------
  typedef R_trait<RecoTrack>::Cl_map Cl_map;
  Cl_map svdclone;
  Cl_map mainclone;
  Cl_map richclone;
  Cl_map ecalclone;
  Cl_map muonclone;

  for( ArteTableIterator<RTRA> rtit = ArteTable<RTRA>::begin();
       rtit != ArteTable<RTRA>::end(); ++rtit ) {
    const ArtePointer<RTRA> rtra = *rtit;

    // do not save reconstructed particles here - they are saved
    // in RecoVertex array
    const unsigned int D = rtra->fit%100;
    if( D == 2 || D == 3 ) continue;
    //    RecoTrack r( *rtra );
    //    Tracks->push_back( &r );
    const ArtePointer<RecoTrack> recotr = 
      ArteTable<RecoTrack>::push_back( RecoTrack(rtra,522) );

    // set relation to ECAL cluster
    if(distance(rtra->rccl.begin(),rtra->rccl.end()) > 0) {
      recotr->set_ecal( getPTR<EcalHit,RCCL>( *rtra->rccl.begin() ));
    }

    // fill SVD clone map
    const int csvd = rtra->csvd;
    if( csvd > 0) {
      fillCloneMap<RecoTrack>(svdclone, csvd, recotr);
    }

    // fill OTR clone map
    const int cmain = rtra->cmain;
    if( cmain > 0 ) {
      fillCloneMap<RecoTrack>(mainclone, cmain, recotr);
    }

    // fill RICH clone map
    const int crich = rtra->crich;
    if( crich > 0 ) {
      fillCloneMap<RecoTrack>(richclone, crich, recotr);
      //      richclone.insert(Cl_map::value_type(crich,recotr));
    }

    // fill ECAL clone map
    const int cecal = rtra->cecal;
    if( cecal > 0 ) {
      fillCloneMap<RecoTrack>(ecalclone, cecal, recotr);
      //      ecalclone.insert(Cl_map::value_type(cecal,recotr));
    }

    // fill Muon clone map
    const int cmuon = rtra->cmuon;
    if( cmuon > 0) {
      fillCloneMap<RecoTrack>(muonclone, cmuon, recotr);
      //      muonclone.insert(Cl_map::value_type(cmuon,recotr));
    }

    //    cout << "recotr->clones: " << recotr->clones << endl;
  } // for RTRA
  

  //----------------------------------------------------------------------------
  // Create Clone Relations
  //----------------------------------------------------------------------------
  //  cout << " svdclone" << endl;
  createCloneRelation<RecoTrack>(svdclone,  Rsegc::vxd );
  //  cout << " mainclone" << endl;
  createCloneRelation<RecoTrack>(mainclone, Rsegc::patt | Rsegc::magt );
  //  cout << " richclone" << endl;
  createCloneRelation<RecoTrack>(richclone, Rsegc::rich );
  //  cout << " ecalclone" << endl;
  createCloneRelation<RecoTrack>(ecalclone, Rsegc::ecal );
  //  cout << " muonclone" << endl;
  createCloneRelation<RecoTrack>(muonclone, Rsegc::muon );

  cout << " clones again: " << endl;
  for(ArteTableIterator<RecoTrack> it = ArteTable<RecoTrack>::begin(); it !=
	ArteTable<RecoTrack>::end(); ++it) {
    if(it->clones.size() > 0) {
      cout << *it << " rtrack->clones: " << it->clones << endl;
    }
  }
#ifdef ClueDEBUG
#endif

#ifdef XXX
  //----------------------------------------------------------------------------
  // RCCL-Clones
  //----------------------------------------------------------------------------
  for( ArteTable<RCCL>::iterator rcit = ArteTable<RCCL>::begin();
       rcit != ArteTable<RCCL>::end(); ++rcit ) {
    ArtePointer<RCCL> rccl = *rcit;

    if( distance(rccl->rtra.begin(), rccl->rtra.end()) > 1 ) {

      const ArtePointer<TrackClones> clones
	= ArteTable<TrackClones>::push_back( TrackClones( Rsegc::ecal ));

      for( ArteRelationIterator<RTRA> rtra = rccl->rtra.begin();
	   rtra != rccl->rtra.end(); ++rtra ) {

	const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>(*rtra);
	cout << " clone: " << *clones << endl;
	cout << " clone->rtrack: " << clones->rtrack << endl;
	cout << " rTrack: " << *rTrack << endl;
	cout << " rTrack->clones: " << rTrack->clones << endl;
	//	createRelation( clones, rTrack, &TrackClones::tracks );
	cout << " now create relation" << endl;
	createRelation( clones, rTrack );
	cout << " created Relation " << endl;
	cout << "xx rTrack->clones: " << rTrack->clones << endl;
	cout << "xx clone->rtrack: " << clones->rtrack << endl;
	for(ClueRelationIterator<RecoTrack> it = clones->rtrack.begin();
	    it != clones->rtrack.end(); ++it)
	  cout << **it << endl;
      }
    }
  }

  //----------------------------------------------------------------------------
  // RSEG-Clones
  //----------------------------------------------------------------------------
  for( ArteTable<RSEG>::iterator rsit = ArteTable<RSEG>::begin();
       rsit != ArteTable<RSEG>::end(); ++rsit ) {
    ArtePointer<RSEG> rseg = *rsit;

    if( distance(rseg->rtra.begin(), rseg->rtra.end()) > 1 ) {

      const ArtePointer<TrackClones> clones
	= ArteTable<TrackClones>::push_back(TrackClones( rseg->cmp ));
      
      for( ArteRelationIterator<RTRA> rtra = rseg->rtra.begin();
	   rtra != rseg->rtra.end(); ++rtra ) {

	const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( *rtra );
	//	createRelation( clones, rTrack, &TrackClones::tracks );
	createRelation( clones, rTrack );
      }
    }
  }
#endif

  //----------------------------------------------------------------------------
  // RSEG
  //----------------------------------------------------------------------------
  if( storeRSEG_Flag ) {
    for( ArteTableIterator<RSEG> rsit = ArteTable<RSEG>::begin();
	 rsit != ArteTable<RSEG>::end(); ++rsit ) {
      const ArtePointer<RSEG> rseg = *rsit;

#ifdef XXX
      cout << "------> RSEG " << rseg << ": ";
      print_cmp(rseg->cmp);
#endif

      if( rseg->cmp & Rsegc::vxd ) {
	ArteTable<VDS_Track>::push_back( VDS_Track(rseg) );
      } else if( (rseg->cmp & Rsegc::patt || rseg->cmp & Rsegc::trit ) 
		 && (rseg->cmp & Rsegc::rich || rseg->cmp & Rsegc::ecal)) {
	// save only OTR/RICH matches!!
	ArteTable<OTR_SpacePoint>::push_back( OTR_SpacePoint(rseg) );
      } else if( rseg->cmp & Rsegc::rich ) {
	ArteTable<RICH_Track>::push_back( RICH_Track(rseg) );
      } else if( rseg->cmp & Rsegc::muon ) {
	ArteTable<MuonTrack>::push_back( MuonTrack(rseg) );
      }
    }
  }


  //----------------------------------------------------------------------------
  // RVER
  //----------------------------------------------------------------------------
  for( ArteTable<RVER>::iterator rvit = ArteTable<RVER>::begin();
       rvit != ArteTable<RVER>::end(); ++rvit ) {
    const ArtePointer<RVER> rver = *rvit;

    const ArtePointer<RecoVertex> rVertex = 
      ArteTable<RecoVertex>::push_back( RecoVertex(rver) );

    double chi2 = 0;
    for( ArteRelationIterator<RREL> rrit = rver->rrel.begin();
	 rrit != rver->rrel.end(); ++rrit ) {
      const ArtePointer<RREL> rrel = *rrit;

      if( rrel->rtra != 0 ) {
	const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( rrel->rtra );
	//	createRelation( rVertex, rTrack, &RecoVertex::out );
	createRelation( rVertex, rTrack );
      }
      chi2 += rrel->chi;
    }
    rVertex->set_chi2(chi2);
  }


  //----------------------------------------------------------------------------
  // MVER
  //----------------------------------------------------------------------------
  for( ArteTableIterator<MVER> mver = ArteTable<MVER>::begin();
       mver != ArteTable<MVER>::end(); ++mver ) {
    ArteTable<MC_Vertex>::push_back( MC_Vertex(*mver) );
  }


  //----------------------------------------------------------------------------
  // MTRA
  //----------------------------------------------------------------------------
  for( ArteTableIterator<MTRA> mtit = ArteTable<MTRA>::begin();
       mtit != ArteTable<MTRA>::end(); ++mtit ) {
    const ArtePointer<MTRA> mtra = *mtit;
    const ArtePointer<MVER> mver = mtra->bver;

    const ArtePointer<MC_Vertex> mVertex = getPTR<MC_Vertex,MVER>( mver );

    MC_Track mTrack( mtra, mVertex );
    ArtePointer<MTRA> btra = mtra->btra;

    mTrack.set_mother( getPTR<MC_Track,MTRA>(btra) );
    
    const ArtePointer<MC_Track> mTrackAP = ArteTable<MC_Track>::push_back( mTrack );
    
    for( ArteRelationIterator<RTRA> rtra = mtra->rtra.begin();
	 rtra != mtra->rtra.end(); ++rtra ) {

      const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( *rtra );
      //      createRelation( rTrack, mTrackAP, &MC_Track::rTrack );
      createRelation( rTrack, mTrackAP );
    }
  }
  return 0;
} // RtEvent::init

