/*
*************************************************************************
**
** Description: GPACK - ARTE INTERFACE package "gart"
**
** Language:	C
**
** Called from: Arte.
** Externals:   gpack.
**
** Author:  	Sergey Moukhine
**          	HERA-B Collaboration
**          	JINR-Dubna
**          	Russia
**
** Created: 	 7 Jun 98
** Updated:	17 Nov 98
**			Output to Notebook with ev. mixing
** Updated:	18 Nov 98
**			Input from Notebook only (NoteIn)
** Updated:     05 Mar 01  (SE)
**                      Notebook creation (stream NOTE_OUT) bug fixed
**                      in cases when input is Public Event Directory
**
** Updated:     25 May 01  (SE)
**     Call of 'gp_settarget' was moved after gp_open() in functions
**     GartOpenDstMiniOut() and GratOpenOthers().  These changes are needed
**     by new version of DATAMGR (> DATAMGR-02-15). 
**
*************************************************************************
*/

#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>


#include "gpack/gpackproto.h"

#include <arte/ArteFlag.h>
#include "gart/gart.h"

#include <arte/Arte.h>
#include <arte/ArteIOlun.h>
#include <auto/EVHD.h>
#include <auto/MEMO.h>
#include <auto/MEMA.h>
#include <auto/MEIO.h>

#include <arte/ZIOCOM.h>

#include <arte/mecCount.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
*************************************************************************
**		Static variables
*************************************************************************
*/

static GartStreamData gartStream[GartStreamMax];

/* Indexing GartStreamData: ARTE LUN ranges 1..99, lun2stream[0] unused */
static GartLun2Stream lun2stream[ARTE_NLUN];

static int  initFlag = 0;
static char errMess[500];
static int  tableNameIntBufer[namelen+1];
static char *tableName = (char *)tableNameIntBufer;
static char tableDesc[TABLE_DESC_LEN+1];
static int  sigWaitRest;
static char selectionCriteriaText[SELECT_CRIT_LEN+1];
static int  selectionLun;
static char *selectionNullText = " ";

/*
*************************************************************************
**		Internal function prototypes
*************************************************************************
*/

static GartStreamKind GartGetStreamKind(int *lun, GartStreamData **sPtr);
static char *GartStreamName(GartStreamKind id);

static void AlarmHandler(int sig);
static void GartInit(void);
static char *ArteTableName(int tab);
static char *ArteTableDesc(int tab, char *tName);

static int  GartShmKeyByName(char *name);
static int  GartOpenDstMiniOut(GartStreamKind id, int lrec);
static int  GartOpenEvIn(GartStreamKind id, int lrec);
static int  GartOpenEvdir(GartStreamKind id, int lrec);
static int  GartOpenNoteOut(int lrec);
static int  GartOpenNoteIn(GartStreamKind id, int lrec);
static int  GartOpenOthers(GartStreamData *sPtr, int lrec);

static int  ArteDatimeToUnixTime(int aDate, int aTime);
static void ArteDatimeFromUnixTime(int datime, int *aDate, int *aTime);
static int  GetTabInfo(GartStreamData *sPtr, int *name);
static int  ReadTab(GartStreamData *sPtr, int tab);
static int  GartDstMiniOut(GartStreamData *sPtr, int tab, int inew);
static int  GartNoteOut();
static char *GetSelectionCriteria();
static void EvdInfoFromHeader(EvdInfo *info, EventHdr *header);
static void HeaderFromEvhd(EventHdr *header);

/*
*************************************************************************
**		External function prototypes
*************************************************************************
*/

extern void mmbook_(int *adr, int *var);
extern void mmdrop_(int *adr);
extern char *ku_gets(void);
extern int   ku_npar(void);
extern void chttoh_ (char *T, int *H, int *N);
extern void chhtot_ (int *H, char *T, int *N);

/*
*************************************************************************
**		Function implementation
*************************************************************************
*/

/***********************************************************************/

void gartopen_(int *lun, char *name, int *shm_flag, int *in_out_flag,
                         int *lrec, int *ios,
                         int  len)
/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	name	 - filename to be opened
**	shm_flag - 
**		   1 for Shared Memory (open / create+open)
**		   0 for other files
**	in_out_flag - 
**		   1 open for output
**		   0 open for input
**	lrec	 - record size (for Shared Memory: size of SHM)
**	ios	 - (output) see GartErrorCode and StreamErrorCode (0 - OK)
**	len	 - lenght of name (needed for FORTRAN strings)
**
** FUNCTION DEFINITION
**	Open stream on lun. 
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
char *firstBlank;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *ios = GART_WRONG_PAR;
   return;
   }

/* Return if already opened */
if (sPtr->isOpen)
   {
   if (sPtr->isOutput != *in_out_flag)
      {
      static int errPtr;
   
      sprintf(errMess, "GART: lun %d is already opened for ",*lun);
      if (sPtr->isOutput)
         strcat(errMess, "output");
      else
         strcat(errMess, "input");
      mecErrorCount(&errPtr, errMess);

      *ios = GART_WRONG_PAR;
      return;
      }
   *ios = GART_OK;
   return;
   }

sPtr->arteLun = *lun;
sPtr->isOutput = *in_out_flag;

/* Copy File Name */
memset(sPtr->fileName,0,sizeof(sPtr->fileName));

len = len > sizeof(sPtr->fileName) ? sizeof(sPtr->fileName) : len;
if (len > 0)
   strncpy(sPtr->fileName,name,len);

firstBlank = strchr(sPtr->fileName,' ');
if (firstBlank)
   *firstBlank = '\0';

/* Shared Memory or File*/
sPtr->isShm = *shm_flag;
      
switch (id)
   {
   case GartStreamOthers:
 /*=======================*/
      {
      *ios = GartOpenOthers(sPtr, *lrec);
      return;
      }
      
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:
 /*=======================*/
      {
      *ios = GartOpenDstMiniOut(id, *lrec);
      return;
      }
      
   case GartStreamNoteOut:
 /*=======================*/
      {
      *ios = GartOpenNoteOut(*lrec);
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:
 /*=====================*/
      {
      *ios = GartOpenEvIn(id, *lrec);
      return;
      }
            
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamNoteIn:
 /*======================*/
      {
      *ios = GartOpenNoteIn(id, *lrec);
      return;
      }
      
   case GartStreamEvdirIn:	/* no break */
 /*==========================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:
 /*===========================*/

      {
      *ios = GartOpenEvdir(id, *lrec);
      return;
      }
   
   default:
      *ios = GART_WRONG_PAR;
      return;         
   }

}

/***********************************************************************/

void gartclos_(int *lun, int *ios)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	ios	 - (output) see GartErrorCode and StreamErrorCode (0 - OK)
**
** FUNCTION DEFINITION
**	Close stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
int		rc;
int		gLun;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *ios = GART_WRONG_PAR;
   return;
   }

/* Return if already closed */
if (!sPtr->isOpen)
   {
   *ios = GART_OK;
   return;
   }

gLun = sPtr->gpackLun;


if (sPtr->isOpen == GartOpenEndOfData)
   rc = 0;
else
   {
#ifdef DEBUG
 printf("gp_close %d\n",gLun);
#endif
   rc = gp_close(gLun);
   }
   
if (rc < 0)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: GPACK close error. rc = %s : %s",
      gp_rcmsg(rc), gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   *ios = rc;
   }

gp_delstream(gLun);
sPtr->isOpen = GartNotOpen;
sPtr->arteLun = 0;

return;
}

/***********************************************************************/

void gartend_(void)

/************************************************************************
**
** PARAMETERS
**	None
**
** FUNCTION DEFINITION
**	Close all streams if necessary.
*************************************************************************
*/

{
int ios;
int i;
GartStreamData  *sPtr;

GartInit();

for (i=0; i<GartStreamMax; i++)
   {
   sPtr = &gartStream[i];
   if (i != GartStreamOthers)
      {
      if (sPtr->isOpen)
         gartclos_(&sPtr->arteLun, &ios);
      }
   else
      {
      while (sPtr)
         {
         if (sPtr->isOpen)
            gartclos_(&sPtr->arteLun, &ios);
         sPtr = sPtr->next;
         }
      }
   }
}

/***********************************************************************/

void gartendev_(int *lun)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**
** FUNCTION DEFINITION
**	End of event for stream on lun.
*************************************************************************
*/

{
int rc;
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);

if (id == GartStreamNoteOut)
   {
   GartNoteOut();
   return;
   }
   
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   return;
   }

/* Flush event */
if (sPtr->isOpen && id != GartStreamNoteOut)
   {
   rc = gp_flush(sPtr->gpackLun);
#ifdef DEBUG
 printf("gp_flush %d rc=%d %s\n",sPtr->gpackLun,rc,gp_rcmsg(rc));
#endif
   if (rc < 0)
      {
      static int errPtr;
         
      sprintf(errMess, "GART: Cannot flush lun %d. rc = %d : %s\n",
         *lun, rc, gp_errmsg(sPtr->gpackLun));
      mecErrorCount(&errPtr, errMess);
      }
   }

/* Write to Notebook if opened (not SHM) */

if (gartStream[GartStreamNoteOut].isOpen && !sPtr->isShm)
   {
   if (id == GartStreamDstOut)
      sPtr->evdInfo.dstPos	= gp_getevtpos(sPtr->gpackLun);
   if (id == GartStreamMiniOut)
   sPtr->evdInfo.miniPos	= gp_getevtpos(sPtr->gpackLun);
         
   gartStream[GartStreamNoteOut].isOpen = GartOpenWritten;
   rc = gp_putevdinfo(gartStream[GartStreamNoteOut].gpackLun,
                      &sPtr->evdInfo);
#ifdef DEBUG
 printf("EVD put: ev=%5d dst=%7d mini=%7d %s\n",
        sPtr->evdInfo.prop.event,sPtr->evdInfo.dstPos,sPtr->evdInfo.miniPos,
        GartStreamName(id));
#endif

   if (rc < 0)
      {
      static int errPtr;
         
      sprintf(errMess, "GART: Cannot put Evd. info. rc = %d: %s\n",
         rc, gp_errmsg(sPtr->gpackLun));
      mecErrorCount(&errPtr, errMess);
      }
   }

return;
}

/***********************************************************************/

void gartsetgpack_(int *lun)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**
** FUNCTION DEFINITION
**	Set stream on lun to be a gpack stream.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   return;
   }

sPtr->arteLun = *lun;
return;

}

/***********************************************************************/

void gartcheck_(int *lun, int *is_gpack)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	is_gpack - (output) =1 for gpack, =0 otherwize
**
** FUNCTION DEFINITION
**	Checks whether stream on lun is a gpack stream.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   *is_gpack = 0;
   return;
   }

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:
 /*=====================*/
      if (sPtr->arteLun)
         *is_gpack = 1;
      else
         *is_gpack = 0;
         
      break;
   
   case GartStreamNoteOut:	/* no break */
 /*=======================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:	/* no break */
 /*==========================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   default:
      *is_gpack = 1;
   }

return;         
}

/***********************************************************************/

void gartothers_(int *lun, int *may_not_gpack)

/************************************************************************
**
** PARAMETERS
**	lun	  - stream number from Arte (1..99)
**	may_not_gp - (output) =1 for not gpack, =0 otherwize
**
** FUNCTION DEFINITION
**	Checks whether stream on lun may be a non-gpack stream.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   *may_not_gpack = 1;
   return;
   }

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:
 /*=====================*/
      *may_not_gpack = 1;
      break;
   
   case GartStreamNoteOut:	/* no break */
 /*=======================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:	/* no break */
 /*==========================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   default:
      *may_not_gpack = 0;
   }

return;         
}

/***********************************************************************/

void gartrewi_(int *lun, int *ios)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	ios	 - (output) see GartErrorCode and StreamErrorCode (0 - OK)
**
** FUNCTION DEFINITION
**	Rewind stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *ios = GART_WRONG_PAR;
   return;
   }

/* OK if not opened */
if (!sPtr->isOpen || sPtr->isOpen == GartOpenEndOfData)
   {
   *ios = GART_OK;
   return;
   }


/* rewind if opened */
#ifndef NOT_IMPLEMENTED
*ios = gp_rewind(sPtr->gpackLun);
if (*ios == GP_NOT_IMPLEMENTED)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Rewind is not implemented for lun=%d\n",*lun);
   mecErrorCount(&errPtr, errMess);
      
   return;
   }
return;
#else
   /*TEMP		gp_rewind function should be written */
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Rewind is not yet implemented in this version\n");
   mecErrorCount(&errPtr, errMess);
      
   *ios = GP_NOT_IMPLEMENTED;
   return;
   }
return;

#endif
}

/***********************************************************************/

void gartlimi_(int *lun, float *limit)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	limit	 - file limit in Mbytes
**
** FUNCTION DEFINITION
**	Set file limit for a stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   return;
   }

/* Store file limit in bytes */
sPtr->limit = (int)((*limit)*1024.*1024.);

return;
}

/***********************************************************************/

void gartwtab_(int *lun, int *tab, int *inew, int *questArr)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	tab	 - table index (data start from iq(jlink(tab)+1) )
**	inew	 - start of event flag: =1 for new event, =0 otherwize
**	questArr - (output) return code:
**	questArr(1)	 0 normal completion
**			+1 "pseudo end of tape"
**			-1 first attempt to write after end_of_data
**			-2 error return
**
** FUNCTION DEFINITION
**	Write one table to stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *questArr = -2;
   return;
   }

/* Error if not opened */
if (!sPtr->isOpen)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Attempt to write to a not opened stream\n");
   mecErrorCount(&errPtr, errMess);
      
   *questArr = -2;
   return;
   }


switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:
 /*=======================*/
      {
      *questArr = GartDstMiniOut(sPtr,*tab,*inew);
      return;
      }
      
   case GartStreamOthers:
 /*=======================*/
      {
      
      if (!sPtr->isOutput)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Attempt to write to an input stream %d",
            *lun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = -2;
         return;
         }

      *questArr = GartDstMiniOut(sPtr,*tab,*inew);
      return;
      }
      
   case GartStreamNoteOut:
 /*=======================*/
      {
      *questArr = 0;
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:	/* no break */
 /*=====================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:	/* no break */
 /*==========================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:
 /*===========================*/

      {
      static int errPtr;
      
      sprintf(errMess, "GART: Attempt to write to an input stream!");
      mecErrorCount(&errPtr, errMess);
      
      *questArr = -2;
      return;
      }
            
   default:
      *questArr = -2;
      return;         
   }

}

/***********************************************************************/

void gartrnext_(int *lun, int *questArr)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	questArr - (output) return code:
**	questArr(1)	 0  normal completion
**			+5  End of Data
**			<0  error return (StreamErrorCode)
**
** CAUTIONS
**	Take care about reading from SHM: 
**	we will wait on a semaphore infinitely after the last event
**	if nobody interrupts us. So alarm call is used to set
**	waiting time.
**
** FUNCTION DEFINITION
**	Read next event header from stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
int		rc;
EventHdr	header;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *questArr = GP_FAILURE;
   return;
   }

if (!sPtr->isOpen)	/* Error if not opened */
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Attempt to read from a not opened stream!\n");
   mecErrorCount(&errPtr, errMess);
      
   *questArr = 5;
   return;
   }

if (sPtr->isOpen == GartOpenEndOfData)
   {
   *questArr = 5;
   return;
   }

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamNoteOut:
 /*=======================*/
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Attempt to read from an output stream!");
      mecErrorCount(&errPtr, errMess);
      
      *questArr = 5;
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:	/* no break */
 /*=====================*/
   case GartStreamOthers:
 /*=====================*/
      {
      if (id == GartStreamOthers && sPtr->isOutput)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Attempt to read from an output stream %d",
            *lun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = 5;
         return;
         }
      
      rc = 0;
      /* 
      ** Take care about reading from SHM: 
      ** we will wait on a semaphore infinitely after the last event
      ** if nobody interrupts us.
      */
      if (sPtr->isShm)
         {
         struct sigaction oldAct;
         struct sigaction newAct;
         
         /* Get default handler */
         sigaction(SIGALRM, NULL, &oldAct);
         
         /* Set our handler */
         newAct = oldAct;
         newAct.sa_handler = AlarmHandler;
         
         /* Activate our handler */
         sigWaitRest = ALRM_WAIT_CNT;
         sigaction(SIGALRM, &newAct, NULL);
         
         while (sigWaitRest > 0)
            {
            alarm(ALRM_WAIT_TIME);
            rc = gp_getevt(sPtr->gpackLun,&header);
#ifdef DEBUG
 printf("gp_getevt %d rc=%d %s\n",sPtr->gpackLun,rc,gp_rcmsg(rc));
#endif
            alarm(0);
            
            if (rc != GP_INTERRUPTED)
               break;
            rc = GP_END_OF_DATA;
            }
         /* Restore default handler */
         sigaction(SIGALRM, &oldAct, NULL);
         }
      else
         {
         rc = gp_getevt(sPtr->gpackLun,&header);
#ifdef DEBUG
 printf("gp_getevt %d rc=%d %s\n",sPtr->gpackLun,rc,gp_rcmsg(rc));
#endif
         }

      sPtr->eventType = header.type;
      if (rc >= 0)
         {
         EvdInfoFromHeader(&sPtr->evdInfo, &header);
         if (header.type == Dst)
            {
            sPtr->evdInfo.dstPos	= gp_getevtpos(sPtr->gpackLun);
            sPtr->evdInfo.miniPos	= -1;
            }
         if (header.type == Mini)
            {
            sPtr->evdInfo.dstPos	= -1;
            sPtr->evdInfo.miniPos	= gp_getevtpos(sPtr->gpackLun);
            }
         sPtr->evdInfo.fileNumber	= 0;
#ifdef DEBUG
	 printf("%s:%d: lun %d, fileName %s\n",__FILE__,__LINE__,
	sPtr->gpackLun,sPtr->fileName);
#endif
         strcpy(sPtr->evdInfo.dataname,sPtr->fileName);
         }
      
      *questArr = 0;

      if (rc < 0)
         *questArr = rc;

      if (rc == GP_END_OF_DATA)
         *questArr = 5;

      if (rc == GP_UNEXPECTED_EOF)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: unexpected EOF on lun = %d",sPtr->arteLun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = 5;
         }

      return;
      }


   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:
 /*==========================*/
      {
      if (MEIO(sPtr->arteLun).nskip * MEIO(sPtr->arteLun).fskip > 0)
	 {
	 gp_skipevt(sPtr->gpackLun, MEIO(sPtr->arteLun).nskip - 1);
	 MEIO(sPtr->arteLun).nskip = 0;
	 MEIO(sPtr->arteLun).fskip = 0;
	 }

      rc = gp_getevt(sPtr->gpackLun,&header);
#ifdef DEBUG
 printf("gp_getevt %d rc=%d %s\n",sPtr->gpackLun,rc,gp_rcmsg(rc));
#endif

      if (rc >= 0)
         {
         int rcInf;
         
         rcInf = gp_getevdinfo(sPtr->gpackLun, &sPtr->evdInfo);
         if (rcInf < 0)
            {
            static int errPtr;
            
            sprintf(errMess, "GART: cannot get Evd. info. lun = %d: %s",
               sPtr->arteLun,gp_errmsg(sPtr->gpackLun));
            mecErrorCount(&errPtr, errMess);
            rc = rcInf;
            }
         }

      *questArr = 0;

      if (rc < 0)
         *questArr = rc;

      if (rc == GP_END_OF_DATA)
         *questArr = 5;

      if (rc == GP_UNEXPECTED_EOF)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: unexpected EOF on lun = %d",sPtr->arteLun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = 5;
         }

      return;
      }

   default:
      *questArr = 5;
      return;         
   }

}

/***********************************************************************/

void gartrtab_(int *lun, int *name, int *questArr)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	name	 - TABLE_NAME_LEN characters of the next table name
**	questArr - (output) return code:
**	questArr(1)	 0  normal completion
**			+4  End of File
**			+5  End of Data
**			100 end of event
**			<0  error return (StreamErrorCode)
**	questArr(11)	always set to zero (no start new event)
**
** CAUTIONS
**	Take care about reading from SHM: 
**	we will wait on a semaphore infinitely after the last event
**	if nobody interrupts us. So alarm call is used to set
**	waiting time.
**
** FUNCTION DEFINITION
**	Read the description of current table from stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
int		rc;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *questArr = GP_FAILURE;
   return;
   }

/* questArr(11)	always set to zero (no start new event) */
questArr[10] = 0;

if (!sPtr->isOpen)	/* Error if not opened */
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Attempt to read from a not opened stream!\n");
   mecErrorCount(&errPtr, errMess);
      
   *questArr = 5;
   return;
   }

if (sPtr->isOpen == GartOpenEndOfData)
   {
   *questArr = 5;
   return;
   }

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamNoteOut:
 /*=======================*/
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Attempt to read from an output stream!");
      mecErrorCount(&errPtr, errMess);
      
      *questArr = 5;
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:	/* no break */
 /*=====================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamOthers:
 /*=====================*/
      {
      if (id == GartStreamOthers && sPtr->isOutput)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Attempt to read from an output stream %d",
            *lun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = 5;
         return;
         }
      
      rc = 0;
      /* 
      ** Take care about reading from SHM: 
      ** we will wait on a semaphore infinitely after the last event
      ** if nobody interrupts us.
      */
      if (sPtr->isShm)
         {
         struct sigaction oldAct;
         struct sigaction newAct;
         
         /* Get default handler */
         sigaction(SIGALRM, NULL, &oldAct);
         
         /* Set our handler */
         newAct = oldAct;
         newAct.sa_handler = AlarmHandler;
         
         /* Activate our handler */
         sigWaitRest = ALRM_WAIT_CNT;
         sigaction(SIGALRM, &newAct, NULL);
         
         while (sigWaitRest > 0)
            {
            alarm(ALRM_WAIT_TIME);
            rc = GetTabInfo(sPtr,name);
            alarm(0);
            
            if (rc != GP_INTERRUPTED)
               break;
            }
         /* Restore default handler */
         sigaction(SIGALRM, &oldAct, NULL);
         }
      else
         rc = GetTabInfo(sPtr,name);

      if (rc >= 0)
         *questArr = 0;

      if (rc < 0)
         *questArr = 5;

      if (rc == GP_END_OF_DATA)
         *questArr = 100;

      if (rc == GP_INTERRUPTED)
         *questArr = 5;

      return;
      }

   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:
 /*==========================*/
      {
      int tab;
      
      tab = (&artelink_.jevhd - &artelink_.jlink[0]) + 1;
      memcpy(name,MEMO(tab).name,TABLE_NAME_LEN);
      *questArr = 0;
      if (sPtr->ncols > 0)
         *questArr = 100;
      sPtr->ncols = 0;
      return;
      }
      
   default:
      *questArr = 5;
      return;         
   }

}

/***********************************************************************/

void gartrdat_(int *lun, int *tab, int *ncols, int *questArr)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	tab	 - table index: mmbook(jlink(tab),...)
**	ncols	 - n of output_columns in the table read
**	questArr - (output) return code:
**	questArr(1)	 0  normal completion
**			<0  error return (StreamErrorCode)
**			+5  End Of File
**			100 end of event ==> skip event
**
** FUNCTION DEFINITION
**	Read data of current table from stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
int		rc;

GartInit();

*questArr = 0;

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *questArr = GP_FAILURE;
   return;
   }

/* Error if not opened */
if (!sPtr->isOpen)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Attempt to read from a not opened stream!\n");
   mecErrorCount(&errPtr, errMess);
      
   *questArr = GP_FAILURE;
   return;
   }

if (sPtr->isOpen == GartOpenEndOfData)
   {
   *questArr = 5;
   return;
   }

*ncols = sPtr->ncols;

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamNoteOut:
 /*=======================*/
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Attempt to read from an output stream!");
      mecErrorCount(&errPtr, errMess);
      
      *questArr = GP_FAILURE;
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:	/* no break */
 /*=====================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamOthers:
 /*=====================*/
      {
      if (id == GartStreamOthers && sPtr->isOutput)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Attempt to read from an output stream %d",
            *lun);
         mecErrorCount(&errPtr, errMess);
      
         *questArr = 5;
         return;
         }
      
      rc = 0;
      /* 
      ** Take care about reading from SHM: 
      ** we will wait on a semaphore infinitely after the last event
      ** if nobody interrupts us.
      */
      if (sPtr->isShm)
         {
         struct sigaction oldAct;
         struct sigaction newAct;
         
         /* Get default handler */
         sigaction(SIGALRM, NULL, &oldAct);
         
         /* Set our handler */
         newAct = oldAct;
         newAct.sa_handler = AlarmHandler;
         
         /* Activate our handler */
         sigWaitRest = ALRM_WAIT_CNT;
         sigaction(SIGALRM, &newAct, NULL);
         
         while (sigWaitRest > 0)
            {
            alarm(ALRM_WAIT_TIME);
            rc = ReadTab(sPtr,*tab);
            alarm(0);
            
            if (rc != GP_INTERRUPTED)
               break;
            rc = GP_END_OF_DATA;
            }
         /* Restore default handler */
         sigaction(SIGALRM, &oldAct, NULL);
         }
      else
         rc = ReadTab(sPtr,*tab);

      if (rc >= 0)
         *questArr = 0;

      if (rc < 0)
         *questArr = rc;

      if (rc == GP_END_OF_DATA)
         {
         *questArr = 5;
         return;
         }

      /*
      ** Fill "flag" field of EVHD and check for selection criteria
      */
      {
      int      tabEvhd;
      EventHdr header;
      EvdInfo  info;

      tabEvhd = (&artelink_.jevhd - &artelink_.jlink[0]) + 1;
      if (tabEvhd == *tab)
         {
         EVHD(1).flag = EVHD(1).flagdst;
         
         /* Check selection criteria for EventIn */
         if (id == GartStreamEventIn || id == GartStreamEvIn1)
            {
            HeaderFromEvhd(&header);
            EvdInfoFromHeader(&info, &header);

            /* No selects for Mixing */
            if (ziocom_.fiomix[4] != 0. &&
               strcmp(selectionNullText,GetSelectionCriteria()) != 0)
               {
               static int errPtr;
      
               sprintf(errMess, "GART: Selection Criteria cannot be set for ev. mixing\n");
               mecErrorCount(&errPtr, errMess);
               rc = 1;
               }
            else
               rc = gp_filter(selectionLun,&info.prop);

            if (rc < 0)
               *questArr = rc;
            if (rc == 0)
               {
               mmdrop_(&artelink_.jevhd);
               *questArr = 100; /* skip event */
               }
            }
         }
      }
      
      return;
      }

   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:
 /*==========================*/
      {
      int tabEvhd;
      int one = 1;
      
      /* must be EVHD */
      tabEvhd = (&artelink_.jevhd - &artelink_.jlink[0]) + 1;
      if (tabEvhd != *tab)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: only EVHD should be read for NoteIn/EvdirIn");
         mecErrorCount(&errPtr, errMess);
      
         *questArr = GP_FAILURE;
         return;
         }


      if (artelink_.jevhd == 0)
         mmbook_(&artelink_.jevhd, &one);
         
      EVHD(1).flag = sPtr->evdInfo.prop.classmask;
      EVHD(1).flagdst = EVHD(1).flag;
      EVHD(1).evt = sPtr->evdInfo.prop.event;
      EVHD(1).run = sPtr->evdInfo.prop.run;
      EVHD(1).exp = sPtr->evdInfo.prop.exp;
      EVHD(1).vfld = 0.;
      ArteDatimeFromUnixTime(sPtr->evdInfo.prop.datime,
                             &(EVHD(1).date), &(EVHD(1).time));
      
      sPtr->ncols = MEMO(tabEvhd).ocol;
      *ncols = sPtr->ncols;
      
      *questArr = 0;
      return;
      }
      
   default:
      *questArr = GP_FAILURE;
      return;         
   }

}

/***********************************************************************/

void gartskiptab_(int *lun, int *ios)

/************************************************************************
**
** PARAMETERS
**	lun	 - stream number from Arte (1..99)
**	ios	 - (output) return code:
**			 0  normal completion
**			<0  error return (StreamErrorCode)
**			+5  End of File / End of Data
**			100 end of event
**
** FUNCTION DEFINITION
**	Skip current table from stream on lun.
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
int		rc;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
if (id == GartStreamWrong)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: lun %d is not suitable for GPACK",*lun);
   mecErrorCount(&errPtr, errMess);

   *ios = GP_FAILURE;
   return;
   }

if (!sPtr->isOpen)	/* Error if not opened */
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Attempt to read from a not opened stream!\n");
   mecErrorCount(&errPtr, errMess);
      
   *ios = GP_FAILURE;
   return;
   }

if (sPtr->isOpen == GartOpenEndOfData)
   {
   *ios = GP_END_OF_DATA;
   return;
   }

switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamNoteOut:
 /*=======================*/
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Attempt to read from an output stream!");
      mecErrorCount(&errPtr, errMess);
      
      *ios = GP_FAILURE;
      return;
      }
      
   case GartStreamEventIn:	/* no break */
 /*=======================*/
   case GartStreamEvIn1:	/* no break */
 /*=====================*/
   case GartStreamEvIn2:	/* no break */
 /*=====================*/
   case GartStreamEvIn3:	/* no break */
 /*=====================*/
   case GartStreamEvIn4:	/* no break */
 /*=====================*/
   case GartStreamDstInEvdir:	/* no break */
 /*==========================*/
   case GartStreamMiniInEvdir:	/* no break */
 /*===========================*/
   case GartStreamDstInNote:	/* no break */
 /*=========================*/
   case GartStreamMiniInNote:	/* no break */
 /*==========================*/
   case GartStreamNoteIn:	/* no break */
 /*======================*/
   case GartStreamEvdirIn:
 /*==========================*/
   case GartStreamOthers:
 /*=====================*/
      {
      if (id == GartStreamOthers && sPtr->isOutput)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Attempt to read from an output stream %d",
            *lun);
         mecErrorCount(&errPtr, errMess);
      
         *ios = GP_FAILURE;
         return;
         }
      
#ifdef DEBUG
 printf("gp_skiptab %d\n",sPtr->gpackLun);
#endif
      rc = gp_skiptab(sPtr->gpackLun);
      if (rc < 0)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: cannot skip table, rc = %s",
            gp_rcmsg(rc));
         mecErrorCount(&errPtr, errMess);
      
         *ios = rc;
         return;
         }
      
      *ios = 0;
      return;
      }

   default:
      *ios = 0;
      return;         
   }

}

/***********************************************************************/

static GartStreamKind GartGetStreamKind(int *lunInp, GartStreamData **sPtr)

/************************************************************************
**
** PARAMETERS
**	lunInp	 - stream number from Arte +-(1..99)
**	sPtr	 - (output) address of internal structure
**
** RETURNS
**	internal stream identification
**
** FUNCTION DEFINITION
**	Converts stream number lun to internal stream identification
**	of type GartStreamKind. Returns address of internal GartStreamData
**	structure for this lun.
*************************************************************************
*/

{
int lun;
GartStreamKind id;

lun = *lunInp;
if (lun < 0)
   lun = -lun;

/* This should never happen */
if ( !lun || lun >= ARTE_NLUN)
   {
   static int errPtr;

   sprintf(errMess, "GART: LUN %d requested out of ARTE 1..%d range",
           *lunInp, ARTE_NLUN-1);
   mecWarningCount(&errPtr, errMess);
   }
else if ( (*sPtr = lun2stream[lun].sData) )
   return id = lun2stream[lun].sKind;

id = GartStreamOthers;


/* Seek the last GartStreamOthers, return already booked non-ARTE luns */
*sPtr = &gartStream[GartStreamOthers];
do {
   if ( lun == (*sPtr)->arteLun )
      return id;
   }
while ( (*sPtr)->next && (*sPtr = (*sPtr)->next) );

if ( !((*sPtr)->next = calloc(1, sizeof(GartStreamData))) )
   {
   static int errPtr;

   sprintf(errMess, "GART: Memory allocation problems for Lun %d", lun);
   mecErrorCount(&errPtr, errMess);
   *sPtr = NULL;
   id = GartStreamWrong;
   }
else
   {
   *sPtr = (*sPtr)->next;
   (*sPtr)->arteLun = lun;
   if (lun && lun < ARTE_NLUN)
      {
      lun2stream[lun].sKind = id;
      lun2stream[lun].sData = *sPtr;
      }
   }

return(id);
}

/***********************************************************************/

static char *GartStreamName(GartStreamKind id)

/************************************************************************
**
** PARAMETERS
**	id	 - stream kind
**
** RETURNS
**	GartStreamKind character name
**
** FUNCTION DEFINITION
**	Returns GartStreamKind character name (for debugging).
*************************************************************************
*/

{
static char *name;

switch (id)
   {
   case GartStreamWrong:
      name = "GartStreamWrong       ";
      break;
      
   case GartStreamDstOut:
      name = "GartStreamDstOut      ";
      break;
      
   case GartStreamMiniOut:
      name = "GartStreamMiniOut     ";
      break;
      
   case GartStreamNoteOut:
      name = "GartStreamNoteOut     ";
      break;
      
   case GartStreamEventIn:
      name = "GartStreamEventIn     ";
      break;
      
   case GartStreamEvIn1:
      name = "GartStreamEvIn1       ";
      break;
      
   case GartStreamEvIn2:
      name = "GartStreamEvIn2       ";
      break;
      
   case GartStreamEvIn3:
      name = "GartStreamEvIn3       ";
      break;
      
   case GartStreamEvIn4:
      name = "GartStreamEvIn4       ";
      break;
      
   case GartStreamDstInNote:
      name = "GartStreamDstInNote   ";
      break;
      
   case GartStreamMiniInNote:
      name = "GartStreamMiniInNote  ";
      break;
      
   case GartStreamNoteIn:
      name = "GartStreamNoteIn      ";
      break;
      
   case GartStreamEvdirIn:
      name = "GartStreamEvdirIn     ";
      break;
      
   case GartStreamDstInEvdir:
      name = "GartStreamDstInEvdir  ";
      break;
      
   case GartStreamMiniInEvdir:
      name = "GartStreamMiniInEvdir ";
      break;
      
   case GartStreamOthers:
      name = "GartStreamOthers      ";
      break;
      
   case GartStreamMax:
      name = "GartStreamMax         ";
      break;
      
   default:
      name = "GartStream?????       ";
      break;
   }

return(name);
}

/***********************************************************************/

static void GartInit(void)

/************************************************************************
**
** PARAMETERS
**	None
**
** FUNCTION DEFINITION
**	Gart initialization.
*************************************************************************
*/

{
int  rc;

if (initFlag)
   {
   if (0) GartStreamName(0);	/* to avoid compiler warnings */
   return;
   }
   
initFlag = 1;

memset(gartStream,0,sizeof(gartStream));


/*
** Now initialize lun2stream index table, fill in known streams.
** The default ARTE I/O unit numbers are set in IOINIT. GART
** is an I/O package, moreover IOINIT is called from ATINIT
** even before definition of most of ARTE tables, initialization
** of KUIP etc. Note, that we neither deallocate GartStreamData,
** nor is ARTE able to change LUN falvour, so there is no need
** for lun2stream[] cleanup at the end of session or in gartclos_().
*/
memset(lun2stream, 0, sizeof(lun2stream));

{
int i;
int tmp[][2] =
   {
	{arteiolun_.ldsto,	GartStreamDstOut	},
	{arteiolun_.lminio,	GartStreamMiniOut	},
	{arteiolun_.lnoteo, 	GartStreamNoteOut	},
	{arteiolun_.levi, 	GartStreamEventIn	},
	{arteiolun_.levi1, 	GartStreamEvIn1		},
	{arteiolun_.levi2, 	GartStreamEvIn2		},
	{arteiolun_.levi3, 	GartStreamEvIn3		},
	{arteiolun_.levi4, 	GartStreamEvIn4		},
	{arteiolun_.ldsti, 	GartStreamDstInNote	},
	{arteiolun_.lminii, 	GartStreamMiniInNote	},
	{arteiolun_.lnotei, 	GartStreamNoteIn	},
	{arteiolun_.lediri, 	GartStreamEvdirIn	},
	{arteiolun_.ldstie, 	GartStreamDstInEvdir	},
	{arteiolun_.lminiie, 	GartStreamMiniInEvdir	},
	{0,			0			}
   };

for (i = 0; i < ARTE_NLUN; i++)
   lun2stream[i].sKind = GartStreamWrong;

for (i = 0; tmp[i][0] || tmp[i][1]; i++)
   {
   int l = tmp[i][0];
   int s = tmp[i][1];

   if (l <= 0 || l >= ARTE_NLUN)
      {
      static int errPtr;

      sprintf(errMess, "GART cannot be properly initialized^@");
      mecErrorCount(&errPtr, errMess);
      }
   lun2stream[l].sKind = s;
   lun2stream[l].sData = &gartStream[s];
   }
}


strcpy(selectionCriteriaText,selectionNullText);

/* Create Gpack unit for event selection */

selectionLun = gp_setstream(PrivateStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",selectionLun);
#endif

if (selectionLun < 0)
   {
   static int errPtr;
   
   rc = selectionLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return;
   }

/*
 *  Null Selection Criteria
 */

#ifdef DEBUG
 printf("gp_select %d\n",selectionLun);
#endif
rc = gp_select(selectionLun,selectionNullText);
   
if (rc  < 0)
   {
   static int errPtr;

   printf("%s\n",gp_errmsg(selectionLun));
   sprintf(errMess, "GART: Select failed. rc=%s\n",
      gp_rcmsg(rc));
   mecErrorCount(&errPtr, errMess);
   return;
   }
}

/***********************************************************************/

static int GartShmKeyByName(char *name)

/************************************************************************
**
** PARAMETERS
**	name	 - the name of file associated with Shared Memory
**
** FUNCTION DEFINITION
**	Returns Shared Memory key associated (by function ftok)
**	with a given filename
*************************************************************************
*/

{
struct stat st;
int rc;
int fDes;

rc = stat(name, &st);
if( rc < 0 )
   {
   fDes = open(name, O_WRONLY|O_CREAT, 0644);
   if(fDes >= 0)
      close(fDes);
   }

return ((int)ftok(name,'G'));
}

/***********************************************************************/

static int GartOpenDstMiniOut(GartStreamKind id, int lrec)

/************************************************************************
**
** PARAMETERS
**	id	 - stream kind
**	lrec	 - record size (for Shared Memory: size of SHM)
**
** FUNCTION DEFINITION
**	Open stream for DstOut and MiniOut. 
*************************************************************************
*/

{
int		rc;
int		gLun;
int		count;
GartStreamData  *sPtr;
GartStreamData  *sPtr2;

sPtr = &gartStream[id];
      
/*
 * Get free stream lun from GPACK 
 */
if (sPtr->isShm)
   gLun = gp_setstream(ShmStream);
else
   {
   /*
   **  Check: if Dst and Mini files are the same
   */
   switch (id)
      {
      case GartStreamDstOut:
         sPtr2 = &gartStream[GartStreamMiniOut];
         if (sPtr2->isOpen)
            {
            if (strcmp(sPtr->fileName,sPtr2->fileName) == 0)
               {
               static int errPtr;
               
               sprintf(errMess,
                  "GART: Cannot write Dst and Mini to the same file");
               mecErrorCount(&errPtr, errMess);
               return GP_NOT_OPEN;
               }
            }
         break;
         
      case GartStreamMiniOut:
         sPtr2 = &gartStream[GartStreamDstOut];
         if (sPtr2->isOpen)
            {
            if (strcmp(sPtr->fileName,sPtr2->fileName) == 0)
               {
               static int errPtr;
               
               sprintf(errMess,
                  "GART: Cannot write Dst and Mini to the same file");
               mecErrorCount(&errPtr, errMess);
               return GP_NOT_OPEN;
               }
            }
         break;
         
      default:
         break;
      }
   
   gLun = gp_setstream(FileStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif
   }
   
if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
 * Set event type: Dst or Mini 
 */
if (id == GartStreamDstOut)
   sPtr->eventType = Dst;
else
   sPtr->eventType = Mini;

#ifdef DEBUG
 printf("gp_setevtype %d\n",gLun);
#endif
gp_setevtype(gLun, sPtr->eventType);

/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %d",rc);
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 *  Set Shared Memory key
 */
 
if (sPtr->isShm)
   {
   sPtr->shmKey = GartShmKeyByName(sPtr->fileName);
   if (sPtr->shmKey == -1)
      {
      static int errPtr;

      sprintf(errMess, "GART: cannot access/create file: %s\n",sPtr->fileName);
      mecErrorCount(&errPtr, errMess);
      return GART_NO_FILE;
      }
   
#ifdef DEBUG
 printf("gp_setkey %d\n",gLun);
#endif
   if ((rc = gp_setkey(gLun, sPtr->shmKey)) < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set SHM key=%d. rc=%d: %s\n",
         sPtr->shmKey, rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
else
/*
 *  Set the file name and open mode
 */
   {
#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
   if ((rc = gp_setname(gLun, sPtr->fileName, WriteOnly)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
         rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/* Open file */
if (!sPtr->isShm)
   {
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   if ((rc = gp_open(gLun)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot open file: %s\n",
         gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
/* Open or Create Shared Memory */
else
   {
   count = 0;
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   while ((rc = gp_open(gLun)) < 0)
      {
      if (count++ > 0)
         {
         static int errPtr;
         
         sprintf(errMess, "GART: Cannot open SHM. rc = %d : %s\n",
            rc, gp_errmsg(gLun));
         mecErrorCount(&errPtr, errMess);
         return rc;
         }
#ifdef DEBUG
 printf("gp_create_ring %d\n",gLun);
#endif
      if ((rc = gp_create_ring(gLun)) < 0)
         {
         static int errPtr;
         
         sprintf(errMess, "GART: Cannot create SHM. rc = %d : %s\n",
            rc, gp_errmsg(gLun));
         mecErrorCount(&errPtr, errMess);
         return rc;
         }
      }
   }

/*
 *  Optionally, immediately after open, one can set target type if he wants: 
 *     IEEE_BigEndian        (SGI, PowerPC, HP, ...)
 *     IEEE_LittleEndian     (Linux PCs, OSF Alpha's)
 *     G_Float_LittleEndian  (OpenVMS Alpha's)
 *
 *  By convention in Arte we use IEEE_LittleEndian
 */
#ifdef DEBUG
 printf("gp_settarget %d\n",gLun);
#endif
if ((rc = gp_settarget(gLun, IEEE_LittleEndian)) < 0)
   {
   static int errPtr;
    
   sprintf(errMess, "GART: Cannot set target type. rc = %d : %s\n",
      rc, gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
/*
** For an output stream, one can also set zipper, if he wants
*/
if ((rc = gp_setzipper(sPtr->gpackLun, sPtr->zipcode)) < 0)
   {
   static int errPtr;

   sprintf(errMess, "GART: cannot set zipper to %d on GPACK lun %d "
           "(gpack rc ==%d)", sPtr->zipcode, sPtr->gpackLun, rc);
   mecErrorCount(&errPtr, errMess);
   }

sPtr->isOpen = GartOpened;
sPtr->isOutput = 1; /* Output unit */
return GART_OK;
}

static int GartOpenOthers(GartStreamData *sPtr, int lrec)

/************************************************************************
**
** PARAMETERS
**	sPtr	 - pointer to stream structure
**	lrec	 - record size (for Shared Memory: size of SHM)
**
** FUNCTION DEFINITION
**	Open stream for Other units. 
*************************************************************************
*/

{
int		rc;
int		gLun;
int		count;

/*
 * Get free stream lun from GPACK 
 */
if (sPtr->isShm)
   gLun = gp_setstream(ShmStream);
else
   {
   gLun = gp_setstream(FileStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif
   }
   
if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
 * Set event type: Dst
 */
sPtr->eventType = Dst;

#ifdef DEBUG
 printf("gp_setevtype %d\n",gLun);
#endif
gp_setevtype(gLun, sPtr->eventType);

/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %d",rc);
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 *  Set Shared Memory key
 */
 
if (sPtr->isShm)
   {
   sPtr->shmKey = GartShmKeyByName(sPtr->fileName);
   if (sPtr->shmKey == -1)
      {
      static int errPtr;

      sprintf(errMess, "GART: cannot access/create file: %s\n",sPtr->fileName);
      mecErrorCount(&errPtr, errMess);
      return GART_NO_FILE;
      }
   
#ifdef DEBUG
 printf("gp_setkey %d\n",gLun);
#endif
   if ((rc = gp_setkey(gLun, sPtr->shmKey)) < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set SHM key=%d. rc=%d: %s\n",
         sPtr->shmKey, rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
else
/*
 *  Set the file name and open mode
 */
   {
   int rwOption;
#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
   
   if (sPtr->isOutput)
      rwOption = WriteOnly;
   else
      rwOption = ReadOnly;
      
   if ((rc = gp_setname(gLun, sPtr->fileName, rwOption)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
         rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/* Open file */
if (!sPtr->isShm)
   {
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   if ((rc = gp_open(gLun)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot open file: %s\n",
         gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
/* Open or Create Shared Memory */
else
   {
   count = 0;
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   while ((rc = gp_open(gLun)) < 0)
      {
      if (count++ > 0)
         {
         static int errPtr;
         
         sprintf(errMess, "GART: Cannot open SHM. rc = %d : %s\n",
            rc, gp_errmsg(gLun));
         mecErrorCount(&errPtr, errMess);
         return rc;
         }
#ifdef DEBUG
 printf("gp_create_ring %d\n",gLun);
#endif
      if ((rc = gp_create_ring(gLun)) < 0)
         {
         static int errPtr;
         
         sprintf(errMess, "GART: Cannot create SHM. rc = %d : %s\n",
            rc, gp_errmsg(gLun));
         mecErrorCount(&errPtr, errMess);
         return rc;
         }
      }
   }

/*
 *  Optionally, immediately after open, one can set target type if he wants: 
 *     IEEE_BigEndian        (SGI, PowerPC, HP, ...)
 *     IEEE_LittleEndian     (Linux PCs, OSF Alpha's)
 *     G_Float_LittleEndian  (OpenVMS Alpha's)
 *
 *  By convention in Arte we use IEEE_LittleEndian
 */
#ifdef DEBUG
 printf("gp_settarget %d\n",gLun);
#endif
if ((rc = gp_settarget(gLun, IEEE_LittleEndian)) < 0)
   {
   static int errPtr;
    
   sprintf(errMess, "GART: Cannot set target type. rc = %d : %s\n",
      rc, gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
/*
** For an output stream, one can also set zipper, if he wants
*/
if (sPtr->isOutput && (rc = gp_setzipper(sPtr->gpackLun, sPtr->zipcode)) < 0)
   {
   static int errPtr;

   sprintf(errMess, "GART: cannot set zipper to %d on GPACK lun %d "
           "(gpack rc ==%d)", sPtr->zipcode, sPtr->gpackLun, rc);
   mecErrorCount(&errPtr, errMess);
   }
      
sPtr->isOpen = GartOpened;
return GART_OK;
}

/***********************************************************************/

static int GartOpenNoteOut(int lrec)

/************************************************************************
**
** PARAMETERS
**	lrec	 - record size
**
** FUNCTION DEFINITION
**	Open stream for NoteOut. 
*************************************************************************
*/

{
int		rc;
int		gLun;
GartStreamData  *sPtr;

sPtr = &gartStream[GartStreamNoteOut];
      
/*
 * Get free stream lun from GPACK 
 */
gLun = gp_setstream(PrivateStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif

if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
 * Set event type: Mini (but this does not matter)
 */
sPtr->eventType = Mini;

#ifdef DEBUG
 printf("gp_setevtype %d\n",gLun);
#endif
gp_setevtype(gLun, sPtr->eventType);

/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %d",rc);
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 *  Set the file name and open mode
 */

#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
if ((rc = gp_setname(gLun, sPtr->fileName, WriteOnly)) < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
      rc, gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
 
/* Open file */

#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
if ((rc = gp_open(gLun)) < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "GART: Cannot open file: %s\n",
      gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
      
sPtr->isOpen = GartOpened;
sPtr->isOutput = 1; /* Output unit */
return GART_OK;
}

/***********************************************************************/

static int GartOpenNoteIn(GartStreamKind id, int lrec)

/************************************************************************
**
** PARAMETERS
**	id	 - stream kind
**	lrec	 - record size
**
** FUNCTION DEFINITION
**	Open stream for NoteIn, DstInNote, MiniInNote. 
*************************************************************************
*/

{
int		rc;
int		gLun;
GartStreamData  *sPtr;
char		*selectionCriteria;

sPtr = &gartStream[id];

/*
 * Get free stream lun from GPACK 
 */
gLun = gp_setstream(PrivateStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif

if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
 * Set event type
 */
switch(id)
   {
   case GartStreamDstInNote:
 /*=========================*/
      sPtr->eventType = Dst;
      break;
      
   case GartStreamMiniInNote:
 /*==========================*/
      sPtr->eventType = Mini;
      break;
      
   case GartStreamNoteIn:
 /*======================*/
      sPtr->eventType = Indx;
      break;

   case GartStreamEvdirIn:
 /*======================*/
      sPtr->eventType = Indx;
      break;

   default:
      sPtr->eventType = Indx;
   }

#ifdef DEBUG
 printf("gp_setevtype %d\n",gLun);
#endif
gp_setevtype(gLun, sPtr->eventType);

/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %d",rc);
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 *  Set the file name and open mode
 */

#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
if ((rc = gp_setname(gLun, sPtr->fileName, ReadOnly)) < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
      rc, gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
 
/*
 *  Set Selection Criteria
 */

selectionCriteria = GetSelectionCriteria(); 
#ifdef DEBUG
 printf("gp_select %d\n",gLun);
#endif
if (selectionCriteria != NULL)
   rc = gp_select(gLun,selectionCriteria);
else
   rc = gp_select(gLun,selectionNullText);
   
if (rc  < 0)
   {
   static int errPtr;

   printf("%s\n",gp_errmsg(gLun));
   sprintf(errMess, "GART: Select failed. rc=%s\n",
      gp_rcmsg(rc));
   mecErrorCount(&errPtr, errMess);
   return GART_BAD_SELECT;
   }
  
/* Open file */

#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
if ((rc = gp_open(gLun)) < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "GART: Cannot open file: %s\n",
      gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
      
sPtr->isOpen = GartOpened;
sPtr->isOutput = 0; /* Input unit */
return GART_OK;
}

/***********************************************************************/

static int GartOpenEvIn(GartStreamKind id, int lrec)

/************************************************************************
**
** PARAMETERS
**	id	 - stream kind
**	lrec	 - record size 
**
** FUNCTION DEFINITION
**	Open stream for EventIn, EvIn1, EvIn2, EvIn3, EvIn4 
*************************************************************************
*/

{
int		rc;
int		gLun;
GartStreamData  *sPtr;

sPtr = &gartStream[id];
      
/*
 * Get free stream lun from GPACK 
 */
if (sPtr->isShm)
   gLun = gp_setstream(ShmStream);
else
   gLun = gp_setstream(FileStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif
if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %d",rc);
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
 * Set event type: it will be reset after reading of EventHdr 
 */
sPtr->eventType = -1;

/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %d",rc);
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 *  Set Shared Memory key
 */
if (sPtr->isShm)
   {
   sPtr->shmKey = GartShmKeyByName(sPtr->fileName);
   if (sPtr->shmKey == -1)
      {
      static int errPtr;

      sprintf(errMess, "GART: file does not exist: %s\n",sPtr->fileName);
      mecErrorCount(&errPtr, errMess);
      return GART_NO_FILE;
      }
   
#ifdef DEBUG
 printf("gp_setkey %d\n",gLun);
#endif
   if ((rc = gp_setkey(gLun, sPtr->shmKey)) < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set SHM key=%d. rc=%d: %s\n",
         sPtr->shmKey, rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
else
/*
 *  Set the file name and open mode
 */
   {
#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
   if ((rc = gp_setname(gLun, sPtr->fileName, ReadOnly)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
         rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/* Open file */
if (!sPtr->isShm)
   {
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   if ((rc = gp_open(gLun)) < 0)
      {
      static int errPtr;
          
      sprintf(errMess, "GART: Cannot open file: %s\n",
         gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
/* Open Shared Memory */
else
   {
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
   if ((rc = gp_open(gLun)) < 0)
      {
      static int errPtr;
         
      sprintf(errMess, "GART: Cannot open SHM. rc = %d : %s\n",
         rc, gp_errmsg(gLun));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }
      
sPtr->isOpen = GartOpened;
sPtr->isOutput = 0; /* Input unit */
return GART_OK;
}

/***********************************************************************/

static int GartOpenEvdir(GartStreamKind id, int lrec)

/************************************************************************
**
** PARAMETERS
**	id	 - stream kind
**	lrec	 - record size 
**
** FUNCTION DEFINITION
**	Open stream for Evdir: EvdirIn, DstInEvdir, MiniInEvdir
*************************************************************************
*/

{
int		rc;
int		gLun;
GartStreamData  *sPtr;
char		*selectionCriteria;

sPtr = &gartStream[id];
      
/* Close if already opened */
if (sPtr->isOpen && sPtr->isOpen != GartOpenEndOfData)
   {
#ifdef DEBUG
 printf("gp_close %d\n",sPtr->gpackLun);
#endif
   gLun = sPtr->gpackLun;
   gp_close(gLun);
   gp_delstream(gLun);
   }
sPtr->isOpen = GartNotOpen;

/*
 * Get free stream lun from GPACK 
 */
gLun = gp_setstream(PublicStream);
#ifdef DEBUG
 printf("gp_setstream %d\n",gLun);
#endif
if (gLun < 0)
   {
   static int errPtr;
   
   rc = gLun;
   sprintf(errMess, "GART: Cannot get GPACK lun. rc = %s",gp_rcmsg(rc));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
sPtr->gpackLun = gLun;

/*
** for DstInEvdir it is necessary to copy file(s) from OSM robot before... 
*/
/*
if (id == GartStreamDstInEvdir)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: DstInEvdir is not yet properly implemented");
   mecErrorCount(&errPtr, errMess);
   }
*/

sPtr->isShm = 0;
/*
 * Set internal buffer size 
 */
if (lrec)
   {
#ifdef DEBUG
 printf("gp_setbuf %d\n",gLun);
#endif
   rc = gp_setbuf(gLun,lrec);
   if (rc < 0)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Cannot set GPACK buffer. rc = %s",gp_rcmsg(rc));
      mecErrorCount(&errPtr, errMess);
      return rc;
      }
   }

/*
 * Set event type 
 */

switch(id)
   {
   case GartStreamEvdirIn:
 /*==========================*/
      sPtr->isOutput = 0; /* Input unit */
      sPtr->eventType = Mini;
      break;
      
   case GartStreamDstInEvdir:
 /*==========================*/
      sPtr->isOutput = 0; /* Input unit */
      sPtr->eventType = Dst;
      break;
      
   case GartStreamMiniInEvdir:
 /*===========================*/
      sPtr->isOutput = 0; /* Input unit */
      sPtr->eventType = Mini;
      break;

   default: /* never reached, added to avoid compiler warnings */
      sPtr->eventType = -1;
   
   }
      
#ifdef DEBUG
 printf("gp_setevtype %d\n",gLun);
#endif
gp_setevtype(gLun, sPtr->eventType);
  

/*
 *  Set file name
 */

#ifdef DEBUG
 printf("gp_setname %d\n",gLun);
#endif
if ((rc = gp_setname(gLun, sPtr->fileName, ReadOnly)) < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "GART: Cannot set filename. rc = %d : %s\n",
      rc, gp_errmsg(gLun));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }

  
/*
 *  Set Selection Criteria
 */

selectionCriteria = GetSelectionCriteria(); 
#ifdef DEBUG
 printf("gp_select %d\n",gLun);
#endif
if (selectionCriteria != NULL)
   rc = gp_select(gLun,selectionCriteria);
else
   rc = gp_select(gLun,selectionNullText);

if (rc  < 0)
   {
   static int errPtr;

   printf("%s\n",gp_errmsg(gLun));
   sprintf(errMess, "GART: Select failed. rc=%s\n",
      gp_rcmsg(rc));
   mecErrorCount(&errPtr, errMess);
   return GART_BAD_SELECT;
   }

/* Open file */
#ifdef DEBUG
 printf("gp_open %d\n",gLun);
#endif
rc = gp_open(gLun);
#ifdef DEBUG
 printf("open -----> %s\n",gp_rcmsg(rc));
#endif
if (rc == GP_END_OF_DATA)
   {
   sPtr->isOpen = GartOpenEndOfData;
   return GART_OK;
   }
   
if (rc < 0)
   {
   static int errPtr;
          
   sprintf(errMess, "%s\n",gp_errmsg(gLun));
   printf("%s",errMess);
   sprintf(errMess, "GART: Cannot open Event Dir.\n");
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
      
sPtr->isOpen = GartOpened;
return GART_OK;
}

/***********************************************************************/

static int GartDstMiniOut(GartStreamData *sPtr, int tab, int inew)

/************************************************************************
**
** PARAMETERS
**	sPtr	 - pointer to a stream structure
**	tab	 - table index (data start from iq(jlink(tab+1)) )
**	inew	 - start of event flag: =1 for new event, =0 otherwize
**
** RETURNS
**			 0 normal completion
**			+1 "pseudo end of tape"
**			-2 error return
**
** FUNCTION DEFINITION
**	Write one table to stream on id.
*************************************************************************
*/

{
int		rc;
int		ncols;
int		nrows;
int		*data;
int		*adr;
char		*tName;
char		*tDesc;
EventHdr	header;

tName = ArteTableName(tab);

if (sPtr->isOpen == GartOpenWritten)	/* something is written to this lun */
   {
   /* check file limit */
   if (!sPtr->isShm && sPtr->limit != 0)
      {
      int size;
   
#ifdef DEBUG
 printf("gp_getevtpos %d\n",sPtr->gpackLun);
#endif
      size = gp_getevtpos(sPtr->gpackLun);
      if (inew && size >= sPtr->limit)
         return (1);
      }
   }
   
/* write header if beginning of event */
if (inew)
   {
   header.type		= sPtr->eventType;
   HeaderFromEvhd(&header);

   /* fill EvdInfo structure */
   
   EvdInfoFromHeader(&sPtr->evdInfo, &header);
   sPtr->evdInfo.dstPos		= -1;
   sPtr->evdInfo.miniPos	= -1;
   sPtr->evdInfo.fileNumber	= 0;
#ifdef DEBUG
   printf("%s:%d: lun %d, fileName %s\n",__FILE__,__LINE__,
	sPtr->gpackLun,sPtr->fileName);
#endif
   strcpy(sPtr->evdInfo.dataname,sPtr->fileName);


   sPtr->isOpen = GartOpenWritten; /* something is written to this lun */

#ifdef DEBUG
 printf("gp_putevt %d\n",sPtr->gpackLun);
#endif
   rc = gp_putevt(sPtr->gpackLun,&header);
   if (rc < 0)
      {
      static int errPtr;
         
      sprintf(errMess, "GART: Cannot put header %s. rc = %d : %s\n",
         tName, rc, gp_errmsg(sPtr->gpackLun));
      mecErrorCount(&errPtr, errMess);
      
      return (-2);
      }
   }

/* 
**write table 
*/
adr = gcbank_.iq + artelink_.jlink[tab-1];

ncols = *adr;
nrows = *(adr+1);
data  = adr+2;

tDesc = ArteTableDesc(tab,tName);
if (tDesc == NULL)
   return (-2);

#ifdef DEBUG
 printf("gp_puttab %d ncols=%5d nrows=%5d %s\n",
    sPtr->gpackLun,ncols,nrows,tName);
#endif

rc = gp_puttab(sPtr->gpackLun,tName,ncols,nrows,tDesc,data);
if (rc < 0)
   {
   static int errPtr;
         
   sprintf(errMess, "GART: Cannot put table %s. rc = %d : %s\n",
      tName, rc, gp_errmsg(sPtr->gpackLun));
   mecErrorCount(&errPtr, errMess);

   return (-2);
   }

return (0);
}

/***********************************************************************/

static int GartNoteOut()

/************************************************************************
**
** PARAMETERS
**	None
**
** RETURNS
**			 0 normal completion
**			+1 "pseudo end of tape"
**			-2 error return
**
** FUNCTION DEFINITION
**	Write to notebook.
*************************************************************************
*/

{
int		rc;
GartStreamData  *sPtr;
GartStreamKind	id;

#ifdef DEBUG
/*TEMP*/ printf("GartNoteOut: begin\n");
#endif

if (!gartStream[GartStreamNoteOut].isOpen)
   return(0);
     
#ifdef DEBUG
/*TEMP*/ printf("GartNoteOut: begin +\n");
#endif
/*
**  Write to Notebook for output streams
*/
sPtr = &gartStream[GartStreamDstOut];
if (sPtr->isOpen && sPtr->isShm)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: DstOut is SHM, cannot write to Notebook!\n");
   mecErrorCount(&errPtr, errMess);
   return(-2);
   } 

sPtr = &gartStream[GartStreamMiniOut];
if (sPtr->isOpen && sPtr->isShm)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: MiniOut is SHM, cannot write to Notebook!\n");
   mecErrorCount(&errPtr, errMess);
   return(-2);
   } 

#ifdef DEBUG
/*TEMP*/ printf("GartNoteOut: begin ++\n");
#endif
if (gartStream[GartStreamDstOut].isOpen ||
    gartStream[GartStreamMiniOut].isOpen)
    return(0);	/* Already treated by gartendev */

/*
**  Write to Notebook for input streams
*/

/* No Notebook output for Mixing*/
 if (ziocom_.fiomix[4] != 0.)
 {
      static int errPtr;
      
      sprintf(errMess, "GART: cannot write to Notebook with event mixing\n");
      mecErrorCount(&errPtr, errMess);
      return(-2);
 }

#ifdef DEBUG
/*TEMP*/ printf("GartNoteOut: begin +++\n");
#endif
 for (id=GartStreamEventIn; id <= GartStreamMiniInEvdir; id++)
 {
      EvdInfo   newInfo;
#ifdef DEBUG
/*TEMP*/ printf("GartNoteOut: %s\n",GartStreamName(id));
#endif
      sPtr = &gartStream[id];
      if( !sPtr->isOpen )
	   continue;

      if( sPtr->isShm )
      {
	   static int errPtr;
   
	   sprintf(errMess,"GART: read from SHM, cannot write to Notebook!\n");
	   mecErrorCount(&errPtr, errMess);
	   return(-2);
      }
      newInfo.prop.year      = sPtr->evdInfo.prop.year;
      newInfo.prop.exp       = sPtr->evdInfo.prop.exp;
      newInfo.prop.run       = sPtr->evdInfo.prop.run;
      newInfo.prop.event     = sPtr->evdInfo.prop.event;
      newInfo.prop.datime    = sPtr->evdInfo.prop.datime;
      newInfo.prop.classmask = EVHD(1).flag;
      newInfo.dstPos         = sPtr->evdInfo.dstPos;
      newInfo.miniPos        = sPtr->evdInfo.miniPos;
      newInfo.fileNumber     = 0;
      gp_getdataname(sPtr->gpackLun, newInfo.dataname);

#ifdef DEBUG
/*TEMP*/ printf("%s:%d: lun %d, dataname <%s>\n",__FILE__,__LINE__,
		gartStream[GartStreamNoteOut].gpackLun,newInfo.dataname);
#endif

      gartStream[GartStreamNoteOut].isOpen = GartOpenWritten;
      rc = gp_putevdinfo(gartStream[GartStreamNoteOut].gpackLun,
			 &newInfo);
   
      if (rc < 0)
      {
	   static int errPtr;
         
	   sprintf(errMess, "GART: Cannot put Evd. info. rc = %d: %s\n",
		   rc, gp_errmsg(sPtr->gpackLun));
	   mecErrorCount(&errPtr, errMess);
	   return(-2);
      }
      return(0);
 }

 {
      static int errPtr;
      
      sprintf(errMess,
	      "GART: Cannot put Evd. info: no input/output streams opened\n");
      mecErrorCount(&errPtr, errMess);
      return(-2);
 }
}

/***********************************************************************/

static int  ArteDatimeToUnixTime(int aDate, int aTime)

/************************************************************************
**
** PARAMETERS
**	aDate	 - Arte date representation:	yymmdd
**	aTime	 - Arte time representation:	hhmmss
**
** RETURNS
**	Unix time representation of Arte date/time
**
** FUNCTION DEFINITION
**	Returns Unix time representation of Arte date/time
*************************************************************************
*/

{
struct tm timeStr;
int	year;
int	mon;
int	day;
int	hour;
int	min;
int	sec;
int	daylightSave;
#if defined(sgi) || defined(sun) 
time_t	altzoneSave;
#endif
time_t	timezoneSave;
time_t	result;


daylightSave = daylight;	daylight = 0;
#if defined(sgi) || defined(sun) 
altzoneSave  = altzone;		altzone  = 0;
#endif
timezoneSave = timezone;	timezone = 0;

year = aDate/10000;
if (year > 96)
   {
   year = 1900 + year;
   }
else
   {
   year = 2000 + year;
   }

mon  = (aDate%10000)/100;
day  = aDate%100;

hour = aTime/10000;
min  = (aTime%10000)/100;
sec  = aTime%100;
   
timeStr.tm_year = year - 1900;
timeStr.tm_mon	= mon - 1;
timeStr.tm_mday	= day;
timeStr.tm_hour	= hour;
timeStr.tm_min	= min;
timeStr.tm_sec	= sec;
timeStr.tm_isdst = -1;

result = mktime(&timeStr);

daylight = daylightSave;
#if defined(sgi) || defined(sun) 
altzone  = altzoneSave;
#endif
timezone = timezoneSave;

return(result);
}

/***********************************************************************/

static void ArteDatimeFromUnixTime(int datime, int *aDate, int *aTime)

/************************************************************************
**
** PARAMETERS
**	datime	 - Unix time representation
**	aDate	 - Arte date representation:	yymmdd
**	aTime	 - Arte time representation:	hhmmss
**
** RETURNS
**	None
**
** FUNCTION DEFINITION
**	Returns Arte date/time representation of Unix time
*************************************************************************
*/

{
struct tm *timeStr;
int	year;
int	mon;
int	day;
int	hour;
int	min;
int	sec;
int	daylightSave;
#if defined(sgi) || defined(sun) 
time_t	altzoneSave;
#endif
time_t	timezoneSave;


daylightSave = daylight;	daylight = 0;
#if defined(sgi) || defined(sun) 
altzoneSave  = altzone;		altzone  = 0;
#endif
timezoneSave = timezone;	timezone = 0;

timeStr = localtime((time_t *)&datime);

daylight = daylightSave;
#if defined(sgi) || defined(sun) 
altzone  = altzoneSave;
#endif
timezone = timezoneSave;

year = timeStr->tm_year;
if (year > 99)
   year = year - 100;

mon  = timeStr->tm_mon + 1;
day  = timeStr->tm_mday;

hour = timeStr->tm_hour;
min  = timeStr->tm_min;
sec  = timeStr->tm_sec;

*aDate = day + 100*mon + 10000*year;
*aTime = sec + 100*min + 10000*hour;

return;
}

/***********************************************************************/

static char *ArteTableName(int tab)

/************************************************************************
**
** PARAMETERS
**	tab	 - table index (data start from iq(jlink(tab)+1) )
**
** RETURNS
**	Arte table name
**
** FUNCTION DEFINITION
**	Returns table name as a null terminated string.
*************************************************************************
*/

{
int  len = namelen;

chhtot_ ((int *)MEMO(tab).name, tableName, &len);
return(tableName);
}

/***********************************************************************/

static char *ArteTableDesc(int tab, char *tName)

/************************************************************************
**
** PARAMETERS
**	tab	 - table index (data start from iq(jlink(tab+1)) )
**	tName	 - table name (for printouts)
**
** RETURNS
**	Arte table description to be used in GPACK
**
** FUNCTION DEFINITION
**	Returns Arte table description to be used in GPACK.
*************************************************************************
*/

{
int	mema;
int	nCol;
int	iCol;
int	col;
int	isFirst;
int	type;
char	oneDesc[100];
int	lastType = -1;
int	lastIlow = 0;
int	lastIupp = 0;
float	lastFlow = 0.0;
float	lastFupp = 0.0;
float	lastAccu = 0.0;
int	nEq;
int	iRpt;
int	changed;


strcpy(tableDesc,"");
isFirst = 1;

mema = MEMO(tab).mema;
nCol = MEMO(tab).ocol;

nEq = 0;
changed = 0;

for (iCol=0; iCol<nCol; iCol++)
   {
   col = mema+iCol;
   type = MEMA(col).type;

   /* relations */
   if (type == 1 || type == 2 || type == 3 || type == 4)
      type = 1;

   /* char */
   if (type == 11)
      type = 8;

   /* short */
   if (type == 9)
      type = 5;

   if (type == 10)
      type = 6;

   /* check possible types */
   if (type <= 0 || type > 11)
      {
      static int errPtr;
      
      sprintf(errMess, "GART: Table %s is not written: unknown type %d in MEMA\n",
         tName,type);
      mecErrorCount(&errPtr, errMess);
      return(NULL);
      }

   /* 
   ** If sequence of equal attributes is empty, save attributes 
   */ 
   if (nEq == 0)
      {
      lastIlow = MEMA(col).ilow;
      lastIupp = MEMA(col).iupp;
      lastFlow = MEMA(col).flow;
      lastFupp = MEMA(col).fupp;
      lastAccu = MEMA(col).accu;
      lastType = type;
      
      changed = 0;
      nEq = 1;
      if (iCol != nCol-1)
         continue;
      }
   /* 
   ** Otherwize check that previos and current attibutes are the same 
   */ 
   else
      {
      changed = 1;
      if (type == lastType)
         {
         switch (type)
            {
            case 5:	/* int */
               if (MEMA(col).ilow == lastIlow && MEMA(col).iupp == lastIupp)
                  {
                  nEq++;
                  changed = 0;
                  if (iCol != nCol-1)
                     continue;
                  }
               break;
            
            case 6:	/* byte */
               if (MEMA(col).iupp == lastIupp)
                  {
                  changed = 0;
                  nEq++;
                  if (iCol != nCol-1)
                     continue;
                  }
               break;
            
            case 7:	/* float */
               if (
                   MEMA(col).flow == lastFlow &&
                   MEMA(col).fupp == lastFupp &&
                   MEMA(col).accu == lastAccu
                  )
                  {
                  changed = 0;
                  nEq++;
                  if (iCol != nCol-1)
                     continue;
                  }
               break;
            
            default:	/* others have no attributes */
               changed = 0;
               nEq++;
               if (iCol != nCol-1)
                  continue;
            }
         }
      }


   /* 
   ** Generate description string
   */
   for (iRpt=0; iRpt<2; iRpt++)
      {
      strcpy(oneDesc,"");
      if (!isFirst)
         strcat(oneDesc,",");
      else
         isFirst = 0;
   
      if (nEq != 1)
         sprintf(oneDesc+strlen(oneDesc),"%d",nEq);

      /* generate */
      switch(lastType)
/*TEMP	ilow, iupp, flow, fupp, accu - not yet implemented: ArteTableDesc */
/*TEMP	compression code - not yet implemented, use default: ArteTableDesc */
         {
         case 5:	/* int */
            strcat(oneDesc,"I1");
            break;
         
         case 6:	/* byte */
            strcat(oneDesc,"B1");
            break;
         
         case 7:	/* float */
            strcat(oneDesc,"F");
            break;
         
         case 8:	/* holl */
            strcat(oneDesc,"H1");
            break;
         
         case 1:	/* relation */
            strcat(oneDesc,"I1");
            break;
         }

      if (strlen(tableDesc)+strlen(oneDesc) > TABLE_DESC_LEN)
         {
         static int errPtr;
      
         sprintf(errMess, "GART: Table %s is not written: too long description\n",
            tName);
         mecErrorCount(&errPtr, errMess);
         return(NULL);
         }
      else
         strcat(tableDesc,oneDesc);
   
      
      /* new sequence started */
      lastIlow = MEMA(col).ilow;
      lastIupp = MEMA(col).iupp;
      lastFlow = MEMA(col).flow;
      lastFupp = MEMA(col).fupp;
      lastAccu = MEMA(col).accu;
      lastType = type;
      
      /* Save the previous too if this is the last item */
      if (!(iCol == nCol-1 && changed == 1))
         break;
      nEq = 1;
      }
   nEq = 1;
   }

return(tableDesc);
}

/***********************************************************************/

static int GetTabInfo(GartStreamData *sPtr, int *name)

/************************************************************************
**
** PARAMETERS
**	sPtr	 - current stream structure
**	name	 - (output) TABLE_NAME_LEN characters of the next table name 
**
** RETURNS
**	GPACK error code
**
** FUNCTION DEFINITION
**	Read the description of next table from sPtr stream.
*************************************************************************
*/

{
int	rc;
int	len = namelen;

if (sPtr->isOpen == GartOpenEndOfData)
   return GP_END_OF_DATA;

#ifdef DEBUG
 printf("gp_gettabinfo %d\n",sPtr->gpackLun);
#endif
rc = gp_gettabinfo(sPtr->gpackLun,tableName,&sPtr->ncols,&sPtr->nrows);

if (rc == GP_END_OF_DATA || rc == GP_INTERRUPTED)
   return rc;

#ifdef DEBUG
 printf("GetTabInfo: ncols = %7d nrows = %7d tableName = %s\n",
    sPtr->ncols,sPtr->nrows,tableName);
#endif
if (rc < 0)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: Cannot get tabinfo for lun = %d %s",
      sPtr->arteLun,gp_rcmsg(rc));
   mecErrorCount(&errPtr, errMess);
   return rc;
   }
   

chttoh_ (tableName, name, &len);

return(GP_SUCCESS);   
}

/***********************************************************************/

static int  ReadTab(GartStreamData *sPtr, int tab)

/************************************************************************
**
** PARAMETERS
**	sPtr	 - current stream structure
**	tab	 - table index (mmbook(jlink(tab),...) )
**
** RETURNS
**	GPACK error code
**
** FUNCTION DEFINITION
**	Read next table from sPtr stream. Creates the table (MMBOOK).
*************************************************************************
*/

{
int	rc;
char	*tName;
int	ncols;
int	nrowsOld;
int	*data;
int	*adr;

if (sPtr->isOpen == GartOpenEndOfData)
   return GP_END_OF_DATA;

tName = ArteTableName(tab);

/* to be prepared for event mixing: read to the end of a bank */
nrowsOld = 0;
if (artelink_.jlink[tab-1] != 0)
   {
   adr = gcbank_.iq + artelink_.jlink[tab-1];
   nrowsOld = *(adr+1);
   }

mmbook_(&artelink_.jlink[tab-1], &sPtr->nrows);

adr = gcbank_.iq + artelink_.jlink[tab-1];

ncols = *adr;
data  = adr + 2 + nrowsOld*ncols;

if (sPtr->ncols > MEMO(tab).ocol)
   {
   static int errPtr;
   
   sprintf(errMess, "GART: table %s changed from %d to %d output columns",
      tName, sPtr->ncols, MEMO(tab).ocol);
   mecErrorCount(&errPtr, errMess);
   return GP_WRONG_DATA_SIZE;
   }

#ifdef DEBUG
 printf("gp_gettab %d   tName = %s\n",sPtr->gpackLun,tName);
#endif
rc = gp_gettab(sPtr->gpackLun,tName,ncols,sPtr->nrows,tableDesc,data);

return(rc);
}

/***********************************************************************/

static void AlarmHandler(int sig)

/************************************************************************
**
** PARAMETERS
**	sig	 - signal
**
** RETURNS
**	None
**
** FUNCTION DEFINITION
**	Alarm signal handler
*************************************************************************
*/

{
float rest;
if (sig == SIGALRM)
   {
   static int errPtr;
      
   rest = (sigWaitRest * ALRM_WAIT_TIME)/60.;
   sigWaitRest--;
   sprintf(errMess, "GART: Read SHM: End of File after %5.1f minutes\n",rest);
   mecWarningCount(&errPtr, errMess);
   }
}

/***********************************************************************/

static char *GetSelectionCriteria()

/************************************************************************
**
** PARAMETERS
**	None
**
** RETURNS
**	Pointer to Selection Criteria string
**
** FUNCTION DEFINITION
**	Returns the current Selection Criteria string for a given id.
**	Returns NULL if Selection Criteria is not set.
*************************************************************************
*/

{
return (selectionCriteriaText);
}

/***********************************************************************/

static void EvdInfoFromHeader(EvdInfo *info, EventHdr *header)

/************************************************************************
**
** PARAMETERS
**	info	-	output EvdInfo structure
**	header	-	input EventHdr structure
**
** RETURNS
**	None
**
** FUNCTION DEFINITION
**	Copies header information to EvdInfo structure.
*************************************************************************
*/

{
int year;
struct tm *t;
  
t = localtime((time_t *)&header->datime);

year = 1900 + t->tm_year;

info->prop.year		= year;
info->prop.exp		= header->experiment;
info->prop.run		= header->run;
info->prop.event	= header->event;
info->prop.datime	= header->datime;
info->prop.classmask	= header->classmask;
}

/***********************************************************************/

static void HeaderFromEvhd(EventHdr *header)

/************************************************************************
**
** PARAMETERS
**	header	-	output EventHdr structure
**
** RETURNS
**	None
**
** FUNCTION DEFINITION
**	Copies header information from EVHD structure.
*************************************************************************
*/

{
header->event		= EVHD(1).evt;
header->classmask	= EVHD(1).flag;
header->run		= EVHD(1).run;
if (header->run <= 0)
   header->run = 1;
header->experiment	= EVHD(1).exp;
if (header->experiment <= 0)
   header->experiment = 1;
header->datime		= ArteDatimeToUnixTime(EVHD(1).date,EVHD(1).time);

#ifdef DEBUG
/*TEMP*/ header->classmask	= header->event;
#endif

}

void gartselact_(void)

/************************************************************************
**
** PARAMETERS
**	None
** FUNCTION DEFINITION
**	Action for KUIP command "select".
*************************************************************************
*/

{
char		*selText;
GartStreamData  *sPtr;
GartStreamKind	id;
int		rc;
int		lrec = 0;
int		lun;

if ( ku_npar() != 1 )
   {
   static int errPtr;

   sprintf(errMess, "GART: wrong number of parameters to SELECT");
   mecErrorCount(&errPtr, errMess);
   }

selText = ku_gets();

/* No selects for Mixing */
if (ziocom_.fiomix[4] != 0.)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Selection Criteria cannot be set for ev. mixing");
   mecErrorCount(&errPtr, errMess);
   selText = selectionNullText;
   }

if (strlen(selText) > SELECT_CRIT_LEN)
   {
   static int errPtr;
      
   sprintf(errMess, "GART: Selection Criteria Text it too long (not changed)");
   mecErrorCount(&errPtr, errMess);
   return;
   }

/* (re)Set Selection Criteria */

GartInit();

#ifdef DEBUG
 printf("gp_select %d\n",selectionLun);
#endif
rc = gp_select(selectionLun,selText);
   
if (rc  < 0)
   {
   static int errPtr;

   printf("%s\n",gp_errmsg(selectionLun));
   sprintf(errMess, "GART: Select failed (not changed)");
   mecErrorCount(&errPtr, errMess);
   gp_select(selectionLun,selectionCriteriaText);
   return;
   }

strcpy(selectionCriteriaText,selText);

/* (re)Set Selection Criteria for Evdir/Notebook */

for (id = GartStreamDstInNote; id <= GartStreamMiniInEvdir; id++)
   {
   sPtr = &gartStream[id];
   if (sPtr->isOpen)
      {
      lun = sPtr->arteLun;
      gartclos_(&lun, &rc);
      sPtr->arteLun = lun;
      if (id < GartStreamEvdirIn)
         GartOpenNoteIn(id, lrec);
      else
         GartOpenEvdir(id, lrec);
      }
   }

return;
}

/***********************************************************************/

void gartsetzipper_(int *lun, char *zipper, int len)

/************************************************************************
**
** PARAMETERS
**	lun	- stream number from Arte (1..99)
**	zipper	- zipper name
**	len	- lenght of zipper (needed for FORTRAN strings)
**
** FUNCTION DEFINITION
**	Set GPACK output stream on lun to be compressed with zipper.
**
*************************************************************************
*/

{
GartStreamKind	id;
GartStreamData  *sPtr;
int rc;

GartInit();

id = GartGetStreamKind(lun,&sPtr);
switch (id)
   {
   case GartStreamDstOut:	/* no break */
 /*======================*/
   case GartStreamMiniOut:	/* no break */
 /*=======================*/
   case GartStreamOthers:
 /*=======================*/
      break;

   case GartStreamWrong:
 /*=======================*/
      {
      static int errPtr;
   
      sprintf(errMess, "GART: lun %d is not suitable for GPACK", *lun);
      mecWarningCount(&errPtr, errMess);
      return;
      }

   default:
      {
      static int errPtr;

      sprintf(errMess, "GART: lun %d supports no compression", *lun);
      mecWarningCount(&errPtr, errMess);
      return;
      }
   }


if      (!strncasecmp(zipper,  "NC", strlen( "NC")))
   sPtr->zipcode = NoCompression;
else if (!strncasecmp(zipper, "LZO", strlen("LZO")))
   sPtr->zipcode = Z_Lzo1x;
else
   {
   /* FORTRAN string handling */
   static int errPtr;
   int l;
   char *s;

   for (l = len-1; l > 0 && zipper[l] == ' '; l--);
   s = alloca(l+2);
   memcpy(s, zipper, l+1);
   s[l+1] = 0;

   snprintf(errMess, sizeof(errMess)-1, "GART: unknown"
            " compression method '%s' requested", s);
   errMess[sizeof(errMess)-1] = 0;
   mecWarningCount(&errPtr, errMess);
   return;
   }

if (!sPtr->isOpen)
   return;

if (sPtr->isOutput)
   {
   if ((rc = gp_setzipper(sPtr->gpackLun, sPtr->zipcode)) < 0)
      {
      static int errPtr;

      sprintf(errMess, "GART: cannot set zipper to %d on ARTE lun %d "
              "(gpack rc ==%d)", sPtr->zipcode, *lun, rc);
      mecErrorCount(&errPtr, errMess);
      }
   }
else
   {
   static int errPtr;

   sprintf(errMess, "GART: lun %d is opened for INPUT", *lun);
   mecWarningCount(&errPtr, errMess);
   sPtr->zipcode = 0;
   }
}

#ifdef  __cplusplus
       }
#endif
