#include <iostream>
#include <algorithm>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00/06/14  Max      First version - geometrical fit, no mass 
//                    or lifetime estimates.
// 01-02-28  Max      Now const safer
// 01-03-13  Max      Added preprocessor directives and updated to use
//                    GroverParticle::operator==


int GroverJpsiSec(GroverParticleList*   allParticles,
		  const GroverWireList* allWires,
		  const GroverCuts*     usercuts,
		  GroverVertexList*     allPrimaries,
		  GroverVertexList*     allSecondaries)

{
  ////////////////////////////////////////////////////////////////////
  //
  //                       GroverJpsiSec
  //
  // Finds secondary vertices which consist of a J/psi candidate
  // plus zero or more charged particles
  // Author M.Sang 14/06/00
  //
  // Arguments: 
  //
  // 1) allParticles   - Any particle flagged as 'not used in primary'
  //                     will be used to build a secondary candidate 
  // 2) allWires       - Pointer to wires
  // 3) cuts           - If the input pointer=0 or absent, default cuts 
  //                     are used.
  // 4) allPrimaries   - On input this is a list of primary vertices.
  //                     On output they may have been modified! 
  //                     (refit, for example)
  // 5) allSecondaries - On input, this is a (possibly empty) list of 
  //                     two-track J/psi candidates. If empty, candidates 
  //                     are found using the Grover standard J/psi
  //                     finder. The output will contain ZERO OR MORE
  //                     secondaries.
  //
  ////////////////////////////////////////////////////////////////////


  if (allPrimaries == 0) return FAILURE;

  const int ASIZE = 1500;
  const float m_jpsi = 3.097, m_K = 0.494, m_pi = 0.1396;


  // Counter and error flag
  static int EventNumber = 0;
  static bool firstcall = true;
  EventNumber++;
  firstcall = (EventNumber==1);

  // Non-default cuts?
  static GroverCuts cuts;
  if (firstcall && usercuts != 0) cuts = *usercuts; //(Memberwise copy)
  static bool verbose = (cuts.Verbose==1);



  // Initialisation
  if (firstcall || verbose)
    cout << "*** GroverJpsiSec (Grover v."
	 << GroverVersion 
	 << "): first call" << endl;


  // Label primaries
  int idx = 0;
  for (GroverVertexList::iterator iprim = allPrimaries->begin();
       iprim != allPrimaries->end(); ++iprim) iprim->SetIndex(++idx);



  // User-defined particle list? If not, fill our own list
  // from RTRA (default) or from RSEG
  int error = 0;
  static bool use_rseg = (cuts.RPTrackChoice==2);
  static bool use_rtra = (cuts.RPTrackChoice==3);
  static bool use_rover =(cuts.RPTrackChoice==1);
  static GroverParticleList standby_par;
  if (!standby_par.empty()) standby_par.clear();
  GroverParticleList* particles = allParticles;
  if (particles == 0) {
    particles = &standby_par;
    // Use all RTRAs unless user specifies RSEGs
    if (use_rseg) error       = GroverFillFromRSEG(particles);
    else if (use_rtra) error  = GroverFillFromRTRA(particles);
    else if (use_rover) error = GroverFillLikeRover(particles); 
    else error = -1;
    if (error != 0) {
      if (verbose) 
	cout << "*** GroverJpsiSec: WARNING - no tracks!";
    }
    else {
      // flag the ones used in primaries! It would save time if this
      // were done in the primary finder, but not guaranteed yet
      for (GroverParticleList::iterator ipar = particles->begin();
	   ipar != particles->end(); ++ipar) {
	for (GroverVertexList::iterator iprim = allPrimaries->begin();
	     iprim != allPrimaries->end(); ++iprim) {
	  for (GroverParticleList::iterator jpar = iprim->particles.begin();
	       jpar != iprim->particles.end(); ++jpar) {
	    if (ipar->Type()==jpar->Type()) {
	      // Good - they are comparable (eg 2 RTRAs)
	      if (*ipar == *jpar) {
		ipar->SetUsedInPrimary(iprim->Index());
	      }
	    }
	  }
	}
      }
    }
  }
  else if (particles->empty()) {
    cout << "*** GroverJpsiSec: FAILED - user passed an empty particle list" 
	 << endl;
    return FAILURE;
  }




  // Label particles with an index and move to local VDS coordinates
  idx = 0;
  for (GroverParticleList::iterator ipar = particles->begin();
       ipar != particles->end(); ++ipar) {
    ipar->SetIndex(++idx);
    ipar->GlobalToLocal();
  }




  if (verbose) cout << "checking for jpsi candidates in usr list " << endl;
  // Any J/psi candidates from the user? If not, find them ourselves
  static GroverVertexList standby_psi;
  if (!standby_psi.empty()) standby_psi.clear();
  GroverVertexList* jpsis = allSecondaries;
  if (jpsis == 0 ) {
    jpsis = &standby_psi;
    if (verbose && firstcall) 
      cout << "*** GroverJpsiSec: Empty J/psi list passed - " 
	   << "I will search for them myself" << endl;
    error = -1;
    //error = GroverFindJpsis(particles,&cuts,jpsis); // Nothing in this yet!
    if (error != 0) {
      if (verbose) cout << "*** GroverJpsiSec: FAILED - "
		     "no J/psi candidates found" << endl;
      return FAILURE;
    }
  }




  if (verbose) cout << "checking against cuts" << endl;
  // If this is a user J/psi candidate list, throw out any which don't
  // have two tracks or which don't come closer than N cm or M sigma
  static float chi2cut  = cuts.RSJpsiSeedMaxChi2;
  static float spacecut = cuts.RSJpsiSeedMaxDist;
  static GroverParticleList::iterator par1, par2;
  if (verbose) cout << jpsis->size() << " jpsi candidates" << endl;
  for (GroverVertexList::iterator ivert = jpsis->begin();
       ivert != jpsis->end(); ++ivert) {
    int ntracks = ivert->particles.size();
    if (ntracks!=2) jpsis->erase(ivert--);
    else {
      par1 = ivert->particles.begin();  par2 = (ivert->particles.begin()+1);
      if (par1->Chi2DistTo(par2) > chi2cut || 
	  par1->DistanceTo(par2) > spacecut  ) {
	jpsis->erase(ivert--);
      }
    }
  }
  if (verbose) cout << "of which " << jpsis->size() << " pass" << endl;



  // If we don't even have candidates, forget it.
  if (jpsis->empty()) {
    if (verbose) cout << "*** GroverJpsiSec: FAILED - "
      "no J/psi candidates pass cuts" << endl;
    return FAILURE;
  }
  else { // label them and fit them (keeping original momenta!)
    int label = 0;
    for (GroverVertexList::iterator isec = jpsis->begin(); 
	 isec != jpsis->end(); ++isec) {
      isec->SetIndex(++label);
      isec->ResetParticles();
      GroverConstrainedFit(isec,-1.0);
    }
  }




  if(verbose) cout << "making seed list" << endl;

  // Make a list of high-ish IPS tracks which are not in primaries and which
  // could be added to the J/psi candidates
  float seedips[ASIZE] = {0.0};  
  GroverParticleList::iterator seedptr[ASIZE] = {0};
  GroverVertexList::iterator  seedprim[ASIZE] = {0};
  int nseeds = 0;
  static float ipscut = cuts.RSMinSeedIPS;
  float ips=0.0, ipsbest = 1e9; 
  bool d2prim = false;
  GroverVertexList::iterator primbest = 0;
  //int kk =0;
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar) {
    if (ipar->UsedInPrimary() != 0 || ipar->UsedInSecondary() == 1234) continue;
    ipsbest = 1e9;
    for (GroverVertexList::iterator iprim = allPrimaries->begin();
	 iprim != allPrimaries->end(); ++iprim) {
      ips = ipar->Chi2DistTo(iprim);
      // Keep the chi2 distance to *closest* primary
      if (ips < ipsbest) {
	d2prim = true;
	ipsbest  = ips;
	primbest = iprim;
      }
    }
    if (ipsbest > ipscut && d2prim) { // keep as a seed track
      seedptr[nseeds] = ipar; 
      seedips[nseeds] = ipsbest;
      seedprim[nseeds]= primbest;
      nseeds++;
    }
  }
  // Make new list of seed tracks
  // then sort in descending order of IPS
  if (verbose) cout << "sorting the " << nseeds << " seeds..." << endl;
  static GroverParticleList SeedTracks;
  SeedTracks.clear();
  //int nsec = 0;
  for (int i=0; i<nseeds; i++) {
    seedptr[i]->SetSecondaryWeight(seedips[i]);
    seedptr[i]->SetMass(m_pi);
    SeedTracks.push_back(*(seedptr[i]));
  }
  sort(SeedTracks.begin(), SeedTracks.end(), SortGPBySecWeight);





  // some miscellaneous variables and pointers
  //int i=0;
  //int lep1type=0;
  //int lep2type=0;
  //int othertype=0;                  // GP types
  //float minchi2_vvsep = cuts.RSVtxVtxMinChi2;// cut for separating 2s from 1s
  //bool isec_dead = false;                    // don't kill the secondary yet




  // Loop over seeds until all track addition possibilities 
  // have been exhausted
  float chi2contribcut   = cuts.RSMaxChi2Contrib;
  float vtx_match_cut    = cuts.RSMaxTrkVtxChi2;
  bool keep_looking = true;
  if (verbose) cout << "starting infinite loop" << endl;
  while (keep_looking) {
    keep_looking = false; // default unless this round is successful
    // loop over unused tracks
    int acounter=0;
    for (GroverParticleList::iterator itrk = SeedTracks.begin();
	 itrk != SeedTracks.end(); ++itrk) {
      if (itrk->UsedInSecondary() != 0) continue;
      if (verbose && false) cout << "checking seed track number "<< ++acounter<< endl;
      bool success = false;
      float best_dist = 1e10; GroverVertexList::iterator best_vtx = 0;
      int acounter2=0;
      for (GroverVertexList::iterator isec = jpsis->begin();
	   isec != jpsis->end(); ++isec) {
	if (verbose && false) cout << "...against jpsi number " << ++acounter2 << endl;
	// find the best matched secondary to this track
	float dist = itrk->Chi2DistTo(isec);
	if (dist < best_dist) {
	  best_dist = dist;
	  best_vtx  = isec;
	}
      }
      if (best_dist < vtx_match_cut) {
	best_vtx->Add(itrk);
	best_vtx->ResetParticles();
	GroverConstrainedFit(best_vtx,0.0);
	if (best_vtx->particles.back().Chi2() < chi2contribcut) {
	  if (verbose && false) cout << "SUCCESS! chi2 = " 
				     << best_vtx->particles.back().Chi2() << endl;
	  itrk->SetUsedInSecondary(best_vtx->Index());
	  success = true;
	  keep_looking = true;
	}
	else {
	  if (verbose && false) cout << "FAILURE! chi2 = " 
				     << best_vtx->particles.back().Chi2() << endl;
	  // reset everything
	  best_vtx->particles.pop_back();
	  best_vtx->ResetParticles();
	  GroverConstrainedFit(best_vtx,0.0);
	  success = false;
	}
      }
    } // looped over all unused tracks...
  }// end of infinite loop
  if (verbose) cout << "end of infinite loop" << endl;
  /*
    if (verbose) cout << "ADDED TRACKS" << endl;
    // Deal with higher multiplicity vertices
    if (!isec_dead) {
      if (isec->particles.size() > 2) {
	// Project all tracks to the z position of the J/psi
	float zpsi = isec->Z();
	for (GroverParticleList::iterator ipar = isec->particles.begin();
	     ipar != isec->particles.end(); ++ipar) ipar->PropagateBeginTo(zpsi);
	// Estimate covariance matrix in plane of J/psi
	isec->EstimateCov(1.0, 0.01);
	// sort particles in descending order of distance to J/psi vtx
	sort(isec->particles.begin(), isec->particles.end(), RevSortGPByDist);
	// Fit again
	if (verbose) if (true) 
	  cout << "3) Vertex has trks = " << jpsis->begin()->Size() << endl;
	isec->ResetParticles();
	int error = GroverConstrainedFit(isec,3.1);
	if (verbose) if (true) 
	  cout << "4) Vertex has trks = " << jpsis->begin()->Size() << endl;
	// mark (un)used particles
	for (GroverParticleList::iterator ipar = isec->particles.begin();
	     ipar != isec->particles.end(); ++ipar)
	  ipar->SetUsedInSecondary(isec->Index());
	for (GroverParticleList::iterator ipar = isec->trashcan.begin();
	     ipar != isec->trashcan.end(); ++ipar)
	  ipar->SetUsedInSecondary(0);
	// skip if very close to primary
	dist = isec->Chi2DistTo(closestprim);
	if (dist < minchi2_vvsep ) {
	  jpsis->erase(isec--);
	  isec_dead = true;
	}
	
	if (verbose) cout << "testing for primary refit" << endl;
	
	// Primary Refitting if required
	if (false && !isec_dead) { // <== NOTE !!!
	  int kkk=0;
	  if (verbose) cout << "looping over sec particles" << endl;
	  for (GroverParticleList::iterator i2par = isec->particles.begin();
	       i2par != isec->particles.end(); ++i2par) {
	    if (verbose) cout << "getting index" << endl;
	    int prim = i2par->UsedInPrimary();
	    if (verbose) cout << "particle #" << ++kkk << " is in prim " << prim << endl;
	    if (prim > 0) {
	      GroverVertexList::iterator pprim = &((*primaries)[prim-1]);
	      for (GroverParticleList::iterator i1par = 
		     pprim->particles.begin();
		   i1par != pprim->particles.end(); ++i1par) {
		if (verbose) cout << "testing rseg" << endl;
		if (i1par->ParentRSEG() == i2par->ParentRSEG()) {
		  if (verbose) cout << "refitting primary" << endl;
		  pprim->trashcan.push_back(*i1par);
		  pprim->particles.erase(i1par--);
		  pprim->ResetParticles();
		  GeoFit(pprim);
		}
	      }
	      if (verbose) cout << "done loop over particles" << endl;
	    }
	  }
	  // now we've removed the secopndry tracks from the primary
	  //GeoFit(

	  // and a final check of compatibility with primary
	  dist = isec->DistanceTo(closestprim);
	  if (dist < minchi2_vvsep && false) {
	    for (GroverParticleList::iterator itrash = isec->trashcan.begin();
		 itrash != isec->trashcan.end(); ++itrash) {
	      closestprim->particles.push_back(*itrash);
	      jpsis->erase(isec--);
	      isec_dead = true;
	    }
	  }
	}
	
      } // done all higher multiplicity vertices
  */


  // done
  if (verbose) cout << "*** GroverJpsiSec: finished!" << endl;
  return SUCCESS;
  
}
