#include <iostream>
#include <algorithm>
#include <math.h>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/higz.h"
#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -------------------------------------------------------------------------
// 00-05-14  Max      First Version   
// 00-05-24  Max      Corrected bug in target alignment
// 00-09-29  Max      Set 'used' particle flags = 0 if <0 at end of function
// 00-10-28  Max      Target following? Try to remove arte dependence soon...
// 00-12-06  Max      More robust wire following mechanism
// 01-01-18  Max      GeoFit(vtx*) => vtx.GeoFit() (reverses change
//                    of May 16 2000)
// 01-02-28  Max      GroverSecondaryFit(vtx*) REMOVED (completely obsolete)
// 01-02-28  Max      Maximum number of vertices set to 32 (no limit previously)
// 01-03-13  Max      Added preprocessor directives
// 01-04-24  Michael  overite sx, sy, sz in allWires list
//
// 02-05-30  DE       The tolerance in WFM is increased

// Wire positions found by vertexing
float wire_xpos[9]  = {0.0},
      wire_ypos[9]  = {0.0},
      wire_zpos[9]  = {0.0},
      wire_xsize[9] = {0.0},
      wire_ysize[9] = {0.0},
      wire_zsize[9] = {0.0};

// Arrays with vertex-rseg relations

const int ROVERMAXTRACKS = 1000;

int grover_rseg_to_rver[ROVERMAXTRACKS] = {0};




int GroverPrimaries(GroverParticleList* allParticles,
		    GroverWireList*     allWires,
		    const GroverCuts*   cuts,
		    GroverVertexList*   allVertices) 
{
  ////////////////////////////////////////////////////////////////////
  //
  //                     GroverPrimaries
  //                               
  // Finds the primary vertices. Algorithm very similar to Rover 1.07.
  //
  // Authors: Dmitry Emilianov, Ivan Kisel, Max Sang, Yuri Vassiliev
  //
  // Inputs (all pointers): A particle list (non-const!)
  //                        A wire list (non-const!)
  //                        A set of cuts 
  // Output: A list of vertices appended to *allVertices
  //         Note that the lists of particles and wires 
  //         will also be changed during execution! The particles
  //         will be labelled with the index of the primary to which
  //         they belong, and will be in local VDS coordinates. The
  //         wires will also be in local coordinates and will be
  //         labelled with the number of tracks and wires found 
  //         on them in the last event. No more than MAXVERTICES
  //         permitted, to save CPU time on the farm.
  //
  // Everything is done in VDS local coordinates!
  // If you pass any arguments in global coordinates they will be
  // changed automatically! BEWARE!
  //
  /////////////////////////////////////////////////////////////////////


  // number of tracks and vertices on a new wire
  // (i.e. some inertia for the wire following)
  const int WRNPAR  = 100;
  const int WRNPRIM = 100;
  // maximum number of tracks+vertices on a wire before resetting
  const int WRMAX = 10000;
  // array size
  const int MAXTRKS = 200;
  // minimum number of hits in a vetex track
  const int MINHITSINTRACK = 5;
  // maximum number of vertices before terminating the search
  const int MAXVERTICES = 32;

  // Get cuts
  bool  verbose           = (cuts->Verbose == 1);
  bool  groverdraw        = (cuts->GroverDraw == 1);
  int   mintracksinvertex = cuts->RPMinTrksInVtx;
  float wtsig             = cuts->RPTrkWireSigmaMax; 


  // Counters for total number of calls and calls/event 
  static int total_ncalls = 0;
  ++total_ncalls;
  bool firstcall = (total_ncalls == 1);
  static int oldrun=0, oldevt=0;
  static int ncalls_thisevt = 1;
  bool duplicall = false;  


#ifndef GROVER_WITHOUT_ARTE
  ArtePointer<EVHD> evhd = *(ArteTable<EVHD>::begin());
  if (firstcall) {
    oldrun = evhd->run; oldevt = evhd->evt; 
    duplicall = false;
    ncalls_thisevt = 1;
  }
  else {

    if (evhd->run == oldrun && evhd->evt==oldevt) {
      duplicall = true; // multiple calls!
      ++ncalls_thisevt;
    }
    else { // first call on this evt
      duplicall = false;
      ncalls_thisevt = 1;
    }
    oldrun = evhd->run; oldevt = evhd->evt; // for next time
  }
#endif



  // welcome message and graphics initialisation
  static bool higz_open = false;
  if (firstcall) {
    cout << "*** GroverPrimaries (Grover v." << GroverVersion 
	 << "): first call" << endl;
    if (groverdraw && !higz_open) {
      graph_ini();
      higz_open = true;
    }
  }



  int nver = 0; // number of vertices found in this event
  GroverParticleList::iterator trackonwire[9][MAXTRKS]; // pointers to particles
  GroverParticleList::iterator trackincluster[MAXTRKS]; //  -"-
  float  trackweight[9][MAXTRKS];// weights for associating tracks to wires
  int    ntronwire[9];           // number of tracks on wire [i]

 


  // wire following stuff
  static GroverWireList localWires;
  GroverWireList* wires = 0;
  bool newwires      = (cuts->GPNewWires == 1);
  bool wirefollowing = (cuts->GPWireFollowing == 1);
  if  (wirefollowing)  wires = &localWires;
  else wires = allWires; // just ignore internal list



  // Transform input wires to local coordinates
  MoveToLocal(wires);

  


  // Get best estimates of wire positions if wire following
  // is switched on
  if (wirefollowing) {
    if (newwires && !duplicall) {
      // Replace all wires if requested
      if (verbose) cout << "*** GroverPrimaries: re-reading wires"<<endl;
      localWires.clear();
      localWires = *allWires;
      if (verbose) cout << "*** GroverPrimaries: found " 
			<< localWires.size() << " wires" <<endl;
    }
    else {
      // OK - no new wires - but use the external positions for any which
      // have not yet been seen by the vertexing OR which have been
      // seen but have mysteriously stopped being seen (e.g. emergency
      // wire retraction)
      if (verbose) cout << "*** GroverPrimaries: checking wires for 'quiets'"<<endl;
      for (GroverWireList::iterator iint = localWires.begin();
	   iint != localWires.end(); ++iint) {
	int wnrint = iint->Number();
	if (wnrint <1 || wnrint >8) continue;
	if (iint->NParticles() <= WRNPAR  && 
	    iint->NPrimaries() <= WRNPRIM   ) {
	  for (GroverWireList::iterator iext = allWires->begin();
	       iext != allWires->end(); ++iext) {
	    if (iext->Number() == wnrint) {
	      iint->SetX(iext->X());
	      iint->SetY(iext->Y());
	      iint->SetZ(iext->Z());
	    }
	  }
	}
      }
    }
  }



  // Draw the target wires
  if (groverdraw) {
    if (!higz_open) { 
      graph_ini();
      higz_open = true;
    }
    draw_targets(wires);
  }



  // If MC data, draw the MC vertices. Also draw the MC 
  // tracks where they intersect with zplane = whatever.
  float zplane = -1.0;
  bool MC = false;

#ifndef GROVER_WITHOUT_ARTE
  MC = ArteTable<MCIN>::booked();
#endif

  if (groverdraw && MC) {
    if (!higz_open) {
      graph_ini();
      higz_open = true;
    }
    draw_mc(zplane);
  }



  // Label all the particles with an index (starting at 1!)
  // In fact, as long as you never removed anything from 
  // *allParticles, you could use e.g. (*allParticles)[i].Momentum()
  // for indexed access to the particles in the list
  // (where i starts at 0, not at 1!)
  // Also check they're all in local coordinates
  int idx = 0;
  MoveToLocal(allParticles);
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar) ipar->SetIndex(++idx);
  if (verbose)  cout << "*** GroverPrimaries: Number of particles = " 
		     << idx << endl; 


  // Calculate new target position using only tracks
  if (wirefollowing) {
    float dx  = 0.4, dy  = 0.4; // tracks > 1mm away are not considered
    float dxt = 0.0, dyt = 0.0, xt=0.0, yt=0.0, zt=0.0;
    float x=0.0, y=0.0;
    int   nt_tr=0, nt_vx=0;
    for (GroverWireList::iterator iwire = wires->begin();
	 iwire != wires->end(); ++iwire) {
      // get current position of wire
      xt = iwire->X();   yt = iwire->Y();   zt = iwire->Z();
      dxt = iwire->SizeX();  dyt = iwire->SizeY();
      nt_tr = iwire->NParticles();  nt_vx = iwire->NPrimaries();
      // -> calculate new (weighted) position
      for (GroverParticleList::iterator ipar = allParticles->begin();
	   ipar != allParticles->end(); ++ipar) {
	ipar->PropagateBeginTo(iwire->Z());
	x = ipar->BeginX();   y = ipar->BeginY();
	if ((fabs(x-xt)<dx) && (fabs(y-yt)<dy)) {
	  xt = xt*(nt_tr+nt_vx)+x;
	  yt = yt*(nt_tr+nt_vx)+y;
	  ++nt_tr;
	  xt/=(nt_tr+nt_vx);  
	  yt/=(nt_tr+nt_vx);
	  dxt = (dxt*(nt_tr-1+nt_vx) + fabs(x-xt)) / (nt_tr + nt_vx);
	  dyt = (dyt*(nt_tr-1+nt_vx) + fabs(y-yt)) / (nt_tr + nt_vx);
	}
      }
      // -> set new position
      iwire->SetX(xt);
      iwire->SetY(yt);
      iwire->SetSizeX(dxt);
      iwire->SetSizeY(dyt);
      // after 10,000 tracks or vertices -> loosen up, man! The target
      // moves slowly, so don't force it to stay in the same place for 
      // ever.
      if (nt_tr+nt_vx > WRMAX)  { nt_tr=WRNPAR+1;  nt_vx = WRNPRIM+1;} 
      iwire->SetNParticles(nt_tr);
      iwire->SetNPrimaries(nt_vx);
    }
  }



  // Associate tracks to wires
  int wnr, ntr;
  for(GroverWireList::iterator iwire = wires->begin();
      iwire != wires->end(); ++iwire) {
    wnr = iwire->Number();
    ntr = 0;
    for(GroverParticleList::iterator ipart = allParticles->begin();
	ipart != allParticles->end(); ++ipart) {
      //ipart->PropagateBeginTo(iwire->Z()); // already inside Chi2DistTo(iwire) !!!
      if(ipart->Chi2DistTo(iwire) <= (wtsig*wtsig)){
	++ntr;
	trackonwire[wnr][ntr] = ipart;
	trackweight[wnr][ntr] = 0.0;
      }
    }
    ntronwire[wnr] = ntr;
    if (verbose) cout << "*** GroverPrimaries: " << ntr 
      << " particles on wire " << iwire->Number() << endl;
  } 



  // just some variables
  int ntrincluster = 0;
  GroverParticleList::iterator  it_best, it, jt;
  GroverWireList::iterator  wire_best;
  float w_best = 0.;
  int max_clu  = 0;
  int tracksonwire;
  int itrnew = 0;
  wnr = 0;


  // Search for clusters to make vertices
  while(true) { // i.e. while there is at least one seed with >=3 tracks

    it_best     = allParticles->begin();
    w_best      = 0.0;
    max_clu     = 0;
    wire_best   = 0;

    for(GroverWireList::iterator iwire = wires->begin();
	iwire != wires->end(); ++iwire) {

      // Propagate all the tracks on this wire to its z plane
      wnr = iwire->Number();
      tracksonwire = ntronwire[wnr];
      if(tracksonwire < mintracksinvertex) continue; // skip to next wire
      itrnew = 0;
      for (int itr=1; itr <= tracksonwire; ++itr){ 
	it = trackonwire[wnr][itr];
	if (it->UsedInPrimary() != 0) continue; // skip used tracks
	++itrnew;
	trackonwire[wnr][itrnew] = trackonwire[wnr][itr];
	trackweight[wnr][itrnew] = trackweight[wnr][itr];
	it->PropagateBeginTo(iwire->Z());
      }
      tracksonwire = ntronwire[wnr] = itrnew;
      if(tracksonwire < mintracksinvertex) // no three-track clusters left
	continue;                          // on this wire -> skip to next.


      // Find clusters (vertex seeds)
      float x1, y1, x2, y2, weight; 
      int ntr_clu = 0;
      for (int itr=1; itr <= tracksonwire; ++itr){
	it = trackonwire[wnr][itr];
	x1 = it->BeginX();
	y1 = it->BeginY();
	ntr_clu = 0;
	weight  = 0.0;
	for (int jtr=1; jtr <= tracksonwire; ++jtr){ // <- is this correct?
	  jt = trackonwire[wnr][jtr];                // yes, it gives minimum 
	  x2 = jt->BeginX();                         // weight = 1, that 
	  y2 = jt->BeginY();                         // characterises number
          float dist2=(x1-x2)*(x1-x2)+(y1-y2)*(y1-y2);// of tracks in
          float cut2 = wtsig*wtsig*                  // a cluster
            ( it->BeginCov(0)+it->BeginCov(2)        
             +jt->BeginCov(0)+jt->BeginCov(2));       
          if(dist2 < cut2){
            weight += ((1.0-(dist2/cut2))*(1.0-(dist2/cut2)));
            ++ntr_clu;
          }
	}
	it->SetPrimaryWeight(weight);
	if (weight > w_best){ // keep best vtx candidate
	  w_best  = weight;   // on this wire
	  it_best = it; 
	  max_clu = ntr_clu; 
	  wire_best = iwire;
	}
      }


    } // finished loop over wires


    if (w_best == 0.0) break; // no vtx candidates left - leave the inf. loop.


    // propagate tracks in the best vtx seed to the appropriate wire
    wnr = wire_best->Number();
    tracksonwire = ntronwire[wnr];
    for (int itr=1; itr <= tracksonwire; ++itr){
      it = trackonwire[wnr][itr];
      it->PropagateBeginTo(wire_best->Z());
    }
    
    // calculate track weights for this (closest) wire
    float x1, y1, x2, y2, weight;
    for (int itr=1; itr <= tracksonwire; ++itr){
      it = trackonwire[wnr][itr];
      x1 = it->BeginX();
      y1 = it->BeginY();
      weight = 0.0;
      for (int jtr=1; jtr <= tracksonwire; ++jtr) {
	jt = trackonwire[wnr][jtr];
	x2 = jt->BeginX();
	y2 = jt->BeginY();
        float dist2 = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
        float cut2 = wtsig*wtsig*
          (it->BeginCov(0) + it->BeginCov(2) +
           jt->BeginCov(0) + jt->BeginCov(2) );
        if(dist2 < cut2)
          weight += ((1.0-(dist2/cut2))*
		     (1.0-(dist2/cut2)) );
      }
      it->SetPrimaryWeight(weight);
    }


    float x_best  = it_best->BeginX();
    float y_best  = it_best->BeginY();
    float c0_best = it_best->BeginCov(0);
    float c2_best = it_best->BeginCov(2);
    float w,x,y;
    float vertx = 0.0, verty = 0.0, vertweight = 0.0;
    ntrincluster = 0;
    for (int itr=1; itr <= tracksonwire; ++itr){ // collect cluster
      it = trackonwire[wnr][itr];
      x = it->BeginX();
      y = it->BeginY();
      w = it->PrimaryWeight();
      float dist2 = (x-x_best)*(x-x_best) + (y-y_best)*(y-y_best);
      float cut2 = wtsig*wtsig*
	(it->BeginCov(0) + it->BeginCov(2) + c0_best + c2_best);
      if(dist2 < cut2){
	vertx += w*x;
	verty += w*y;
	vertweight += w;
	trackincluster[++ntrincluster] = it;
      }
    }

    if (ntrincluster < mintracksinvertex)  break;


    // Make a new GroverVertex in the wire plane and 
    // fill with the candidate tracks
    // Convenient notation but local coordinates ASSUMED!
    
    vertx /= vertweight;  verty /= vertweight;
    GroverVertex vertex(vertx, verty, wire_best->Z());
    vertex.SetWire(wire_best->Number());
    for (int i=1; i<=ntrincluster; ++i) vertex.Add(trackincluster[i]);
    
    vertex.EstimateCov(1.0, 0.01);
    
    // Calculate distances from tracks to vertex in xy plane
    vertex.UpdateDistances();
        
    // Sort particle list according to distance from 
    // mean vertex position
    //sort(vertex.particles.begin(), vertex.particles.end(), GroverDistSort);
    
    // FIT VERTEX and see if it's still good
    int vertex_ok = 0;
    vertex.GeoFit();
    int refsize = 0;
    for (GroverParticleList::iterator ipart = vertex.particles.begin();
	 ipart != vertex.particles.end(); ++ipart) {
      if (ipart->Nhit() >= MINHITSINTRACK) ++refsize;
    }

    
    if (refsize >= mintracksinvertex)	{
      // We found a vertex!!!
      vertex_ok = 1;
      ++nver;
      vertex.SetIndex(nver); // give it a number
      allVertices->push_back(vertex); // add it to the global vertex list
    }
    
    
    if (vertex_ok == 1) {
      if (verbose) cout << "*** GroverPrimaries: found good vertex" << endl;
      // Flag the used tracks as 'used' in the global particle list
      // by labelling them with the vertex number
      for (GroverParticleList::iterator ipar = vertex.particles.begin();
	   ipar != vertex.particles.end(); ++ipar) {
	if (ipar->UsedInPrimary() != 0) { // not discarded from vertex
	  int id = ipar->Index();
	  (*allParticles)[id-1].SetUsedInPrimary(nver);
	  if (id>=0&&id<ROVERMAXTRACKS) grover_rseg_to_rver[id] = nver;
	  //ipar->PropagateBeginTo(vertex.Z()+0.5);
	  if (groverdraw) {
	    if (!higz_open) {
	      graph_ini();
	      higz_open = true;
	    }
	    draw_track(ipar,Black); //max//
	  }
	}
      }
      if (groverdraw) {
	if (!higz_open) {
	  graph_ini();
	  higz_open = true;
	}
	draw_vertex(&vertex,Red);
      }
    }
    else {
      // If it's a bad vertex, also flag the _unused_ tracks as 'used'.
      for (GroverParticleList::iterator ipar = vertex.particles.begin();
	   ipar != vertex.particles.end(); ++ipar) {
	if (ipar->UsedInPrimary() != 0)
	  (*allParticles)[(ipar->Index())-1].SetUsedInPrimary(-1);
      }
      if (!vertex.trashcan.empty()) {
	for (GroverParticleList::iterator ipar = vertex.trashcan.begin();
	     ipar != vertex.trashcan.end(); ++ipar) {
	  (*allParticles)[(ipar->Index())-1].SetUsedInPrimary(-1);
	}
      }
    }
    

    if (vertex_ok == 1 && wirefollowing) {
      // Update target position using tracks and vertices
      // -> get current position of wire
      float xt    = wire_best->X();
      float yt    = wire_best->Y();
      float zt    = wire_best->Z();
      float dxt   = wire_best->SizeX();
      float dyt   = wire_best->SizeY();
      float dzt   = wire_best->SizeZ();
      int   nt_tr = wire_best->NParticles();
      int   nt_vx = wire_best->NPrimaries();
      int   nv    = vertex.Size();
      float xv    = vertex.X();
      float yv    = vertex.Y();
      float zv    = vertex.Z();
      // -> calculate new position
      float dx = 0.1, dy = 0.1;
      if ((fabs(xv-xt)<dx)&&(fabs(yv-yt)<dy)){
	xt = xt*(nt_tr+nt_vx)+xv*3*nv;
	yt = yt*(nt_tr+nt_vx)+yv*3*nv;
	zt = zt*nt_vx+zv*3*nv;
	nt_vx += 3*nv;
	xt /= (nt_tr+nt_vx);  
	yt /= (nt_tr+nt_vx);
	zt /=  nt_vx;

	dxt = (dxt*(nt_tr+nt_vx-3*nv) + nv*fabs(xv-xt)) / (nt_tr + nt_vx);
	dyt = (dyt*(nt_tr+nt_vx-3*nv) + nv*fabs(yv-yt)) / (nt_tr + nt_vx);
	dzt = (dzt*(nt_vx-3*nv) + 3*nv*fabs(zv-zt)) / nt_vx;
      }
      // -> set new position
      wire_best->SetX(xt);
      wire_best->SetY(yt);
      wire_best->SetZ(zt);
      wire_best->SetSizeX(dxt);
      wire_best->SetSizeY(dyt);
      wire_best->SetSizeZ(dzt);   
      // -> Update global wire position monitoring
      int thiswire = wire_best->Number();
      wire_xpos[thiswire]  = xt;
      wire_ypos[thiswire]  = yt;
      wire_zpos[thiswire]  = zt;
      wire_xsize[thiswire] = dxt;
      wire_ysize[thiswire] = dyt;
      wire_zsize[thiswire] = dzt;
      // after 10k tracks or vertices -> loosen up, man! The target
      // moves slowly, so don't force it to stay in the same place for 
      // ever.
      if (nt_tr+nt_vx > WRMAX)  { nt_tr=WRNPAR+1;  nt_vx = WRNPRIM+1;} 
      wire_best->SetNParticles(nt_tr);
      wire_best->SetNPrimaries(nt_vx);

      if (!duplicall && !newwires) {
	// Overwrite positions of external wires with local best
	// estimates
	for (GroverWireList::iterator iext = allWires->begin();
	     iext != allWires->end(); ++iext) {
	  if (iext->Number() == thiswire) {
	    iext->SetX(xt);
	    iext->SetY(yt);
	    iext->SetZ(zt);
	    iext->SetSX(dxt);
	    iext->SetSY(dyt);
	    iext->SetSZ(dzt);
	    iext->SetNPrimaries(nt_vx);
	    iext->SetNParticles(nt_tr);
	  }
	}
      }
    }

    if (nver >= MAXVERTICES) break; // safety

  } // end of while(true)


  // Clean up 'used' flag (Max, 29 Sep 00)
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar)
    if (ipar->UsedInPrimary() < 0) ipar->SetUsedInPrimary(0);
  

  if (verbose) cout << "*** GroverPrimaries: found " << nver << " primary "
		    << ((nver==1) ? "vertex" : "vertices") << endl;
  

  return SUCCESS;
}



////////////////////////////////////////////////////////////////////



int GeoFit(GroverVertex* vtx)
{
  return vtx->GeoFit();
}






///////////////////////////////////////////////////////////////////////////


