#ifndef GROVER_INLINES
#define GROVER_INLINES

//
// This file contains inlined member functions of the Grover classes.
// see grover.hh for context and GroverParticle.C, GroverVertex.C
// and GroverWire.C for the rest of the member functions
//
// Please remember that inlining is a trade-off. In principle it allows
// the compiler to substitute real code instead of a function call.
// However: a) the compiler might ignore the inline specifier
//          b) It can lead to 'code bloat' (object files get fat)
// If you are responsible for the code and you hate inlines, move these
// functions to their respective class definitions (GroverParticle.C, 
// for example) and remove the inline specifier (essential). However, 
// you should check it doesn't slow any of the routines down...
//
// Created January 19 2001 by M.Sang
//
// Modification Log:
//
// (YY-MM-DD) Person Modification(s) 
//  01-03-13  Max    Added preprocessor directives
//  01-04-24  Michael Added functions to access sx,sy,sz in GroverWire
//





/////////////////////////////////////////////////////////////////////
//                                                                 //
//                            GroverParticle                       //
//                                                                 //
/////////////////////////////////////////////////////////////////////

//            ***  public ***

inline bool  GroverParticle::InLocal(void)  const 
{return inlocal;}

inline bool  GroverParticle::InGlobal(void) const 
{return inglobal;}

inline float GroverParticle::Momentum(void) const 
{return p;}

inline float GroverParticle::InvMomentum(void) const 
{return (p!=0.0) ? 1.0/p : 0.0;}

inline int GroverParticle::SetMomentum(const float p_new)
{if(p_new>0) {p=p_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline int GroverParticle::SetInvMomentum(const float invp_new)
{if(invp_new>0) {p=1./invp_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline int GroverParticle::SetOMomentum(const float p_new)
{if(p_new>0) {op=p_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline int GroverParticle::SetOInvMomentum(const float invp_new)
{if(invp_new>0) {op=1./invp_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline float GroverParticle::BeginX(void) const 
{return xb;}

inline float GroverParticle::BeginY(void) const 
{return yb;}

inline float GroverParticle::BeginZ(void) const 
{return zb;}

inline float GroverParticle::BeginSlopeX(void) const 
{return axb;}

inline float GroverParticle::BeginSlopeY(void) const 
{return ayb;}

inline float GroverParticle::BeginCov(const int i) const
{if (i>=0&&i<=20) return covb[i]; else return FAILURE;}

inline float GroverParticle::EndX(void) const 
{return xe;}

inline float GroverParticle::EndY(void) const 
{return ye;}

inline float GroverParticle::EndZ(void) const 
{return ze;}

inline float GroverParticle::EndSlopeX(void) const 
{return axe;}

inline float GroverParticle::EndSlopeY(void) const 
{return aye;}

inline float GroverParticle::EndCov(const int i) const 
{if (i>=0&&i<=20) return cove[i]; else return FAILURE;}

inline float GroverParticle::OriginalBeginZ(void) const 
{return ozb;}

inline float GroverParticle::OriginalEndZ(void) const 
{return oze;}

inline int GroverParticle::Nhit(void) const 
{return nhit;}

inline int GroverParticle::SetNhit(const int newnhit) 
{nhit=newnhit; return SUCCESS;}

inline int GroverParticle::Charge(void) const 
{return q;}

inline int GroverParticle::SetCharge(const int newq) 
{q=newq; return SUCCESS;}

inline float GroverParticle::Chi2(void) const
{return chi2;}

inline int GroverParticle::SetChi2(const float c) 
{chi2 = (c>=0.0) ? c:0.0; return SUCCESS;}

inline float GroverParticle::Mass(void) const 
{return mass;}

inline int GroverParticle::SetMass(const float mass_new)
{if(mass_new>=0) {mass=mass_new; return SUCCESS;} else return FAILURE;}

#ifndef GROVER_WITHOUT_ARTE
inline int GroverParticle::SetTopp(const ArtePointer<TOPP> t)
{return this->SetMass(t);}

inline ArtePointer<TOPP> GroverParticle::Topp(void) const 
{return topp;}

inline ArtePointer<RTRA> GroverParticle::ParentRTRA(void) const 
{return parRTRA;}

inline ArtePointer<RSEG> GroverParticle::ParentRSEG(void) const 
{return parRSEG;}

inline ArtePointer<RCCL> GroverParticle::ParentRCCL(void) const 
{return parRCCL;}

inline ArtePointer<MTRA> GroverParticle::ParentMTRA(void) const 
{return parMTRA;}
#endif

inline int GroverParticle::Type(void) const 
{return type;}

inline GroverParticle* GroverParticle::ParentGP(void)   const 
{return parGP;}

inline GroverParticleList* GroverParticle::ParentGPL(void)  const 
{return parGPL;}

inline GroverVertex* GroverParticle::ParentGV(void)   const 
{return parGV;}

inline int GroverParticle::Index(void) const 
{return index;}

inline int GroverParticle::SetIndex(const int i) 
{index=abs(i); return (i>=0) ? 0 : -1;}

inline float GroverParticle::Dist(void) const 
{return dist;}

inline int GroverParticle::SetDist(const float d) 
{dist=fabs(d); return SUCCESS;}

inline int GroverParticle::UsedInPrimary(void) const
{return usedinprimary;}

inline int GroverParticle::SetUsedInPrimary(const int m) 
{usedinprimary = m; return SUCCESS;}

inline int GroverParticle::UsedInSecondary(void) const 
{return usedinsecondary;}

inline int GroverParticle::SetUsedInSecondary(const int m) 
{usedinsecondary = m; return SUCCESS;}

inline float GroverParticle::PrimaryWeight(void) const 
{return wprim;}

inline int GroverParticle::SetPrimaryWeight(const float wprim_new) 
{wprim = wprim_new; return SUCCESS;}

inline float GroverParticle::SecondaryWeight(void) const 
{return wsec;}

inline int GroverParticle::SetSecondaryWeight(const float wsec_new) 
{wsec = wsec_new; return SUCCESS;}

//            ***  protected ***

inline int GroverParticle::SetGlobal(const bool newg) 
{inglobal=newg; tampered=true; return SUCCESS;}

inline int GroverParticle::SetLocal(const bool newl) 
{inlocal=newl; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginX(const float xb_new)
{xb = xb_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginY(const float yb_new)
{yb = yb_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginZ(const float zb_new)
{zb = zb_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndX(const float xe_new)
{xe = xe_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndY(const float ye_new)
{ye = ye_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndZ(const float ze_new)
{ze = ze_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginSlopeX(const float axb_new)
{axb = axb_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginSlopeY(const float ayb_new)
{ayb = ayb_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndSlopeX(const float axe_new)
{axe = axe_new; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndSlopeY(const float aye_new)
{aye = aye_new; tampered=true; return SUCCESS;}

inline int GroverParticle:: SetBeginCov(const float new_covb[COVDIM])
{for(int i=0;i<COVDIM;i++) covb[i]=new_covb[i]; tampered=true; return SUCCESS;}

inline int GroverParticle::SetBeginCov(const int index, const float covb_new)
{if(index>=0&&index<=20) {covb[index]=covb_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline int GroverParticle::SetEndCov(const float new_cove[COVDIM])
{for(int i=0;i<COVDIM;i++) cove[i]=new_cove[i]; tampered=true; return SUCCESS;}

inline int GroverParticle::SetEndCov(const int index, const float cove_new)
{if(index>=0&&index<=20) {cove[index]=cove_new; tampered=true; return SUCCESS;} else return FAILURE;}

inline bool GroverParticle::Tampered(void) const
{return tampered;}



/////////////////////////////////////////////////////////////////////
//                                                                 //
//                            GroverVertex                         //
//                                                                 //
/////////////////////////////////////////////////////////////////////

//            ***  public ***

inline bool GroverVertex::InLocal(void) const  
{return inlocal;}

inline bool  GroverVertex::InGlobal(void) const 
{return inglobal;}

inline float GroverVertex::Chi2(void) const 
{return chi2;}

inline float GroverVertex::X(void) const 
{return x;}

inline float GroverVertex::Y(void) const 
{return y;}

inline float GroverVertex::Z(void) const 
{return z;}

inline float  GroverVertex::Cov(const int i) const 
{if(i>=0&&i<COVDIM) return cov[i]; else return FAILURE;}

inline float GroverVertex::Mass(void) const
{return m;} 

inline float GroverVertex::MassError(void) const 
{return dm;}

inline int GroverVertex::Nconstraints(void) const 
{return n_constraints;}

inline float GroverVertex::ConstraintMass(const int i) const
{return (i>=0&&i<CONSTDIM) ? constr_mass[i] : 0.0;}

#ifndef GROVER_WITHOUT_ARTE
inline ArtePointer<TOPP> GroverVertex::ConstraintTOPP(const int i) const
{return (i>=0&&i<CONSTDIM) ? constr_topps[i] : 0;}

inline ArtePointer<TOPP> GroverVertex::ConstraintTOPP(void) const
{return (n_constraints==1) ? constr_topps[0] : 0;}
#endif

inline int GroverVertex::Index(void) const
{return index;}

inline int GroverVertex::SetIndex(const int i) 
{index=abs(i); return (i>=0)?0:-1;}

inline int GroverVertex::Wire(void) const 
{return wnr;}

inline int GroverVertex::SetWire(const int wire) 
{if (wire>0&&wire<9) {wnr=wire; return SUCCESS;} else return FAILURE;}


//            ***  protected ***

inline int GroverVertex::SetChi2(const float newc) 
{chi2=newc; return SUCCESS;}

inline int GroverVertex::SetX(const float xx) 
{x=xx; return SUCCESS;}

inline int GroverVertex::SetY(const float yy) 
{y=yy; return SUCCESS;}

inline int GroverVertex::SetZ(const float zz) 
{z=zz; return SUCCESS;}

inline int GroverVertex::SetMomentum(const float newp) 
{p=fabs(newp);return SUCCESS;}

inline int GroverVertex::SetSlopeX(const float newax) 
{ax=newax;return SUCCESS;}

inline int GroverVertex::SetSlopeY(const float neway) 
{ay=neway;return SUCCESS;}

inline int GroverVertex::SetCov(const float value[COVDIM])
{for (int i=0;i<COVDIM;i++) cov[i] = value[i]; return SUCCESS;}

inline int GroverVertex::SetCov(const int element, const float value)
{if(element>=0&&element<COVDIM){cov[element]=value;return SUCCESS;} else return FAILURE;}

inline int GroverVertex::SetBx(const float bxbx)
{bx=bxbx;return SUCCESS;}

inline int GroverVertex::SetBy(const float byby)
{by=byby;return SUCCESS;}

inline int GroverVertex::SetDecayLength(const float len)
{decay=len; return SUCCESS;}

inline int GroverVertex::SetMass(const float newm)
{m=fabs(newm); return SUCCESS;}

inline int GroverVertex::SetMassError(const float newdm)
{dm=fabs(newdm); return SUCCESS;}

inline float GroverVertex::Bx(void) const 
{return bx;}

inline float GroverVertex::By(void) const
{return by;}

inline float GroverVertex::DecayLength(void) const 
{return decay;}

inline float GroverVertex::GCov(const unsigned int n) const
{if (n>=0&&n<gcov.size()) return gcov[n]; else return 1e9;}




/////////////////////////////////////////////////////////////////////
//                                                                 //
//                             GroverWire                          //
//                                                                 //
/////////////////////////////////////////////////////////////////////

//            ***  public ***

inline int  GroverWire::Number(void) const 
{return wnr;}

inline bool GroverWire::InLocal(void) const  
{return inlocal;}

inline bool GroverWire::InGlobal(void) const 
{return inglobal;}

inline float GroverWire::X(void) const 
{return x;}

inline float GroverWire::Y(void) const 
{return y;}

inline float GroverWire::Z(void) const 
{return z;}

inline float GroverWire::SizeX(void) const 
{return dx;}

inline float GroverWire::SizeY(void) const 
{return dy;}

inline float GroverWire::SizeZ(void) const 
{return dz;}

inline float GroverWire::SX(void) const 
{return sx;}

inline float GroverWire::SY(void) const 
{return sy;}

inline float GroverWire::SZ(void) const 
{return sz;}

inline int GroverWire::Shape(void) const 
{return shape;}

inline bool GroverWire::Horizontal(void) const
{return horizontal;}

inline bool GroverWire::Vertical(void) const 
{return vertical;}

inline int GroverWire::SetX(const float xx) 
{x=xx; return SUCCESS;}

inline int GroverWire::SetY(const float yy) 
{y=yy; return SUCCESS;} 

inline int GroverWire::SetZ(const float zz) 
{z=zz; return SUCCESS;} 

inline int GroverWire::SetSizeX(const float ddx) 
{if (ddx>0){dx=ddx;return SUCCESS;} else return FAILURE;}

inline int GroverWire::SetSizeY(const float ddy) 
{if (ddy>0){dy=ddy;return SUCCESS;} else return FAILURE;}

inline int GroverWire::SetSizeZ(const float ddz) 
{if (ddz>0){dz=ddz;return SUCCESS;} else return FAILURE;}

inline int GroverWire::SetSX(const float xx) 
{sx=xx; return SUCCESS;}

inline int GroverWire::SetSY(const float yy) 
{sy=yy; return SUCCESS;} 

inline int GroverWire::SetSZ(const float zz) 
{sz=zz; return SUCCESS;} 

inline int GroverWire::Mobile(void) const 
{return mobile;}

inline int GroverWire::Stick(void) 
{mobile = 0; return SUCCESS;}

inline int GroverWire::Unstick(void) 
{mobile = 1; return SUCCESS;}

inline int GroverWire::SetMobility(const int m) 
{mobile = m; return SUCCESS;}

inline float GroverWire::Rot(const int i) const 
{if(i>=0&&i<9) return rot[i]; else return FAILURE;}

inline int GroverWire::NParticles(void) const 
{return ntr;}

inline int GroverWire::IncrementNParticles(void) 
{return ++ntr;}

inline int GroverWire::DecrementNParticles(void) 
{return --ntr;}

inline int GroverWire::SetNParticles(const int nn) 
{if (nn>=0) {ntr = nn; return SUCCESS;} else return FAILURE;}

inline int GroverWire::NPrimaries(void) const 
{return nprim;}

inline int GroverWire::IncrementNPrimaries(void) 
{return ++nprim;}

inline int GroverWire::DecrementNPrimaries(void) 
{return --nprim;}

inline int GroverWire::SetNPrimaries(const int nn) 
{if (nn>=0) {nprim = nn; return SUCCESS;} else return FAILURE;}

inline int GroverWire::NSecondaries(void) const 
{return nsec;}

inline int GroverWire::IncrementNSecondaries(void) 
{return ++nsec;}

inline int GroverWire::DecrementNSecondaries(void) 
{return --nsec;}

inline int GroverWire::SetNSecondaries(const int nn) 
{if (nn>=0) {nsec = nn; return SUCCESS;} else return FAILURE;}

#endif
