#ifndef __chamber_boundaries_hh
#define __chamber_boundaries_hh 1

/*
  This file defines SetupChamber_Boundary = vector<BoundaryLine>;
  BoundaryLine = {double start, delta; Direction direction}
  enum Direction {DOWN=0, RIGHT, UP, LEFT};

  In this way one can define the outer boundary of the chamber.
  Start is x for LEFT/RIGHT or y for UP/DOWN, this is the coordinate which
  changes along this BoundaryLine;
  delta is positive for RIGHT/UP and negative for LEFT/DOWN
  The other coordinate which is constant for this BoundaryLine can be found
  as start+delta from the previous BoundaryLine; for the first element
  in vector<BoundaryLine> it is start+delta of the last element (BoundaryLine's
  make a ring).
  There is a function which finds boundary automatically for any geometry:
  
  template <class CellT, class SectionT, class ChamberT>
  inline void find_boundary(SetupChamber<CellT, SectionT, ChamberT>& ch,
			    SetupChamber_Boundary* boundary);
*/
// Author V.Balagura (balagura@mail.desy.de)   (17.11.2001)

#include "ptset/high_pt.hh"

enum Direction {DOWN=0, RIGHT, UP, LEFT};

struct BoundaryLine {
  Direction direction;
  double start; // x for LEFT,RIGHT; y for UP,DOWN
  double delta; // delta is negative for LEFT & DOWN
  BoundaryLine() {}
  BoundaryLine(double fStart, double fDelta, Direction fDirection) :
    start(fStart), delta(fDelta), direction(fDirection) {}
};
struct SetupChamber_Boundary : public vector<BoundaryLine> {};

inline ostream& operator<<(ostream& os, const BoundaryLine& bl) {
  static string dir_names[4] = {"Down", "Right", "Up", "Left"};
  return os << bl.start << " " <<  dir_names[(unsigned int)bl.direction]
	    << " " << bl.delta << '\n';
}
inline ostream& operator<<(ostream& os, const SetupChamber_Boundary& b) {
  copy(b.begin(), b.end(), ostream_iterator<BoundaryLine>(cout, ""));
}

enum _ToFrom_ {_TO_=0, _FROM_=1};
inline void _rotate_(_ToFrom_ to_from, Direction dir, Point* p) {
  // Rotation TO/FROM RIGHT direction
  double tmp;
  switch ( dir ) {
  case RIGHT:  break;
  case UP:    tmp = p->x;                      // +-pi/2 rotation
    if ( to_from == _TO_ ) {p->x =   p->y ; p->y = -tmp;}
    else                   {p->x = -(p->y); p->y =  tmp;}
    break;
  case LEFT: p->x = -(p->x); p->y = -(p->y); break; // pi rotation
    
  case DOWN:  tmp = p->x;                      // +-3*pi/2 rotation
    if ( to_from == _TO_ ) {p->x = -(p->y); p->y =  tmp;}
    else                   {p->x =   p->y ; p->y = -tmp;}
    break;
  }
}
inline void _rotate_(_ToFrom_ to_from, Direction dir, Grid* grid) {
  // Grid is different from Point since its grid.x,y are always positive
  if ( dir == UP || dir == DOWN ) {
    double tmp = grid->x; grid->x = grid->y; grid->y = tmp;
  }
  _rotate_(to_from, dir, &grid->base_offset);
}  
  
template <class CellT, class SectionT, class ChamberT>
inline void find_boundary(SetupChamber<CellT, SectionT, ChamberT>& ch,
			  SetupChamber_Boundary* boundary) {
// first element in *boundary will be directed to the RIGHT,
// second - UP, last - DOWN;
// inner part of the chamber is to the left from every element
  typedef SetupChamber<CellT, SectionT, ChamberT> Chamber;
  typename Chamber::ISection_Cell isc_first; // lower left cell
  Point corner = (*ch[0].begin()).first * ch[0].grid;
  isc_first = typename Chamber::ISection_Cell(ch.begin(), ch[0].begin());
  for (unsigned int is=1; is<ch.size(); ++is) {
    Point corner_is = (*ch[is].begin()).first * ch[is].grid;
    if (  corner.y  > corner_is.y ||
	  (corner.y == corner_is.y && corner.x > corner_is.x) ) {
      isc_first = typename Chamber::ISection_Cell(ch.begin()+is,
						  ch[is].begin());	
      corner = corner_is;
    }
  }
  boundary->push_back(BoundaryLine(corner.x, (*isc_first.isection).grid.x,
				   RIGHT));

  typename Chamber::ISection_Cell isc = isc_first;
  do {
    //    copy(boundary->begin(), boundary->end(),
    //    	 ostream_iterator<BoundaryLine>(cout, ""));
    BoundaryLine& last = boundary->back();
    Grid grid = (*isc.isection).grid;
    Index ind = (*isc.icell).first;
    switch (last.direction) {
    case RIGHT: ++ind.ix; break;  case UP:   ++ind.ix; ++ind.iy; break;
    case LEFT:  ++ind.iy; break;  case DOWN:                     break;
    }
    Point corner = ind * grid; // next corner
    _rotate_(_TO_, last.direction, &grid);   // rotate to direction 
    _rotate_(_TO_, last.direction, &corner); // corresponding to RIGHT
    Point p[2]; p[0] = corner; p[1] = corner; // two probes
    {
      double step_x = grid.x/16; double step_y = grid.y/16;
      p[0].x += step_x; p[1].x += step_x;
      p[0].y -= step_y; p[1].y += step_y; // p[0]->DOWN; p[1]->RIGHT
    }
    _rotate_(_FROM_, last.direction, &p[0]); // rotate from RIGHT direction back
    _rotate_(_FROM_, last.direction, &p[1]);
    
    typename Chamber::ISection_Cell new_isc;
    unsigned int new_dir;
    for (new_dir=DOWN; new_dir<UP; ++new_dir) {
      new_isc = ch.isection_cell(p[new_dir]);
      if ( new_isc.isection != ch.end() ) break;
    } // new_dir after the loop should be just new direction calculated from 
    if ( new_dir != UP ) { // RIGHT new cell is found
      if ( new_isc.isection != isc.isection ) {// get rotated grid if the 
	grid = (*new_isc.isection).grid;        // section is new.
	_rotate_(_TO_, last.direction, &grid);
      }
      isc = new_isc; 
    }
    
    if ( new_dir == RIGHT ) {
      if (last.direction == RIGHT || last.direction == UP) last.delta += grid.x;
      else                                                 last.delta -= grid.x;
    }
    else {
      new_dir = (int(last.direction) + new_dir - int(RIGHT) + 4) % 4;
      // new_dir - RIGHT gives the shift; +4 to make everything positive
      _rotate_(_FROM_, last.direction, &corner);
      double start;
      if ( new_dir == LEFT || new_dir == RIGHT ) start = corner.x;
      if ( new_dir == UP   || new_dir == DOWN  ) start = corner.y;
      double delta = grid.y;
      if ( new_dir == LEFT || new_dir == DOWN  ) delta = -delta;
      boundary->push_back(BoundaryLine(start, delta, (Direction)new_dir));
    }  
  } while (isc != isc_first || boundary->back().direction != RIGHT);
  boundary->pop_back();
}
  

#endif
