// 
//  Author:  Wouter Hulsbergen                 Created:  04-Jul-98
//  Email:   W.Hulsbergen@nikhef.nl            HERA-B Collaboration
// 
// Description: Set of global routines used to fill the hit patterns
// in RTRA/RTRE. Called from rangerTrack::storeback(..)
//
// Modifications:
// - A small bug in printing was fixed                        08/04/02 A.Lanyov
//   Also additional comments were added to the printout making 
//   easier for a human reader to decipher the meaning of bits.
// - Window for OTR hitspattern reduced to <=3 closest cells   02/05/02 W.H.
// included the ITR in the filling of the hit patterns 21.05.02 W.Hulsbergen
//   return if recoOTR != 1 && 3                         19.08.02  A.Spiridonov
//

#include <gean/ZGESL.h>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <digi/zgdig.hh>

#include <rang/Track.hh>
#include <rang/Node.hh>
#include <rang/Hit.hh>
#include <rang/HitPattern.hh>
#include <rang/Geom.hh>
#include <rang/Volume.hh>

#include <wax/DecodeManager.hh>
#include <Otr/OtrI.h>
#include <iomanip>
#include "reco/rcflags.hh"

extern int OtrEventChannelValid(int ci);

// This is a set of routines to fill the hit patterns in vds and
// otr. The inner tracker is not yet included.

namespace rangerHitPattern {
  
  int gverboselevel_ = 0 ;
  int ghitmode_      = 0 ; // 0: take all close hits, 1: only reco hits


  //***********************************************************************
  //
  //  Routine:      prepareRTRE
  //                
  //  Author:       M. Dressel
  //
  //  Description:  
  //  Create entry in RTRE table for this rtra
  //
  //  Last modifications:
  //
  //***********************************************************************

  ArtePointer<RTRE> prepareRTRE(ArtePointer<RTRA> rtra)
  {
    // Create new entry in RTRE table
    int index = rtra.index() ;
    int pos   = ArteTable<RTRE>::size() ;
    if(pos < index ) {
      RTRE rtre ;
      memset((void*) &rtre,0,sizeof(RTRE));
      ArteTable<RTRE>::allocate(index-pos) ;
      for(ArteTable<RTRE>::iterator irtre(pos+1) ; 
	  irtre != ArteTable<RTRE>::end() ; ++irtre) *(*irtre) = rtre ;
    }
    return ArtePointer<RTRE>(index) ;
  }

  //***********************************************************************
  //
  //  Routine:      SiPlaneIsHit
  //                
  //  Author:       M. Dressel
  //
  //  Description:  
  //  Calculate if vds plane is hit. 
  //
  //  Last modifications:
  //
  //***********************************************************************

  int SiPlaneIsHit( ArtePointer<GEDE> sigede, float xx, float yy){
    // SiPlaneIsHit checks if the track coordinates xx and yy
    // are in the sensitive area of the VDS GEDE plane sigede.
    //  
    // Input: all input is expected to be in global coordinates
    //        GEDE pointer to VDS plane: sigede 
    //        coordinates of a track on the GEDE plane: xx, yy
    //
    // return values: 
    // 1: VDS GEDE plane is crossed by extrapolated track
    // 0: VDS GEDE plane is not crossed.
    // negative values flag errors:
    // -1: unvalid GEDE plane
    
    const float LIMIT = 0.005; // to be 50 micron from the nominal boarders
    
    if ( sigede->cmp != 1) return -1; 
    
    int Nstrips = sigede->nwi[0];
    
    // sensitive area
    float sxmn = sigede->sxmn+LIMIT;
    float sxmx = sigede->sxmx-LIMIT;
    float symn = sigede->symn+LIMIT;
    float symx = sigede->symx-LIMIT;
    
    // check if track coordinates are on the sensitive area
    if ( xx < sxmn ||
	 xx > sxmx ||
	 yy < symn ||
	 yy > symx ) return 0;
    
    // projection of sensitive area on hitb->x coordinate
    float pit = sigede->pit;
    float xfw = sigede->xfw[0]; 
    float hbxmax = xfw - 0.5 * pit + pit*Nstrips   - LIMIT;
    float hbxmin = xfw - 0.5 * pit                 + LIMIT;
    
    // rotate positions of track
    float xx_rot = sigede->rot[0]*xx + sigede->rot[3]*yy;
    float yy_rot = sigede->rot[1]*xx + sigede->rot[4]*yy;
    
    // hitb->x of track coordinate
    float hbx = sigede->cos*xx_rot - sigede->sin*yy_rot;
    
    if ( hbx < hbxmin || hbx > hbxmax ) return 0;
    
    return 1;
    
  }

  //***********************************************************************
  //
  //  Routine:      fillVdsHitPattern
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill rtra->bse
  //
  //  Last modifications:
  //
  //***********************************************************************

  void fillVdsHitPattern(const rangerTrack* track, ArtePointer<RTRA> rtra)
  {
    typedef rangerTrack::rangerNodeConstIterator rangerNodeConstIterator ;
    unsigned int rc=0; 
    for(rangerNodeConstIterator nodeIt = track->begin() ;
	nodeIt != track->end() ; nodeIt++) {
      int plane = (*nodeIt)->plane() ;
      if( plane > 0 ) {
	ArtePointer<GEDE> gede = ArtePointer<GEDE>(plane) ;
	if(gede->cmp == LGESL_SVD) {
	  const rangerState& state = (*nodeIt)->getState() ;
	  float dz = 0.5*(gede->zmin+gede->zmax) - state.z() ;
	  float xx = state.x() + dz*state.tx() ;
	  float yy = state.y() + dz*state.ty() ;
	  // (bye the way, dz seems to be always zero)
	  if(  SiPlaneIsHit(gede,xx,yy)>0 ) {
	    int layer = 4 * (gede->gesl->nsl - 1) + gede->lay - 1;
	    rc |= 1 << layer ;
	  } 
	}
      }
    }
    rtra->bse = rc ;
  }
  
  //***********************************************************************
  //
  //  Routine:      fillOtrHitPattern
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill hit pattern entries in rtre
  //
  //  Last modifications:
  //
  //***********************************************************************

  unsigned char otrLayToBit( int geslnsl, int lay)
  {
    static unsigned char layindex[13][6]  ;
    static bool init=false ;
    if(!init) {
      init=true ;
      unsigned char nsinglelay[13][6] ;
      for( unsigned char isl=0; isl<13; isl++)
	for( unsigned char ilay=0; ilay<6; ilay++)
	  nsinglelay[isl][ilay]=0 ;
      unsigned char lastmc(0),lastpc(0) ;

      for(ArteTable<GESL>::iterator igesl = ArteTable<GESL>::begin() ;
	  igesl != ArteTable<GESL>::end() ; ++igesl) 
	if( igesl->cmp==2 ) {
	  for(ArteRelationIterator<GEDE> igede = (*igesl)->gede.begin() ;
	      igede != (*igesl)->gede.end() ; ++igede) 
	    if((*igede)->sec>0 &&
	       nsinglelay[igesl->nsl-1][(*igede)->lay-1]<(*igede)->nsl/2)
	      nsinglelay[igesl->nsl-1][(*igede)->lay-1]=(*igede)->nsl/2 ;
	  if( igesl->name[0] == 'M' ) lastmc = igesl->nsl-1 ;
	  if( igesl->name[0] == 'P' ) lastpc = igesl->nsl-1 ;
	}
      
      unsigned char thislayindex=0;
      for( unsigned char isl=0; isl<13; isl++) {
	for( unsigned char ilay=0; ilay<6; ilay++) {
	  layindex[isl][ilay]=thislayindex ;
	  thislayindex += nsinglelay[isl][ilay] ;
	}
	if( isl==lastmc) thislayindex=32 ;
	if( isl==lastpc) thislayindex=64 ;
      }
    } 
    unsigned char rc =  layindex[geslnsl-1][lay-1] ;
    return rc ;
  }

  unsigned char itrLayToBit(int geslnsl, int lay)
  {
#define MAXNSL 15
    static unsigned char sllayoffset[MAXNSL]  ;
    static bool init=false ;
    if(!init) {
      init=true ;
      const char itr2otrname[MAXNSL][5] = {"MC01","XX02","MC03","XX03","MC05","MC06","XX07","XX08",
					     "XX09","PC01","PC02","PC03","PC04","TC01","TC02"} ;
      // map these names to indices
      unsigned int itr2otrgeslnsl[MAXNSL] ;
      for(int i=0; i<MAXNSL; ++i) itr2otrgeslnsl[i]=0 ;
      for(ArteTable<GESL>::iterator igesl = ArteTable<GESL>::begin() ;
	  igesl != ArteTable<GESL>::end() ; ++igesl) 
	if( igesl->cmp==2 )
	  for(int i=0; i<MAXNSL; ++i) 
	    if( strncmp(igesl->name,itr2otrname[i],4)==0 ) {
	      itr2otrgeslnsl[i] = igesl->nsl ;
	      break ;
    }
      for(ArteTable<GESL>::iterator igesl = ArteTable<GESL>::begin() ;
	  igesl != ArteTable<GESL>::end() ; ++igesl) 
	if( igesl->cmp==3 ) {
	  unsigned int msi = std::atoi(igesl->name+2) ;
	  if(msi <= MAXNSL && itr2otrgeslnsl[msi-1])
	    sllayoffset[igesl->nsl-1] = otrLayToBit(itr2otrgeslnsl[msi-1],1) ;
	  else
	    std::cout << "rangerHitPattern::itrLayToBit: cannot find otr layer for " 
		      << igesl->name << " " << msi << std::endl;
	}
    }
    return sllayoffset[geslnsl-1] + (lay-1);
#undef MAXNSL
  }

  bool otrGedeWireIsGood(ArtePointer<GEDE> gede, int mono, int cell)
  {
    static int hitprepflag=-1 ;
    if( hitprepflag<0 ) {
      char* f ;
      if ((f= getenv("OtrHitPreparation"))) hitprepflag = atoi(f);
      else hitprepflag=0 ; //OTrI is default
    }
    bool rc=false ;
    if( hitprepflag==1 ) { // wax
      static waxDecodeManager* dm = waxDecodeManager::GetMe() ;
      const waxDecodeWire* wire = dm->GetGede(gede.index())->GetWire(cell+32*mono) ;
      if( wire && wire->IsConnected() && wire->GetChannel()->GetStatus()==0 )
	rc = true ;
    } else {               // OTrI
      int ci = OtrIFromNedeLayerWire(gede.index(),mono,cell) ;
      if( !OtrIError() && ci>0 && OtrEventChannelValid(ci)) rc=true ;
    } 
    return rc ;
  }

  void fillOtrHitPattern(const rangerTrack* track, ArtePointer<RTRE> rtre)
  {

    if(recoOTR != 1 && recoOTR != 3) return;
    
    // calculates the number of crossed layers, and the number of found
    // hits in PC, MC, TC
    const float nsigmay = 2 ;
    const float nsigmax = 2 ;
    
    // dangerous not to reset them, but they are partially used from somewher else...
    //rtre->mchi = rtre->pchi = rtre->tchi = 0 ;
    //rtre->mche = rtre->pche = rtre->tche = 0 ;
   
    // we will use this routine in two modes:
    // 0) For tracks without `patt': a `found hit' in a layer is any close hit
    // 1) For tracks  with   `patt': a `found hit' in a layer is a reconstructed hit
    ArtePointer<RTRA> rtra =  ArtePointer<RTRA>(rtre.index()) ;
    int thishitmode = (rtra->cmp&Rsegc::patt)&&ghitmode_>0 ? 1 : 0 ;
    
    // loop over all gedes and see if they are really hit and if there is a hit
    ArtePointer<GEDE> gede ;
    float cosa,sina ;
    float z,x,y,tx,ty,winy,winx,dzgede ;
    const rangerVolume* vol ;
    for(rangerTrack::rangerNodeConstIterator inode = track->begin() ;
	inode != track->end() ; inode++) {
      int plane = (*inode)->plane() ;
      if(plane>0 && 
	 ((gede=ArtePointer<GEDE>(plane))->cmp==3 ||
	  (gede->cmp==2 && gede->sec>0 ) ) ) {
	vol  = rangerGeom::vol( gede ) ;
	cosa = vol->Cos();
	sina = vol->Sin();
	const rangerState& state = (*inode)->getState() ;
	z  = state.z() ; 
	x  = state.x()*cosa  - state.y()*sina ;
	y  = state.y()*cosa  + state.x()*sina ;
	tx = state.tx()*cosa - state.ty()*sina ;
	ty = state.ty()*cosa + state.tx()*sina ;
	const HepSymMatrix& cov = state.cov() ;
	winy = nsigmay*sqrt(cosa*cosa*cov.fast(2,2) +
				  2*cosa*sina*cov.fast(2,1) + sina*sina*cov.fast(1,1)) ;
	winx = nsigmax*sqrt(cosa*cosa*cov.fast(1,1) -
				  2*cosa*sina*cov.fast(2,1) + sina*sina*cov.fast(2,2)) ;
	dzgede = 0.5*(gede->zmin+gede->zmax)-z ;
	
	if(gverboselevel_>=2) {
	  float sLeng = vol->sLengthFast(state) ;
	  if(sLeng<=0) { 
	    gede->hwc[13]='\0' ;
	    std::cout << "volume no longer on track !!" << rtra << gede->zmin << " "
		      << (*inode)->hasHitb() << " " << gede->hwc << " " << plane << std::endl ;
	    std::cout << x+dzgede*tx << " " << gede->sxmn << " " << gede->sxmx << " | "
		      << y+dzgede*ty << " "
		      << gede->symn << " " << gede->symx << std::endl ;
	  }
	}

	bool slhitexpected[2] = {false,false} ;
	bool slhitfound[2]    = {false,false} ;
	
	// Do we expect a hit ?
	if( gede->symn + winy < y+dzgede*ty && y+dzgede*ty < gede->symx - winy &&
	    gede->sxmn + winx < x+dzgede*tx && x+dzgede*tx < gede->sxmx - winx ) {

	  if(gede->cmp==2) { // OTR
	    // max distance for closest cell to track
	    const float rmin = gede->pit/4 ; 
	    
	    // are there any good cells on the track ? (within `small' window)
	    for(unsigned char mono=0; mono<gede->nsl; ++mono) {
	      // find closest cell in this mono layer
	      float thisx = x + (gede->zfw[mono]-z)*tx ;
	      // (the '3' is for safety: we don't want to truncate negative values)
	      unsigned char cell = int((thisx - gede->xfw[mono])/gede->pit+0.5+3) - 3 ;
	      if( cell<gede->nwi[mono] &&
		  fabs( gede->xfw[mono] + cell*gede->pit - thisx ) < rmin 
		  && otrGedeWireIsGood(gede,mono,cell) ) 
		slhitexpected[mono/2] = true ;
	    }
	  } else {           // ITR
	    slhitexpected[0] = true ;
	  }
	}

	// Is there a hit ?
	if( thishitmode==1 ) {
	  if( (*inode)->hasHitb() ) { // there is a reconstructed hit
	    ArtePointer<HITB> hitb = (*inode)->getHitb() ;
	    int sl = gede->nsl>2 && hitb->z > 0.5*(gede->zfw[1] + gede->zfw[2]) ? 1 : 0 ;
	    slhitfound[sl]=true ;
	  }
	} else {
	  const float otrrmax = gede->pit/3 + winx ;//std::min(gede->pit/3 + winx,3*gede->pit/4) ;
	  const float itrrmax = gede->pit + winx ;//std::min(gede->pit + winx,2*gede->pit) ;
	  float rmax  = gede->cmp==2 ? otrrmax : itrrmax ;
	  for(ArteTable<HITB>::iterator ihitb = gede->hit1 ;
	      ihitb != gede->hitl+1 ; ihitb++) 
	    if ( fabs(ihitb->x-(x + (ihitb->z-z)*tx) ) < rmax ) {
	      int sl = gede->nsl>2 && ihitb->z > 0.5*(gede->zfw[1] + gede->zfw[2]) ? 1 : 0 ;
	      slhitfound[sl]=true ;
	    }
	}

	// Fill the bits for each single layer
	int layindex = (gede->cmp==2) ?
	  otrLayToBit(gede->gesl->nsl,gede->lay) :
	  itrLayToBit(gede->gesl->nsl,gede->lay) ;
	for(unsigned char isl=0; isl<(gede->nsl+1)/2 ; ++isl, ++layindex) {
	  switch(layindex/32) {
	  case 0:
	    if(slhitexpected[isl]) rtre->mche |= (1 << (layindex%32)) ;
	    if(slhitfound[isl])    rtre->mchi |= (1 << (layindex%32)) ;
	    break;
	  case 1:
	    if(slhitexpected[isl]) rtre->pche |= (1 << (layindex%32)) ;
	    if(slhitfound[isl])    rtre->pchi |= (1 << (layindex%32)) ;
	    break;
	  case 2:
	    if(slhitexpected[isl]) rtre->tche |= (1 << (layindex%32)) ;
	    if(slhitfound[isl])    rtre->tchi |= (1 << (layindex%32)) ;
	    break;
	  }
	}
      }
    }
  }

  //***********************************************************************
  //
  //  Routine:      fillTrackEndPoint
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill state vector at last hit in rtre
  //
  //  Last modifications:
  //
  //***********************************************************************

  void fillTrackEndPoint(const rangerTrack* track, ArtePointer<RTRE> rtre)
  {
    for(rangerTrack::rangerNodeConstIterator nodeIt = track->begin() ;
	nodeIt != track->end() ; ++nodeIt) 
      if( (*nodeIt)->dim()>0 ) {
	const rangerState& state = (*nodeIt)->getState() ;
	rtre->ze  = state.z() ;
	rtre->xe  = state.x() ;
	rtre->ye  = state.y() ;
	rtre->txe = state.tx() ;
	rtre->tye = state.ty() ;
	rtre->pe  = state.qP() ;
	break;
      }
  }

  //***********************************************************************
  //
  //  Routine:      printHitPatterns(RTRA)   
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Print all hit patterns
  //
  //  Last modifications:
  //
  //***********************************************************************

  inline void printbitpattern(unsigned int hp, int len=32)
  {
    for( int bit=0; bit<len; bit++)
      std::cout <<  (( hp & (1<<bit) ) ? '*' : '.') ;
    std::cout << std::endl ;
  }

  void printHitPatterns(ArtePointer<RTRA> rtra)
  {
    ArtePointer<RTRE> rtre(rtra.index()) ;
    std::cout << "rtra,hsi,hou,hin,hca: " << rtra << " "
	      << rtra->hsi << " " << rtra->hou << " "
	      << rtra->hin << " " << rtra->hca << std::endl ;
    
    std::cout << "VDS layers: [01][02][03][04][05][06][07][08]" << std::endl;
    std::cout << "rtra->bse:  " ; printbitpattern(rtra->bse) ;
    std::cout << "rtra->bsi:  " ; printbitpattern(rtra->bsi) ;
    
    std::cout << "MC  layers: [1][2][3][4]5[6][8]" << endl;
    std::cout << "rtre->mche: " ; printbitpattern(rtre->mche,19) ;
    std::cout << "rtre->mchi: " ; printbitpattern(rtre->mchi,19) ;

    std::cout << "PC  layers: [ PC01  ][PC02][PC03][ PC04  ]" << std::endl;
    std::cout << "rtre->pche: " ; printbitpattern(rtre->pche,30) ;
    std::cout << "rtre->pchi: " ; printbitpattern(rtre->pchi,30) ;

    std::cout << "TC  layers: [TC01][TC02]" << endl;
    std::cout << "rtre->tche: " ; printbitpattern(rtre->tche,12) ;
    std::cout << "rtre->tchi: " ; printbitpattern(rtre->tchi,12) ;
  }

  //***********************************************************************
  //
  //  Routine:      storeRtra    
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill rtra->bse and rtre table
  //
  //  Last modifications:
  //
  //***********************************************************************
  
  int storeRtra( rangerTrack* track, ArtePointer<RTRA> rtra)
  {  
    // setverboselevel( (rtra->hsi>0 && (rtra->hou+rtra->hin)>0) ? 1 : 0 ) ; 

    // lets' make this a uniform approach
    // prepare storing endpoint of track
    ArtePointer<RTRE> rtre = prepareRTRE(rtra) ;
 
    // 1) extrapolate track to vds and fill vds hit pattern
    if( rtra->cmp & Rsegc::vxd ) {
      track->extrapolateUpstream() ;
      fillVdsHitPattern(track,rtra) ;
    }
    
    // 2) extrapolate track to TC and fill otr hit pattern
    if( rtra->cmp & (Rsegc::patt | Rsegc::ecal) 
	/*&& (rtra->cmp & Rsegc::vxd)*/ ) {
      track->extrapolateDownstream() ;
      fillOtrHitPattern(track,rtre) ;
    }

    // 3) fill track parameters at last hit
    fillTrackEndPoint(track,rtre) ;
    
    if(gverboselevel_>=1) printHitPatterns(rtra) ;
    return 0;
  }

}
