// ********************************************************************
// Routine: rcRTRAmatchMTRA 
// Author: Wouter Hulsbergen (November 2000)
//
// Description: Match vds, tracker and vds-tracker RTRAs to MTRA. Only
// one relation per rtra is created. Using a global variable called
// 'rcRTRAmatchMTRA_MatchMode' the matching can be switched between to
// modes, which are described below. One method uses the RSEG-MTRA
// match. The other uses HITB-MIMP-MTRA relations.
//
//  Modifications:
//
// ********************************************************************

#include <set>
#include <map>
#include <arte/arte.hh>
#include <gean/ZGESL.h>
#include <arte/RSEG_cmp.hh>

int rcRTRAmatchMTRA_rseg( ArtePointer<RTRA> ) ;
int rcRTRAmatchMTRA_hitb( ArtePointer<RTRA> ) ;
int rcRTRAmatchMTRA_MatchMode = 0 ;

int rcRTRAmatchMTRA() {
  // Choose the method based on HITB-MIMP-MTRA
  const int mode = rcRTRAmatchMTRA_MatchMode ;

  int rc = 0;
  if ( ArteTable<MTRA>::size() == 0 )  rc = -1;
  else {
    for( ArteTable<RTRA>::iterator rtraIt = ArteTable<RTRA>::begin() ;
	 rtraIt != ArteTable<RTRA>::end() ; rtraIt++) 
      if( (rtraIt->cmp & ( Rsegc::vxd | Rsegc::patt ) ) ) {
	if ( mode == 0 ) rc |= rcRTRAmatchMTRA_hitb(*rtraIt) ;
	else             rc |= rcRTRAmatchMTRA_rseg(*rtraIt) ;
      }
  }

  return rc ;
}

// **********************************************************************
//
// Routine: rcRTRAmatchMTRA_rseg
// Author:  Wouter Hulsbergen ( November 2000 )
//
// Description: Match vds, tracker or vds-tracker RTRAs to MTRA using
// RSEG-MTRA relations. Only one relation per rtra is created. If no
// relation is created for a vds-only of tracker-only RTRA this means
// that the segment is a ghost.  If no relation is created for a
// vds-tracker RTRA, this implies that either one of the two segments
// is a ghost, or that it is a ghost match.
//
// **********************************************************************

int rcRTRAmatchMTRA_rseg( ArtePointer<RTRA> rtra ) {

  // Loop over all tracker and vds segments. Check that they have mtra
  // match. Check that all segments have same mtra match. Only then, create
  // relation.

  int imtra = 0 ;
  for( ArteRelationIterator<RSEG> rsegIt = rtra->rseg.begin() ;
       rsegIt != rtra->rseg.end() && imtra >=0 ; ++rsegIt)
    if( (*rsegIt)->fit==0 && ((*rsegIt)->cmp & ( Rsegc::vxd | Rsegc::patt )) ) {

      int thisimtra=0 ;
      for( ArteRelationIterator<MTRA> mtraIt = (*rsegIt)->mtra.begin() ;
	   mtraIt != (*rsegIt)->mtra.end() && thisimtra >=0; ++mtraIt)
	if( !thisimtra ) thisimtra = (*mtraIt).index() ;
	else if( thisimtra != int((*mtraIt).index()) ) thisimtra = -1 ;
      
      if( thisimtra < 0 || (imtra >0 && thisimtra != imtra ) ) imtra = -1;
      else imtra = thisimtra ;
    }

  if(imtra > 0) createRelation( rtra, ArtePointer<MTRA>(imtra) , &RTRA::mtra ) ;
  

  return 0; 
}

// **********************************************************************
//
// Routine: rcRTRAmatchMTRA_hitb
// Author:  Wouter Hulsbergen ( November 2000 )
//
// Description: Match vds, tracker or vds-tracker RTRA to MTRA using
// HITB-MIMP-MTRA relations. Only one relation per rtra is created. If
// no relation is created this means that there were not enough hits
// in the rtra corresponding to the same mtra. The routine contains
// two parameters: the minimum fraction of good hits and the maximum
// number of faults. The values 0.7 and 4 respectively ensure that a
// wrong vds-tracker match will not be matched to MTRA.
//
// **********************************************************************

int rcRTRAmatchMTRA_hitb( ArtePointer<RTRA> rtra ) {
  
  const float minfrac = 0.7 ; // minimum number of hits in rtra corresponding to mtra
  const int   maxflt  = 4   ; // maximum number of wrong hits in rtra
  
  // Loop over all hitbs and count for every mtra how many times it
  // appears.
  map<ArtePointer<MTRA>,int> mtramap ;
  int nhits(0) ;
  for(ArteRelationIterator<HITB> hitbIt = rtra->hitb.begin() ;
	hitbIt != rtra->hitb.end() ; hitbIt++) {
    int lgesl = (*hitbIt)->gede->cmp ;
    if(lgesl==LGESL_SVD || lgesl==LGESL_OTR || lgesl==LGESL_ITR) {
      nhits++ ;
	for(ArteRelationIterator<MIMP> mimpIt = (*hitbIt)->mimp.begin() ;
	    mimpIt != (*hitbIt)->mimp.end() ; mimpIt++)
	  mtramap[(*mimpIt)->mtra]++ ;
    }
  }
  
  // Loop over all related mtras and check whether number of
  // found hits large enough
  for( map<ArtePointer<MTRA>,int>::iterator mapIt = mtramap.begin() ;
       mapIt != mtramap.end() ; mapIt++) 
    if(mapIt->second >= nhits*minfrac && nhits-mapIt->second <=maxflt )
      createRelation( rtra, mapIt->first , &RTRA::mtra ) ;
  
  return 0; 
}
