
// *****************************************************************************
//
// source:    rcSelectSLT.C
//
// type:      source code
//
// created:   26 Jul 2000
//	
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcSelectSLT
//
// called by: rcModSel and rcClassSLT
//
// purpose:   classification according to SLT record
//
// hints:     return value rc = 0     nothing classified
//                              bit 1 0x1  SLTEE
//                              bit 2 0x2  SLTMM
//                              bit 3 0x4  SLTDV
//                              bit 4 0x8  SLTEEB
//                              bit 5 0x10 SLTCUT
//                              bit 6 0x20 SLTYPS
//
// changes:
// 26 Jul 2000 (AG) creation (code by Brian Aagaard Petersen)
// 03 Aug 2000 (AG) polished code 
// 04 Aug 2000 (BP) Added mass calculation
// 08 Aug 2000 (AG) clarified SLT bits
// 16 Aug 2000 (AG) introduced new class l+l- > 8.0 GeV ECLASS_SLTYPS
//  4 Dec 2000 (AG) polished printout
//
// *****************************************************************************

// standard includes
#include <iostream.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"

// SLT includes
extern "C" {
#include "sltinfo.h"
#include "l2field.h"
}

// externals
extern int FADEB;

// prototypes
static int rcSelectSLTBrems( SLT_SILITRACK &sili );
static float SLTMass(SLT_TRIGGER &trig);

// constants
const float SLTMassCut  = 2.0; // J/Psi search
const float SLTMassCut2 = 8.0; // Ypsilon search

//
// function code
//
int
rcSelectSLT()
{
  ARTE_mark( "rcSelectSLT 0", 0 );

  // event number
  static int old = 0;
  int        evt = faEVHDevt();
  
  // print
  if( FADEB ) cout << "SELECT: evt=" << evt << " old=" << old << endl;

  // new event
  static int slt = 0;
  if( evt != old ){

    // store event number
    old = evt;
    
    // reset slt result for new events
    slt = 0;

    // check SLT record
    ARTE_mark( "SLTReadDSLT 0", 0 );
    if( SLTReadDSLT() == SLT_OK ){

      SLT_STATUS  stat;
      SLT_TRIGGER trig;

      if( SLTStatus( &stat ) == SLT_OK ){
	for( int i=0; i<stat.ntriggers; i++ ){

	  if( SLTTrigInfo( i, &trig ) != SLT_OK )  continue;
	  if( trig.status == SLTTRIG_REJECT )      continue;
	  
	  // e+ e-
	  if( trig.type == SLTTYPE_DIELECTRON ){
	    slt |= 0x1;
	    for( int j=0; j<trig.ntracks; j++ ){
	      SLT_TRACK track;
	      SLT_SILITRACK sili;
	      if( SLTTrackInfo( trig.trackindex[j], &track ) != SLT_OK ) continue;
	      if( track.status == SLTTRACK_ACCEPT ){
		if( SLTSiliTrackInfo( trig.trackindex[j], &sili) != SLT_OK ) continue;
		if( rcSelectSLTBrems( sili ) ) slt |= 0x8;
	      }
	    }
	  }
	  
	  // mu+ mu-
	  if( trig.type == SLTTYPE_DIMUON )        slt |= 0x2;
	  
	  // dv
	  if( trig.info & SLTTRIG_DETACHEDVERTEX ) slt |= 0x4;

	  // mass
	  float mass = SLTMass( trig );
	  if( mass > SLTMassCut )  slt |= 0x10;
	  if( mass > SLTMassCut2 ) slt |= 0x20;

	}
      }
    }
    ARTE_mark( "SLTReadDSLT 1", 0 );
  }

  // print
  if( slt ) cout << " SLT: rc=0x" << hex << slt << dec << endl;
  
  ARTE_mark( "rcSelectSLT 1", 0 );
  return slt;
}

//
// function
//
static int 
rcSelectSLTBrems( SLT_SILITRACK &sili )
{
  //-----------------------------------------------------------------
  //   select two body decays
  //-----------------------------------------------------------------
  
  float energy    = 0;
  float minImpact = 100000.;
  
  float xECAL, yECAL, zECAL;
  float xTrack, yTrack, zTrack, txTrack, tyTrack;
  float yimpact, ximpact, rimpact;

  for( ArteTable<RCCL>::iterator rccl = ArteTable<RCCL>::begin();
       rccl != ArteTable<RCCL>::end();
       ++rccl ){
        
    xECAL   = (*rccl)->x;
    yECAL   = (*rccl)->y;
    zECAL   = (*rccl)->z;
    xTrack  = sili.bx;
    yTrack  = sili.by;
    zTrack  = 0;
    txTrack = sili.tx;
    tyTrack = sili.ty;
    
    yimpact = yTrack + tyTrack * (zECAL - zTrack) - yECAL;
    ximpact = xTrack + txTrack * (zECAL - zTrack) - xECAL;
    rimpact = sqrt( ximpact*ximpact + yimpact*yimpact );
    if( rimpact < minImpact ){
      minImpact = rimpact;
      energy    = (*rccl)->esum;
    }
  }
  
  float ycut = ( 1 + 3 * ( 1 - energy / 100 ) ) / sqrt( 2 );
  
  if( minImpact > ycut ) energy = 0;
  
  return energy > 0;
}

//
// function
//
static float
SLTMass( SLT_TRIGGER &trig )
{

  SLT_SILITRACK sili[2];
  SLT_REFITTRACK refit[2];
  SLT_TRACK track[2];
  float p[2];

  // Get track parameters and check tracks were accepted 

  if ( trig.ntracks != 2) return 0.;
  for(int j=0;j<2;j++) {
    if (SLTTrackInfo(trig.trackindex[j],&track[j])!=SLT_OK) return 0.0;
    if (track[j].status!=SLTTRACK_ACCEPT) return 0.0;
    if (SLTRefitTrackInfo(trig.trackindex[j],&refit[j])!=SLT_OK) return 0.0;
    if (SLTSiliTrackInfo(trig.trackindex[j],&sili[j])!=SLT_OK) return 0.0;

//               Redo Momentum calculation:
    float h[4];
    float pinv;
    
    pinv=1/sili[j].pxpz;
	       
    for (int i=0;i<3;i++) {
      FieldGetValue(sili[j].tx,sili[j].ty,pinv,h);
      pinv=((refit[j].tx-sili[j].tx)/h[0]);
    }
    p[j]=fabs(1./pinv); 
  }

  /* mass */
  double a,b,c,d,e,mycos;
  float mass;
  
  a = sqrt(sili[0].tx*sili[0].tx + sili[0].ty*sili[0].ty + 1.);
  b = sqrt(sili[1].tx*sili[1].tx + sili[1].ty*sili[1].ty + 1.);
  c = sili[0].tx*sili[1].tx + sili[0].ty*sili[1].ty + 1.;
  
  d = sqrt(p[0]*p[0]);
  e = sqrt(p[1]*p[1]);
   
  mycos = c/(a*b);
  
  mass =  sqrt( 2*(d*e - p[0]*p[1] * mycos));

  return mass;


}


