/*********************************************************
 *
 * Main L2-Formatter code for CnA, including the handlers 
 *
 * Authors: Peter Kreuzer, Jose Hernandez
 *
 * Date: Spring 2000
 *
 * Functionalities: Main L2Formatter function, running 
 *                  online (L2_CnA) or offline (NO L2_CnA).
 *
 *                  For online mode, the code should be compiled 
 *                  with the L2_CnA pre-compiler flag and run with 
 *                  the "-L4Odummy" argument on the command line. 
 *                  This ARTE mode does not expect event looping.
 *
 *                  In offline mode, the default is the SLTsim mode, 
 *                  in which the SLT formatted tables are initialized 
 *                  via the CnAKey and piped to the EDC executable. 
 *                  However, also a stand-alone mode is forseen offline, 
 *                  (set via the "-L4Odummy" run argument like online). 
 *                  This mode allows us to test L2FormArte without all 
 *                  the whole CnA machinery (CnA manager, client etc) 
 *                  and to dump the formatted tables in ASCII or BINARY 
 *                  format.
 *                             
 *                  The handlers for Geo/Buffmap/Trigger-type/Trigger-symbol
 *                  tables are also defined here
 *
 * Modifications: -The Geo/Buffmap handler (l2fmt_hdl) is 
 *                now uniquely defined and called for all 
 *                CnA modes (Jan 23 2001, PK)
 *                -The TrigTab handler (tfmt_hdl) was 
 *                added (Feb 2001, TN)
 *                -Clean-up and adding otrsltwritetofile
 *                for offline OTR calib debugging 
 *                (May 30 2001, PK) 
 *                -Add the functionality needed in SLTsim
 *                mode (June 5 2001, PK/LS)
 *                -Update the overall steering related to 
 *                StandAlone/SLTsim mode (June 28 2001, PK)
 *                -Update the handling of the CnAkey, now 
 *                 extracted from general ARTE runnumber
 *                (Sep 9 2001, PK)
 *                -Reduce to minimum the pre-proc. flags,
 *                replace by runparameters instead:  
 *                (Nov 4 2001, PK):
 *                --> SLTsim mode incl. piping to EDC: default
 *                --> CnA mode: pre-proc. flag "L2_CnA"
 *                --> StandAlone mode: "-StandAlone" in run par
 *                --> MC/data modes are distinguished by test 
 *                on MCIN table size or can be forced via 
 *                "-mcmode" run parameter)
 *                (Feb 4 2002, PK):
 *                added lots of comments here and there....
 *                (22 April 2002, KR)  
 *                -->  General housekeeping.
 ********************************************************/
#include "l2formarte.hh"

/***
 * l2form_PipeInit:
 * Function to initialize the Pipe for SLTsim mode
 * Author: Lars Soezueer (April 2001)
 *
 ***/

static int l2io_fd;
static char* fifo;
#define MAXTRIES 30
#define NAPTIME 1
#define MAXPIPE 4096

int l2form_PipeInit() 
{
  int fd=-1, tries;
  char* slt_dom;
  char tmpstr[80], mess[80];
  
  slt_dom = getenv("SLT_DOM");
  
  if (slt_dom==NULL)
    { 
      ARTE_error(GRR_SERROR,"Couldn't get SLT_DOM",0);
      return 1;
    }
  if (slt_dom[0]=='/')  slt_dom = slt_dom + 1;
  if (slt_dom[strlen(slt_dom)-1]=='/')  slt_dom[strlen(slt_dom)-1]='\0';
  sprintf(tmpstr,"/tmp/fifo_%s",slt_dom);
  fifo = tmpstr;

  if (mkfifo(fifo,0666)==-1)
    {
      if (errno!=EEXIST)
	{
	  ARTE_error(GRR_SERROR,"Couldn't create fifo",0);
	  return 1;
	}
      else ARTE_error(GRR_SINFO, "Using existing fifo",0);
    }
  
  sprintf(mess,"l2form_PipeInit: fifo %s created\n",fifo); 
  ARTE_error(GRR_SINFO, mess,0);
  
  for (tries=1; tries <= MAXTRIES; tries++)
    {
      if ((fd=open(fifo, O_WRONLY|O_NDELAY))!=-1) break;
      if (errno!=ENXIO)
	{
	  if (errno>0&&errno<sys_nerr)
	    {
	      sprintf(mess,"ERROR: open (%d; %s)",errno,sys_errlist[errno]);
	      ARTE_error(GRR_SERROR,mess,0);
	    }
	  return 1;
	}
      sprintf(mess,
	      "fifo not ready for sending. Will wait %d seconds.\n",
	      NAPTIME*tries);
      ARTE_error(GRR_SINFO, mess,0);
      sleep(NAPTIME*tries);
    }
  if (fd==-1)
    {
      ARTE_error(GRR_SERROR,"Couldn't open fifo",0);
      return 1;
    }
  if (fcntl(fd,F_SETFL,O_WRONLY)==-1)
    {
      ARTE_error(GRR_SERROR,"Couldn't set permissions",0);
      return 1;
    }
  ARTE_error(GRR_SINFO,"l2form_PipeInit: fifo successfully opened",0);
  
  l2io_fd = fd;
  
  return 0;
}

/***
 *
 * Trig_Sym_major_minor get_major_minor:
 * Special function used in 2000 simulation (SLTsim),
 * to extract major/minor of 
 * the trigger/symbol tables, needed since these tables were missing in the 
 * keytable during 2000 run 
 * - if the input runnumber is smaller than the first 2000 
 * trigger run (15033), then 0's are returned.
 * - if the input runnumber is within the 2000-trigger run 
 * series, then the major/minor are found via a hard coded 
 * look-up table and returned.
 * - if the input runnumber is larger than the latest 
 * 2000-Run (17267), then 0's are returned.
 * Finally a special case is forseen to force the trigger/sybol tables to be 
 * generated according to a list ("runnumber"= major from 1 to 7) in the kumac
 * via the old "/SLT/SIM/runnumber <rununmber>" card (See SLTsim docu)
 *
 * Author:   Lars Soezueer (April 2001)
 * Modified: Peter Kreuzer (May 2001): bug fix in the look-up 
 *           table.  
 * Modified: Peter Kreuzer (Sep 9 2001): new handling of CnAkey 
 * Modified: Peter Kreuzer (Nov 4 2001): modified return code; 
 *           moved error handling to tfmt_hdl
 * 
 ***/

Trig_Sym_major_minor get_major_minor(int runnumber)
{
  
  Trig_Sym_major_minor r;
  //  Tables database;
  //  int rc, tx, n_max = 255, n_used, i;
  //  char **servernames, **tablenames;
  //  int *transaction, *major, *minor;
  //  char mess[256];
  //  bool got_tt=false;
  // bool got_st=false;
  char *navail="No SLT Tables available for this runnumber.";
  //  char line[256];
  r.tt_major = 0;
  r.st_major = 0;
  r.tt_minor = 0;
  r.st_minor = 0;
  
  if (runnumber>=17267)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=17242)
    {r.tt_major=SINGLE_LEPTON_FLT; r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17214)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=17211)
    {r.tt_major=DI_LEPTON_FLT;     r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17190)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=17173)
    {r.tt_major=DI_LEPTON_FLT;     r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber==17172)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=17158)
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17153)
    {r.tt_major=DI_LEPTON_FLT;     r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17138) 
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber==17137) 
    {r.tt_major=SINGLE_LEPTON;     r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17127)
    {r.tt_major=SINGLE_LEPTON_FLT; r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17110)
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=17100)
    {r.tt_major=HARD_PHOTON;       r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=16863)
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=16799)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=16012)
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=15986)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=15713)
    {r.tt_major=DI_LEPTON;         r.st_major=DEFAULT_SYM; return r;}
  else if (runnumber>=15687)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=15487)
    {r.tt_major=DI_LEPTON_L2SILI;  r.st_major=L2SILI_SYM; return r;}
  else if (runnumber>=15454)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=15240)
    {r.tt_major=DI_LEPTON_L2SILI;  r.st_major=L2SILI_SYM; return r;}
  else if (runnumber>=15231)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=15112)
    {r.tt_major=DI_LEPTON_L2SILI;  r.st_major=L2SILI_SYM; return r;}
  else if (runnumber>=15091)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}
  else if (runnumber>=15033)
    {r.tt_major=DI_LEPTON_L2SILI;  r.st_major=L2SILI_SYM; return r;}
  else if ((runnumber >= 1) && (runnumber <= 7))
    {r.tt_major=runnumber;r.st_major=(runnumber==7 ? 2 : 1);return r;}
  else if (runnumber <= 0)
    {ARTE_error(GRR_SERROR, navail, 0); return r;}

  // should never reach here
  ARTE_error(GRR_SERROR, navail, 0);
  return r;
}

extern "C" {
  
  /***
   *
   * Main L2Formarte code:
   *
   * The code is divided into L2_CnA / non-L2_CnA modes. The 
   * former is expecting rpm messages inside an "infinite while 
   * loop" in order to invoke the L2Formatter handlers which 
   * have been referenced to in the CIF subscribtion/initialization:
   * - tfmt_hdl in formatter_trigger_init() for L2Symbol and 
   * L2Trigger tables 
   * - l2fmt_hdl in formatter_l2_init() for l2Geo and L2BuffMap 
   * - otrsltFormatInit() for L2OtrCalib table
   *
   * The Monte-Carlo mode can be forced via the "-mcmode" parameter 
   * on the command line.
   *
   * The StandAlone mode cxan be forced via "-L40dummy" parameter;
   * no event input file should be given in that care. This mode 
   * should be used for the real online L2FormArte and for offline 
   * CnA table generation/tests. 
   *
   ***/

  void L2FormArte()
  {
    static int L2_CnA_FlagIsSet=0, CnA_Debug_FlagIsSet=0;
#ifdef L2_CnA
    L2_CnA_FlagIsSet=1;
#endif
#ifdef DEBUG_CNA
    CnA_Debug_FlagIsSet=1;
#endif

    static int magOn = 1;  
    static int mode = 0;
    static int siliConfig = 0;
    char mess[200];
    int KTstatus = 0;    
    int irun;
    
    // L2-geometry configuration set-up 
    //L2Mode_set(L2MODE_NOITRPIX);
    
    // Get run argument to determine L2Formatter mode (Standalone or SLTsim) 
    for( int i=1; i<nArg; i++ )
      {
	if( !strncmp( vArg[i], "-L4Odummy", 9 ) )
	  {
	    StandAlone = 1;
	    sprintf(mess,
		    "\nL2FormArte: Skipping EMG partition!");
	    ARTE_error(GRR_SINFO,mess,0); 
	    sprintf(mess,
		    "  (no event input expected)\n");
	    ARTE_error(GRR_SINFO,mess,0); 
	  }
      }
    // Get Run argument to check if mcmode or not
    for( int i=1; i<nArg; i++ ){
      if( !strncmp( vArg[i], "-mcmode", 7 ) ){
	mcflag = 1;
	sprintf(mess,"\nL2FormArte: mcmode !\n");
	ARTE_error(GRR_SINFO,mess,0); 
      } 
    }
    if (ArteTable<MCIN>::size()!=0) mcflag = 1;
    
    if(L2_CnA_FlagIsSet)   //CnA Online Mode    
      {
	int rc;
	char name[200];
	rpm_stat(RPM_DOMAIN,200,name);
	rpm_stat(RPM_PROCESS_NAME,200-strlen(name),name+strlen(name));
	grr_exec(name);
	
	// Set-up publication client interface     
	rc = pub_cinit();
	if ( rc ) ARTE_error(GRR_SFATAL,"pub_cinit complained",rc);
	
	// Set-up CIF formatter interface 
	rc = CIF_msg_init_formatter (Kcna);
	if (rc) ARTE_error(GRR_SFATAL,"CIF_msg_init_formatter returned error",rc);
	
	// Register formatted tables:
	
	// trigger/symbol formatter
	rc = formatter_trigger_init();
	if (rc) ARTE_error(GRR_SERROR,"formatter_user_setup for l2 trigger/symbol table complains",rc);      
	
	// L2OtrCalib formatter
	rc = otrsltFormatInit();
	if (rc) ARTE_error(GRR_SERROR,"formatter_user_setup for sltotr_CALIB complains",rc);      
	
	// l2geo/l2map formatter
	rc = formatter_l2_init();
	if (rc) ARTE_error(GRR_SERROR,"formatter_user_setup for l2 complains",rc);      
	
	// Patiently waiting for order from CnA manager
	ARTE_error(GRR_SINFO,"\nL2FormArte/usevnt: entering CnA Loop !",0);
	while(1){
	  char data[RPM_LMAX];
	  rpm_hdr hdr;
	  
	  rc=rpm_receive(&hdr,data,RPM_LMAX,BLOCK_FOREVER);
	  
	  if (  rc == GRR_CONTINUE )
	    continue;
	  else if ( rc ){
	    ARTE_error(GRR_SERROR,"rpm_receive complaines",rc);
	    continue;
	  }
	  
	  // load in memory current keytable before ftab handlers are called 
	  if(hdr.type == CNA_MSG_END_UPDATES){
	    L2FormARTE_KTstatus = KeyLoadTables(((CNA_msg_end_updates*)data)->key);
	    sprintf(Key,"%d",((CNA_msg_end_updates*)data)->key);
	    sprintf(mess,"L2FormArte/usevnt: input key from CnA manager:%d, KTstatus %d\n",((CNA_msg_end_updates*)data)->key, L2FormARTE_KTstatus);
	    grr_error(GRR_SINFO,mess,0);
	  }
	  rc = rpm_dispatch(&hdr, data);
	  if ( rc && rc != GRR_UNKNOWN){
	    char line[256];
	    sprintf(line,"L2FormArte/usevnt: rpm_dispatch complaines about msg type 0x%x",hdr.type);
	    ARTE_error(GRR_SERROR,line,rc);
	  }
	}
      } // End of L2_CnA online mode.
    
    if(!L2_CnA_FlagIsSet)   // No L2_CnA
      {
	/** 
	 *
	 * The non-L2_CnA mode is used offline, both by SLTsim (Standalone==0) 
	 * and in Standalone mode (for offline CnA test).
	 *
	 * The DB table versions used in NO L2_CnA mode are handled
	 * in the following way:
	 * - In SLTsim mode, the CnAkey is obtained from the global 
	 * run cards in the kumac ("/CNA/RUN2CNAKEY [runnumber]" and 
	 * "/CNA/READ CNA ON"). If "runnumber" is <0, then it is expected 
	 * to correspond to an existing CnAKey and treated as such.  
	 * The EDC Trig Type is set accordingly in get_major_minor. 
	 * Details are given in that function. 
	 * - In StandAlone mode, the CnAkey is obtained in first priority 
	 * via the kumac as for previous point, or via the "-cna_no" parameter on 
	 * the command line, or from KeyLastMajor(0) if nothing else was set.  
	 *
	 * The L2 geometry is generated according to the CnAkey 
	 * (alignment correction)  
	 * or directly from the ARTE-GEO in mcmode.
	 *
	 * The various L2Form handlers (for L2Symbol, L2Trigger, L2Geo, 
	 * L2BuffMap and L2OtrCalib tables) are called explicitely in the 
	 * non-L2_CnA mode.
	 *
	 **/ 
	
	if (StandAlone==1)
	  { //Standalone mode enabled via -L4Odummy run par
	    // CnA key from kumac runnumber or from -cna_no:<RunNO>
	    if ((CnA_K=CnAkey)<=0)  
	      if ((CnA_K=cna_no)<=0) CnA_K = KeyLastMajor(0);
	    if (KeyBookWasCalled == 1)
	      { 
		KTstatus = KeyLoadTables(CnA_K);
		sprintf(mess,"L2FormArte: got KTstatus=%d from CnAkey=%d (StandAlone mode)",
			KTstatus,CnA_K);
		ARTE_error(GRR_SINFO,mess,0);
	      } 
	    else
	      {
		sprintf(mess,
		   "L2FormArte: KEYBOOK not specified, Please check your kumac! (StandAlone mode)");
		ARTE_exit(GRR_SFATAL,mess,0);
	      }
	  } 
	else
	  { //SLTsim mode (default of non-L2_CnA)
	    KTstatus = 1;
	    if ((CnA_K=CnAkey)<=0)
	      {
		sprintf(mess,
			"L2FormArte/SLTsim: Unvalid CnAKey given (CnAkey = %d) EXIT ARTE!",CnA_K);
		ARTE_exit(GRR_SFATAL,mess,0);
	      }
	    else
	      {
		sprintf(mess,"L2FormArte: got KTstatus=%d from CnAkey=%d (SLTsim mode)",KTstatus,CnA_K);
		ARTE_error(GRR_SINFO,mess,0);
	      }
	  }
	sprintf(Key,"%d",CnA_K);
	// Call handlers
	// L2Symbol and L2Trigger 
	tfmt_hdl (NULL, KTstatus, NULL, 0, NULL);
	// L2Geo and L2BuffMap 
	l2fmt_hdl (NULL,KTstatus, NULL, 0, NULL);
	
	// The following part should be hidden in otrslt: in SLTsim mode (Standalone==0) 
	// the otrslt code is not opening the pipe by itself, so it has to be done here. 
	// In Standalone mode (Standalone==1), the OtrCalib table needs to be explicitely dumped. 
	if (StandAlone==0) {// Pipe only in SLTsim mode
	  //L2OtrCalib
	  //open pipe 
	  if (l2form_PipeInit() != 0)
	    {
	      sprintf(mess,
		      "L2FormArte/usevnt: Problems initializing the pipe for OTR-CALIB\n");
	      ARTE_error(GRR_SFATAL,mess,0);
	    }
	  //invoke handler 
	  sprintf(mess,
		  "L2FormArte: Now calling the formatter for OTR-CALIB (pipe mode)\n");
	  ARTE_error(GRR_SINFO,mess,0);
	  otrsltWriteToFD(l2io_fd);
	  close(l2io_fd);
	}
	else
	  {
	    char fName[100];
	    int dump=0;
	    for( int i=1; i<nArg; i++ ){
	      if( !strncmp( vArg[i], "-sltdumpdir:", 12 ) ){
		dump++;
		strcpy(fName, vArg[i]+12);
		sprintf(mess,
			"\nL2FormArte/otrcalib_hdl: dumping the OTRiSltCalib (l2in.otr) file to %s\n",
			fName);
		ARTE_error(GRR_SINFO,mess,0); 
	      } 
	    }
	    if (dump == 0)
	      {
		sprintf(mess,
			"\nL2FormArte/otrcalib_hdl: Couldn't read address of sltdumpdir directory\n");
		ARTE_error(GRR_SERROR,mess,0);
	      }
	    else
	      { 
		strcat(fName,"l2in.otr_");
		strcat(fName,Key);
		otrsltWriteToFile(fName);
	      }  
	  }
      } // End of no L2_CnA.
  } // End of L2FormArte.
  
  /*****************
   *
   * formatter_l2_init:
   * Formatter user initialization code for the l2geo table
   * (standard CIF used for CnA)
   *
   ***/
  
  int formatter_l2_init(void){
    int rc;
    rc = CIF_add_formatted_tab ("l2tab",    // name of formatted table
    				"name:0B",  // rpm format 
    				&l2fmt_sub);// subscriptin handler 
    if (rc) grr_error(GRR_SERROR,"could not subscribe to ftab (rc=%d)\n",rc);
    return rc;
    
  }
  
  /*****************
   *
   * formatter_trigger_init:
   * Formatter user initialization code for the trigger table
   * (standard CIF used for CnA)
   *
   ***/
  
  int formatter_trigger_init(void){
    int rc;
    rc = CIF_add_formatted_tab ("ttab",     // name of formatted table
                                "name:0B",  // rpm format 
                                &tfmt_sub); // subscriptin handler 
    if (rc) grr_error(GRR_SERROR,"could not subscribe to ftab (rc=%d)\n",rc);
    return rc;
    
  }

  /*****************
   *
   * l2fmt_sub:
   * l2 formatted table subscription handler 
   * (maps formatted tables to raw tables)
   *
   ***/
  
  int l2fmt_sub(const char *tab,int id,CIF_format_tab **formatter){
    int rc = GRR_OK;
    int i;
    
    for (i=0;i<N_RTABS-2;i++) {
      if ((rc = CIF_formatter_need_raw(id, tablename[i], &l2_rawtabid[i])) 
	  != GRR_OK) return rc; 
    }
    
    *formatter = &l2fmt_hdl;
    ftabid  = id;
    
    return rc;
  }
  
  /*****************
   *
   * tfmt_sub:
   * trigger formatted table subscription handler
   * (maps formatted tables to raw tables)
   *
   ***/
  
  int tfmt_sub(const char *tab,int id,CIF_format_tab **formatter){
    int rc = GRR_OK;
    int i;
    
    for (i=N_RTABS-2;i<N_RTABS;i++) {
      if ((rc = CIF_formatter_need_raw(id, tablename[i], &t_rawtabid[i]))
          != GRR_OK) return rc;
    }
    
    *formatter = &tfmt_hdl;
    ttabid  = id;
    
    return rc;
  }
  

  /*****************
   *
   * l2fmt_hdl:
   * l2 formatter handler for L2Geo and l2BuffMap tables  
   *
   ***/
  
  int l2fmt_hdl(const CIF_raw_tab *urt, int N_raw, 
		CIF_formatted_tab *uft, int N_max, int *N_updated)
  {
    static int L2_CnA_FlagIsSet=0, CnA_Debug_FlagIsSet=0;
#ifdef L2_CnA
    L2_CnA_FlagIsSet=1;
#endif
#ifdef DEBUG_CNA
    CnA_Debug_FlagIsSet=1;
#endif
    int bufflength1=0;
    int bufflength2=0;
    char mess[200];
    int flagVDS    = 0;
    int flagGLOBAL = 0;
    int flagITR = 0;
    int flagOTR = 0;
    int flagMUON = 0;
    unsigned int alcaUpdate = 0;
    int * l2formgeo = NULL;
    int * l2formmap = NULL;
    int * l2formtab = NULL;
    int j, k; // L2_CNA variables (?)
    
    if (l2formgeo) free(l2formgeo);
    if (l2formmap) free(l2formmap);
    if (l2formtab) free(l2formtab);
    
    if ( N_raw>0 )
      {
	if(L2_CnA_FlagIsSet) //CnA Online mode
	  {
	    *N_updated = 0;
	    if ( N_max < 1 ) return GRR_FISHY; 
	    
	    if(CnA_Debug_FlagIsSet) // Debugging!
	      {
		for(j=0;j<N_raw;j++)
		  {        // loop over the updated rtabs 
		    for(k=0;k<N_RTABS-2;k++)
		      {  // loop over all rtabs to identify the ID 
			if ( urt[j].id == l2_rawtabid[k])
			  {
			    sprintf(mess,
				    "L2FormArte/l2fmt_hdl: raw table name:%s, id:%d, major:%d, minor:%d\n",tablename[k],l2_rawtabid[k],urt[j].version[0],urt[j].version[1]);
			    grr_error(GRR_SINFO,mess,0);
			  }
		      }
		  }
	      }// End of debugging.
	  } // End of CNA Mode.
	if(!L2_CnA_FlagIsSet)L2FormARTE_KTstatus = 2;

	// Put comment here
	// Generation of formatted tables for L2 
	if (mcflag == 1)
	  {
	    if(l2GeoFormat(&l2formgeo,&bufflength1)!=0)
	      {
		sprintf(mess,
		       "L2FormArte/l2fmt_hdl: Problems with L2GeoFormatter\n");
		grr_error(GRR_SFATAL,mess,1);
		return GRR_FISHY;
	      }
	    else
	      {
		sprintf(mess,
			"L2FormArte/l2fmt_hdl: l2geo tables generated with length %d\n",bufflength1);
		grr_error(GRR_SINFO,mess,0);
	      }
	  }  
	if (StandAlone==1)
	  {
	    if(L2FormARTE_KTstatus >0)
	      {
		flagVDS = acReadVDS(L2FormARTE_KTstatus);
		flagITR = acReadITR(L2FormARTE_KTstatus);
		flagOTR = acReadOTR(L2FormARTE_KTstatus);
		flagMUON = acReadMuon(L2FormARTE_KTstatus);
		flagGLOBAL = acReadGlobal(L2FormARTE_KTstatus);
	      } else
		{
		  sprintf(mess,
			  "L2FormArte_handler:  L2FormARTE_KTstatus=%d !!! Something fishy !!!!\n",L2FormARTE_KTstatus);
		  grr_error(GRR_SINFO,mess,0);
		}
	    if (flagVDS    >= 1) alcaUpdate |=  COMP_M_SVD;
	    if (flagITR    >= 1) alcaUpdate |=  COMP_M_ITR;
	    if (flagOTR    >= 1) alcaUpdate |=  COMP_M_OTR;
	    if (flagMUON   >= 1) alcaUpdate |=  COMP_M_MUON;
	    if (flagGLOBAL >= 1) alcaUpdate |=  COMP_M_GLOBAL;
	    
	    cout << "calling GEDE update for global alignment !!!!" << endl;
	    // Global alignment of GEDE's
	    if (alcaUpdate != 0) acGlobalAlignmentForGEDE(alcaUpdate);
	  }
	if (mcflag == 0)
	  {
	    if(l2GeoFormat(&l2formgeo,&bufflength1)!=0)
	      {
		sprintf(mess,
			"L2FormArte/l2fmt_hdl: Problems with L2GeoFormatter\n");
		grr_error(GRR_SFATAL,mess,1);
		return GRR_FISHY;
	      }
	    else
	      {
		sprintf(mess,
			"L2FormArte/l2fmt_hdl: l2geo tables generated with length %d\n",bufflength1);
		grr_error(GRR_SINFO,mess,0);
	      }
	  }
	
	// The following is a temporary hack for SLTsim: DB table versions
	// from OtrRunnumber (acread) are used if the global kumac 
	// run-number is nor valid

	if (mcflag==1 && StandAlone == 0 && CnAkey<=0)
	  {
	    sprintf(mess,
		    "L2FormArte/l2fmt_hdl: WARNING: CnAKey not valid, acRead use key from OtrEventnumber for l2buffmap");
	    ARTE_error(GRR_SINFO,mess,0);
	    L2FormARTE_KTstatus = 2; 
	    flagVDS = acReadVDS(L2FormARTE_KTstatus);
	    flagITR = acReadITR(L2FormARTE_KTstatus);
	    flagOTR = acReadOTR(L2FormARTE_KTstatus);	
	    flagMUON = acReadMuon(L2FormARTE_KTstatus);	
	  }
	
	if (l2MapFormat(&l2formmap,&bufflength2)!=0)
	  {
	    sprintf(mess,
		    "L2FormArte/l2fmt_hdl: Problems with L2MapFormatter\n");
	    grr_error(GRR_SFATAL,mess,1);
	    return GRR_FISHY;
	  }
	else
	  {
	    sprintf(mess,
	       "L2FormArte/l2fmt_hdl: l2map table generated with length %d\n",
		    bufflength2);
	    grr_error(GRR_SINFO,mess,0);
	  }
	
	l2formtab = (int *) malloc(bufflength1+bufflength2);
	if(l2formtab==NULL)
	  {
	    sprintf(mess,
	     "L2FormArte/l2fmt_hdl: Problems to allocate memory for ftabs\n");
	    grr_error(GRR_SFATAL,mess,1);
	    return GRR_FISHY;
	  }
	else
	  {
	    memcpy(l2formtab,l2formgeo,bufflength1);
	    memcpy((char*)l2formtab+bufflength1,l2formmap,bufflength2);
	  }
	if(L2_CnA_FlagIsSet) //CnA Online Mode
	  {
	    uft[0].id   = ftabid;
	    uft[0].data = l2formtab;
	    uft[0].len  = bufflength1+bufflength2;
	    *N_updated  = 1;
	  }
	if(!L2_CnA_FlagIsSet) //NO L2_CnA
	  {
	    if (StandAlone==0)
	      {
		// Pipe l2tabs in SLTsim mode only
		if (l2form_PipeInit() != 0)
		  return GRR_FISHY;
	
		write(l2io_fd,
		      (char *)l2formgeo,
		      sizeof(L2FORMGEOHEAD) );
		
		int geosize = bufflength1-sizeof(L2FORMGEOHEAD);
		int parts = geosize/MAXPIPE;
		int tail = geosize%MAXPIPE;
		
		printf("L2FormArte/l2fmt_hdl: Piping l2formgeo (length %d)\n", geosize);
		for (int part = 0; part < parts; part++) {
		  write(l2io_fd,
			(char *)l2formgeo+sizeof(L2FORMGEOHEAD)+part*MAXPIPE,
			MAXPIPE);
		}
		write(l2io_fd,
		      (char *)l2formgeo+sizeof(L2FORMGEOHEAD)+parts*MAXPIPE,
		      tail );
		
		write(l2io_fd,
		      (char *)l2formmap,
		      sizeof(L2FORMMAPHEAD) );
		
		int length_bm = bufflength2-sizeof(L2FORMMAPHEAD);
		parts = length_bm/MAXPIPE;
		tail = length_bm%MAXPIPE;
		
		printf("L2FormArte/l2fmt_hdl: Piping l2formmap (total length %d)\n", length_bm);
		for (int part = 0; part < parts; part++) {
		  write(l2io_fd,
			(char *)l2formmap+sizeof(L2FORMMAPHEAD)+part*MAXPIPE,
			MAXPIPE);
		}
		write(l2io_fd,
		      (char *)l2formmap+sizeof(L2FORMMAPHEAD)+parts*MAXPIPE,
		      tail );
		
		close(l2io_fd);
		printf("L2FormArte/l2fmt_hdl: Piping l2formgeo/l2formmap done ! \n");
	      }
	  }
	//#endif //NO L2_CnA
	free(l2formgeo);
	free(l2formmap);
      }
    return GRR_OK;
  }
  
  /*****************
   *
   * tfmt_hdl:
   * l2 formatter handler for L2Symbol and L2Trigger tables  
   *
   ***/
  
  int tfmt_hdl (const CIF_raw_tab *urt, int N_raw,
		CIF_formatted_tab *uft, int N_max, int *N_updated)
  {
    static int L2_CnA_FlagIsSet=0, CnA_Debug_FlagIsSet=0;
#ifdef L2_CnA
    L2_CnA_FlagIsSet=1;
#endif
#ifdef DEBUG_CNA
    CnA_Debug_FlagIsSet=1;
#endif

    int bufflength1, bufflength2, j, k, length_st, irun=0;
    char mess[200];
    int * l2formtt = NULL;
    int * l2formst = NULL;
    int * l2formtab = NULL;
    
    L2FORMTHEAD* header2;
    Trig_Sym_major_minor ts;
    L2SYMBOL *theSym;
    char *rdPtr;
    
    
    if(l2formtt) free(l2formtt);
    if(l2formst) free(l2formst);
    if(l2formtab) free(l2formtab);
    
    if ( N_raw>0 )
      {
	if(!L2_CnA_FlagIsSet)
	  { // No L2_CnA.
            // For 2000 analysis purpose, use get_major_minor lookup, 
	    // since the trig/symb tables were not part of keytable
	    if ((((irun=RunNoToSimu)>0 || 
		  (irun=kasltsim::get_runnumber())>0) && 
		 irun<=DB_START_RUN))
	      {
		ts=get_major_minor(irun);
		if ((ts.tt_major == 0) && 
		    (ts.tt_minor == 0) && 
		    (ts.st_major == 0) && 
		    (ts.st_minor == 0))
		  {
		    if (StandAlone == 0)
		      {
			sprintf(mess,"L2FormArte/tfmt_hdl: 0's return, INVALID Run Number for 2000 trig tables %d, exit !",irun);
			ARTE_exit(GRR_SFATAL,mess,0);
		      }
		    else
		      {
			sprintf(mess,"L2FormArte/tfmt_hdl WARNING: generated Trigger/Symbol tables do NOT correspond to 2000 run number %d",irun);
			ARTE_error(GRR_SERROR,mess,0);
			ts.tt_major = 1;
			ts.tt_minor = 0;
			ts.st_major = 1;
			ts.st_minor = 0;
		      }
		  }
	      }
	    else if (CnA_K > DB_START_KEY)
	      { 
		// rely on keytable for major/minor of new runs
		ts.tt_major = -1;
		ts.tt_minor = -1;
		ts.st_major = -1;
		ts.st_minor = -1;
	      }
	    else
	      {
		if (StandAlone == 0)
		  {
		    sprintf(mess,"L2FormArte/tfmt_hdl: 0's return, INVALID Run Number for trig tables (irun %d, CnA_K %d), exit !",irun,CnA_K);
		    ARTE_exit(GRR_SFATAL,mess,0);
		  } 
	      }
	    char mess1[200];
	    sprintf(mess1,"L2FormArte/tfmt_hdl: tt_major=%d tt_minor=%d\n",ts.tt_major,ts.tt_minor);
	    grr_error(GRR_SINFO,mess1,0);
	    sprintf(mess1,"L2FormArte/tfmt_hdl: st_major=%d st_minor=%d\n",ts.st_major,ts.st_minor);
	    grr_error(GRR_SINFO,mess1,0);
	  }
	if(L2_CnA_FlagIsSet)
	  { // L2_CnA mode.
	    *N_updated = 0;
            if ( N_max < 1 ) return GRR_FISHY;
            if(CnA_Debug_FlagIsSet)
	      {
		sprintf(mess,"tfmt_hdl: Number of updated raw tables:%d\n",N_raw);
		grr_error(GRR_SINFO,mess,0);
	      }
	    // loop over the updated rtabs
	    for(j=0;j<N_raw;j++)
	      {  
		// loop over all rtabs to identify the ID
		for(k=N_RTABS-2;k<N_RTABS;k++)
		  {  
		    if(urt[j].id == t_rawtabid[k])
		      {
			if(1||CnA_Debug_FlagIsSet)
			  {
			    sprintf(mess,"Yo, Kendall, L2FormArte/tfmt_hdl: raw table name:%s, id:%d, major:%d, minor:%d\n",tablename[k],t_rawtabid[k],urt[j].version[0],urt[j].version[1]);
			    grr_error(GRR_SINFO,mess,0);
			  }
		      }
		  }
	      }
	    ts.tt_major = -1;
	    ts.tt_minor = -1;
	    ts.st_major = -1;
	    ts.st_minor = -1;
	  }//End of L2_CnA
	// Generation of formatted tables for L2
	if(L2TTFormat(&l2formtt,&bufflength1,ts.tt_major,ts.tt_minor)!=0)
	  {
	    sprintf(mess,"tfmt_hdl: Problems with L2TTFormatter\n");
	    grr_error(GRR_SFATAL,mess,1);
	    return GRR_FISHY;
	  }
	else
	  {
	    sprintf(mess,
		    "tfmt_hdl: trigger table generated with length %d\n",
		    bufflength1);
	    grr_error(GRR_SINFO,mess,0);
	  }
	if (L2STFormat(&l2formst,&bufflength2,ts.st_major,ts.st_minor)!=0)
	  {
	    sprintf(mess,"tfmt_hdl: Problems with L2STFormatter\n");
	    grr_error(GRR_SFATAL,mess,1);
	    return GRR_FISHY;
	  }
	else
	  {
	    sprintf(mess,"tfmt_hdl: l2symbol table generated with length %d\n",
		    bufflength2);
	    grr_error(GRR_SINFO,mess,0);
	  }
	
	if(CnA_Debug_FlagIsSet)
	  {
	    // TESTING  that the table is properly filled 
	    header2 = (L2FORMTHEAD*) l2formst;
	    rdPtr = (char*) (header2 + 1);
	    length_st = header2->nentries*sizeof(L2SYMBOL);
	    theSym = (L2SYMBOL*)malloc(length_st);
	    memcpy(theSym, (L2SYMBOL*) rdPtr,length_st);
	    L2TS_WriteToFile(0,theSym, header2->nentries);  
	    free(theSym);
	  }
	
	l2formtab = (int *) malloc(bufflength1+bufflength2);
	if(l2formtab==NULL)
	  {
	    sprintf(mess,"tfmt_hdl: Problems to allocate memory for ftabs\n");
	    grr_error(GRR_SFATAL,mess,1);
	    return GRR_FISHY;
	  }
	else
	  {
	    memcpy(l2formtab,l2formtt,bufflength1);
	    memcpy((int*)((char*)l2formtab+bufflength1),l2formst,bufflength2);
	  }
      }
    
    if(L2_CnA_FlagIsSet)
      {
	uft[0].id   = ttabid;
	uft[0].data = l2formtab;
	uft[0].len  = bufflength1+bufflength2;
	*N_updated  = 1;
      }
    
    if(!L2_CnA_FlagIsSet)
      {
	if (StandAlone==0)
	  {
	    // Do the piping of l2tabs here
	    if (l2form_PipeInit() != 0) return GRR_FISHY;
	    write(l2io_fd,(char *)l2formtt,sizeof(L2FORMTHEAD));
	    header2 = (L2FORMTHEAD *) l2formtt;
	    int ttsize = sizeof(L2TENTRY) * header2->nentries;
	    printf("L2FormArte/tfmt_hdl: Piping l2formtt with length %d.\n", 
		   ttsize);
	    write(l2io_fd,(char *)(header2+1),ttsize);
	    write(l2io_fd,(char *)l2formst,sizeof(L2FORMTHEAD));
	    header2 = (L2FORMTHEAD *) l2formst;
	    int stsize = sizeof(L2SYMBOL) * header2->nentries;
	    printf("L2FormArte/tfmt_hdl: Piping l2formst with length %d.\n", 
		   stsize);
	    write(l2io_fd,(char *)(header2+1),stsize);
	    close(l2io_fd);
	    printf("L2FormArte/tfmt_hdl: Piping l2formtt/l2formst done ! \n");
	  } // End of StandAlone==0 case.
      } // End of No L2_CnA case.
    free(l2formtt);
    free(l2formst);
    return GRR_OK;
  }
}
