/************************************************************
  AUTHOR       :: P. Kreuzer
  CREATED      :: 15/03/00        
  LAST MODIFIED:: 18/03/00
               :: 28/05/01 : commented out all "targ" functions
               ::            because new l2libs don't contain
               ::            l2targ.c anymore AND because the 
               ::            target information is irrelevant 
               ::            in the L2Geo file
  DESCRIPTION  :: L2GEO-formatter (l2GeoFormat) and code to generate   
               :: the l2geo table. The table is composed of
               :: SUPERLAYER and PLANE structures.
               :: (CnA, Binary and ASCii modes are foreseen)
  DOMAIN       :: L2FormArte       
  ************************************************************/

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <strings.h>
#include <sys/stat.h>

#include "arte/arte.hh"

extern "C" {
#include "pltable.h"
#include "sltable.h"
#include "l2format.h"
#include "rpm.h"
#include "grr.h"
#include "pub.h"
#include "cif.h"
#include "l2geo.h"
#include "errno.h"
extern char Key[20];
extern int errno;
}
#include "farm/farm.hh"
#include "database/keytable.h"
int acReadVDS(int);
void acReadGlobal(int KTstatus);
extern int  nArg;
extern char **vArg;
extern "C" 
{
  static void FillSLTable(void);
  static void FillPlTable(void);
  static void FillGeoTable(void);
  static int MakeL2Geo(int ** l2geo,int * length);
  static int Write_GeoAscii(void);
  static int Write_GeoBinary(int *l2geo, int size);
  
  

  /*****************
   *
   * l2GeoFormat:
   * Function called by the l2geo_fmt handler  
   * to generate l2geo formatted table for the SLP. 
   * Options to produce ASCii and/or BINARY file dumps 
   * are available via compilation flags DUMP_ASCII/BINARY.
   * The DUMP_NODE flag should be used ONLY online !
   *
   ***/
  
  int l2GeoFormat(int ** l2geo,int * length)
    {
      int *l2formgeo;
      int size;
      char mess[200];
      char ModeName[10];

      /* Destructors */
      SL_destroy();
      Pl_destroy();
            
      /*  initialisation of geometry data */
      FillGeoTable(); 
      
      if(MakeL2Geo(&l2formgeo,&size)!=0){
	sprintf(mess,"L2FormArte/l2GeoFormat: MakeL2Geo failed !\n"); 
	grr_error(GRR_SFATAL,mess,1);
	return 1;
      }
      *l2geo = l2formgeo;
      *length = size;

      /* decode the SltDumpMode to be used for the L2tabs */
      for( int i=1; i<nArg; i++ ){
	if( !strncmp( vArg[i], "-sltdumpmode:", 13 ) ){
	  strcpy(ModeName, vArg[i]+13);
	  sprintf(mess,"L2FormArte/l2GeoFormat: sltdumpmode %s\n",ModeName);
	  grr_error(GRR_SINFO,mess,0); 
	} 
      }
      if (!strncmp(ModeName,"ASCII",5)){
	if (Write_GeoAscii()!=0) {
	  sprintf(mess,"L2FormArte/l2GeoFormat: Problems to write ASCii file (l2in.geo)"); 
	  grr_error(GRR_SERROR,mess,0);
	} else {
	  sprintf(mess,"L2FormArte/l2GeoFormat: ASCii file (l2in.geo ) generated"); 
	  grr_error(GRR_SINFO,mess,0);
	}
      } else if (!strncmp(ModeName,"BINARY",6)){
	if (Write_GeoBinary(l2formgeo,size)!=0) {
	  sprintf(mess,"L2FormArte/l2GeoFormat: Problems to write Binary file (l2bin.geo)"); 
	  grr_error(GRR_SERROR,mess,0);
	} else {
	  sprintf(mess,"L2FormArte/l2GeoFormat: Binary file (l2bin.geo ) generated"); 
	  grr_error(GRR_SINFO,mess,0);
	}
      } else {
	sprintf(mess,"L2FormArte/l2GeoFormat: NO l2geo File will be dumped !!!\n");
	grr_error(GRR_SINFO,mess,0);
      }
      return 0;
    }


  /*****************
   * 
   * Create SUPERLAYER and PLANE structures,  
   * based on GTAR, GESL and GEDE Arte-tables
   *
   ***/

  static void FillGeoTable()
    {
      
      /* Fill in the various l2 geometry tables */
      
      FillSLTable();
      FillPlTable();
      
    }  

      
  /*****************
   *
   * Create l2geo structure used by formatter
   *
   ***/

  static int MakeL2Geo(int ** l2geo,int * length)
    {
      L2FORMGEOHEAD *header;
      int nsuplrs;
      int nplanes;
      int length_sl;
      int length_pl;
      char *wrPtr;
      char mess[200];
      int flag = 1;

      nsuplrs = SLGetNbEntries();
      nplanes = PlGetNbEntries();
      /* Allocate memory for CnA-formatted tables */
      length_sl=nsuplrs*sizeof(SUPERLAYER);
      length_pl=nplanes*sizeof(PLANE);
      *l2geo = (int *) malloc(sizeof(L2FORMGEOHEAD)+
			      (length_sl+length_pl));
      *length = sizeof(L2FORMGEOHEAD)+length_sl+length_pl;
      /* Call formatters */
      sprintf(mess,"L2FormArte/MakeL2Geo: nsuplrs=%d, nplanes=%d (length=%d)\n",
	      nsuplrs,nplanes,*length); 
      grr_error(GRR_SINFO,mess,0);
      if (l2geo != NULL) {
	header = (L2FORMGEOHEAD *) *l2geo;
	header->ntargs = 0;
	header->nsuplrs = nsuplrs;
	header->nplanes = nplanes;
	wrPtr = (char*) (header + 1);
	SL_formatCNA((SUPERLAYER*) wrPtr);
	wrPtr += length_sl;
	Pl_formatCNA((PLANE*) wrPtr);
	wrPtr += length_pl;
	flag = 0;
      } else {
	sprintf(mess,"L2FormArte/MakeL2Geo: couldn't allocate space for l2geo table\n"); 
	grr_error(GRR_SFATAL,mess,0);
	flag = 1;
      }
      return flag;
    }


  /*****************
   *
   * Function to dump l2geo table to ASCii file 
   *
   ***/
  
  static int Write_GeoAscii() 
  {
    FILE *fdout;
    char fName[100];
    char mess[200];
    int flag = 1;
    int dump = 0;
    
    for( int i=1; i<nArg; i++ ){
      if( !strncmp( vArg[i], "-sltdumpdir:", 12 ) ){
	dump++;
	strcpy(fName, vArg[i]+12);
	sprintf(mess,"L2FormArte/Write_GeoAscii: dumping the L2Geo ASCII file (l2in.geo) to %s\n",fName);
	grr_error(GRR_SINFO,mess,0); 
      } 
    }
    if (dump == 0) {
      sprintf(mess,"L2FormArte/Write_GeoAscii: Couldn't read address of sltdumpdir directory\n");
      grr_error(GRR_SERROR,mess,0);
    } 
    strcat(fName,"l2in.geo_");
    strcat(fName,Key);
    umask(0111);

    /*    open file for geometry information: ---------------- */
    fdout = fopen(fName, "w");
    if( fdout == NULL) {
      sprintf(mess,"L2FormArte/Write_GeoAscii: unable to open %s\n",fName); 
      grr_error(GRR_SERROR,mess,1);
      sprintf(mess,"L2FormArte/Write_GeoAscii: error %s\n",strerror(errno)); 
      grr_error(GRR_SERROR,mess,1);
      flag = 1;
    } else {
      SL_PrintTable(fdout);
      Pl_PrintTable(fdout);
      fclose(fdout);
      flag = 0;
    }
    return flag;
  }
  
  
  /*****************
   *
   * Function to dump l2geo table to Binary file 
   *
   ***/
  
  static int Write_GeoBinary(int *l2geo, int size) 
  {
    FILE *fdout;
    char fName[100];
    char mess[200];
    int flag = 1;
    int dump = 0;

    for( int i=1; i<nArg; i++ ){
      if( !strncmp( vArg[i], "-sltdumpdir:", 12 ) ){
	dump++;
	strcpy(fName, vArg[i]+12);
	sprintf(mess,"L2FormArte/Write_GeoBinary: dumping the L2Geo BINARY file (l2bin.geo) to %s\n",fName);
	grr_error(GRR_SINFO,mess,0); 
      } 
    }
    if (dump == 0) {
      sprintf(mess,"L2FormArte/Write_GeoBinary: Couldn't read address of sltdumpdir directory\n");
      grr_error(GRR_SERROR,mess,0);
    } 
    strcat(fName,"l2bin.geo_");
    strcat(fName,Key);
    umask(0111);
    /*    open file for geometry information: ---------------- */
    fdout = fopen(fName, "wb");
    if( fdout == NULL) {
      sprintf(mess,"L2FormArte/Write_GeoBinary: unable to open %s\n",fName); 
      grr_error(GRR_SERROR,mess,1);
      sprintf(mess,"L2FormArte/Write_GeoBinary: error %s\n",strerror(errno)); 
      grr_error(GRR_SERROR,mess,1);
      flag = 1;  
    } else {
      fwrite(l2geo, size, 1, fdout);
      fclose(fdout);
      flag = 0;
    }
    return flag;
  }
  
}


  /*****************
   *
   * Function to copy GESL table into l2 SUPERLAYER structure
   *
   ***/

static void FillSLTable(void)
{
   SUPERLAYER *sl;
   char mess[200];

  for (ArteTableIterator<GESL> gesl; gesl.valid(); gesl.next())
  {

/*  allocate memory for superplane structure: ------------ */
    sl = (SUPERLAYER *) (malloc(sizeof(SUPERLAYER)) );
    if (sl==NULL) {
      sprintf(mess,"L2FormArte/FillSLTable: no room for superlayer structure\n"); 
      grr_error(GRR_SFATAL,mess,1);
      exit(EXIT_FAILURE);
    }

/*  copy ARTE data in l2simu C structure: ---------------- */
    sl->l2nsl = gesl.index();
    strncpy(sl->name,gesl->name,4);
    /*    sl->name= gesl->name;*/
    sl->cmp = gesl->cmp;
    sl->nsl = gesl->nsl;
    sl->mfl = gesl->mfl;
    sl->lay = gesl->lay;
    /* sl->xo = gesl->xo; no longer in the arte bank
       sl->yo = gesl->yo; */
    sl->xo = 0;
    sl->yo = 0;
    sl->z1 = gesl->z1;
    sl->z2 = gesl->z2;
    sl->npl = gesl->npl;
    sl->fpl = gesl->fpl;

/*  Insert superplane in superplane table: --------------- */
    SLInsert( sl );

#if 0
    cout <<"SUPERLAYER "<< " " << gesl->name << " " << gesl.index() <<" "<< gesl->cmp <<" "<< gesl->nsl <<" " 
 << gesl->lay <<" "<< gesl->fpl << endl;
#endif

  }
}


  /*****************
   *
   * Function to copy GEDE table into l2 PLANE structure
   *
   ***/

static void FillPlTable(void)
{
   PLANE *pl;
   char imod[3];
   char mess[200];
   char imuon[3];
   char jmuon[2];
   char pad_id[2];

  for (ArteTableIterator<GEDE> gede; gede.valid(); gede.next())
  {
    /* consider only planes with existing hardware implementation (PK)
       ...sounds a bit funny, but for the OTR the simulation seems to have
       more gede's defined as existing in a full superlayer..these have sec=0  */
    ArtePointer<NEDE> nede (gede.index());
    if (strncmp(gede->hwc,"none",4)==0) continue;
    /*    if (gede->sec == 0) continue; */

/*  allocate memory for plane structure: ----------------- */
    pl = (PLANE *) (malloc(sizeof(PLANE)) );
    if (pl==NULL) {
      sprintf(mess,"L2FormArte/FillPlTable: no room for plane structure\n"); 
      grr_error(GRR_SFATAL,mess,1);
      exit(EXIT_FAILURE);
    }

/*  copy ARTE data in l2simu C structure: ---------------- */
    pl->l2npl = gede.index();
    pl->l2nsl = gede->gesl.index();
    pl->cmp = gede->cmp;
    pl->lay = gede->lay;
    pl->sec = gede->sec & 0xFFFF;
    pl->mod = gede->sec >> 16;
    if (gede->cmp == 2) {
      strncpy(imod,&gede->hwc[6],2);
      strcpy(&imod[2],"\0");
      pl->mod = atoi(imod);
    }

/* Define mod of the muon from the hwc. jmuon indicates the half (+y/-y) in
   each layer and imuon the element */

    if (gede->cmp == 8){
      strncpy(imuon, &gede->hwc[6],2);
      strncpy(jmuon, &gede->hwc[4],1);
      pl->mod = atoi(jmuon)*100.;
      pl->mod = atoi(imuon) + pl->mod;
      pl->sec = 0;
    }
    if (gede->cmp == 9){
      strncpy(jmuon, &gede->hwc[4],1);
      pl->mod = atoi(jmuon);
      pl->lay = 1;
      pl->sec = 0;
    }

    ArtePointer<GESL> gesl (gede->gesl);
    if (gede->cmp == 3) {
      if (strncmp(gesl->name,"MS14",4)!=0 && strncmp(gesl->name,"MS15",4)!=0) {
	if ((gede->sec == 1) || (gede->sec == 3)) pl->mod = 1;
	if ((gede->sec == 6) || (gede->sec == 8)) pl->mod = -1;
	if ((gede->sec == 2) || (gede->sec == 4)) pl->mod = 2;
	if ((gede->sec == 5) || (gede->sec == 7)) pl->mod = -2;
      } else {
	if ((gede->sec == 1) || (gede->sec == 3)) pl->mod = 10;
	if ((gede->sec == 6) || (gede->sec == 8)) pl->mod = -10;
	if ((gede->sec == 2) || (gede->sec == 4)) pl->mod = 20;
	if ((gede->sec == 5) || (gede->sec == 7)) pl->mod = -20;
      }
    }
    pl->shp = gede->shp;
    pl->xmin = gede->xmin;
    pl->xmax = gede->xmax;
    pl->ymin = gede->ymin;
    pl->ymax = gede->ymax;
    pl->zmin = gede->zmin;
    pl->zmax = gede->zmax;
    pl->sxmn = gede->sxmn;
    pl->sxmx = gede->sxmx;
    pl->symn = gede->symn;
    pl->symx = gede->symx;
    pl->cos = gede->cos;
    pl->sin = gede->sin;
    pl->rot = gede->rot;
    if (pl->cmp ==3) {
      pl->rot = nede->rot;
    }
    pl->acc = gede->acc;
    pl->eff = gede->eff;
    pl->dod = gede->dod;
    pl->pit = gede->pit;
    pl->ms = gede->ms;
    pl->xfw = gede->xfw;
    pl->zfw = gede->zfw;
    pl->nwi = gede->nwi;
    pl->nsl = gede->nsl;


/*  add angle information for MSGCs and MUONs: --------------------- */
    if (pl->cmp ==3 || pl->cmp ==8 || pl->cmp ==9) {
        pl->cos = pl->rot[0];
        pl->sin = pl->rot[1];
    }

    /* for silicon, rotate into global hera-b coordinate system,
       exchange min and max if needed then reset the rot matrix
       to diagonal, since the plane is redefined in the overall
       system. The slt assumes the silicon is in the overall
       system, but the "new" geometry no longer does. I'm still
       assuming rotations only about z-axis. Hope that doesn't
       break down! 26.4.98 mm

       I think the xfw's are defined in the plane coordinate
       system so they shouldn't need to be modified. Not sure,
       but since the slt code doesn't use them, I'll be lazy.
    */

    if( pl->cmp == 1) {
      float xp_min = pl->xmin * pl->rot[0] + pl->ymin * pl->rot[1];
      float yp_min = pl->xmin * pl->rot[3] + pl->ymin * pl->rot[4];

      float xp_max = pl->xmax * pl->rot[0] + pl->ymax * pl->rot[1];
      float yp_max = pl->xmax * pl->rot[3] + pl->ymax * pl->rot[4];

      if(xp_min < xp_max) {
	pl->xmin = xp_min;
	pl->xmax = xp_max;
      } else {
	pl->xmin = xp_max;
	pl->xmax = xp_min;
      }

      if(yp_min < yp_max) {
	pl->ymin = yp_min;
	pl->ymax = yp_max;
      } else {
	pl->ymin = yp_max;
	pl->ymax = yp_min;
      }

      xp_min = pl->sxmn * pl->rot[0] + pl->symn * pl->rot[1];
      yp_min = pl->sxmn * pl->rot[3] + pl->symn * pl->rot[4];

      xp_max = pl->sxmx * pl->rot[0] + pl->symx * pl->rot[1];
      yp_max = pl->sxmx * pl->rot[3] + pl->symx * pl->rot[4];

      if(xp_min < xp_max) {
	pl->sxmn = xp_min;
	pl->sxmx = xp_max;
      } else {
	pl->sxmn = xp_max;
	pl->sxmx = xp_min;
      }

      if(yp_min < yp_max) {
	pl->symn = yp_min;
	pl->symx = yp_max;
      } else {
	pl->symn = yp_max;
	pl->symx = yp_min;
      }
    }
    if(pl->cmp == 1 || pl->cmp == 2) {
      float cos_tmp = pl->rot[0] * pl->cos + pl->rot[3] * pl->sin;
      float sin_tmp = pl->rot[1] * pl->cos + pl->rot[4] * pl->sin;
      
      pl->cos = cos_tmp;
      pl->sin = sin_tmp;
      
      pl->rot[0] = pl->rot[4] = pl->rot[8] = 1.;
      pl->rot[1] = pl->rot[2] = pl->rot[3] = pl->rot[5] = pl->rot[6] = pl->rot[7] = 0.;
    }

/*  Insert plane in plane table: ------------------------- */
    PlInsert( pl );

#if 0
   cout << "PLANE : " << pl->l2npl <<" "<< pl->l2nsl <<" "<< pl->cmp <<" "
   << pl->lay <<" "<< pl->sec <<" "
   << pl->shp <<" "<< pl->xmin <<" "<< pl->xmax <<" "<< pl->ymin <<" "
   << pl->ymax <<" "<< pl->zmin <<" "<< pl->zmax <<" "<< pl->sxmn <<" "
   << pl->sxmx <<" "<< pl->symn <<" "<< pl->symx <<" "<< pl->cos <<" "
   << pl->sin <<" "<< "ROT[]" <<" "<< pl->acc <<" "<< pl->eff <<" "
   << pl->dod <<" "<< pl->pit <<" "<< pl->ms <<" "<< "XFW" <<" "
   << "ZFW" <<" "<< "NWI" <<" "<< pl->nsl <<"\n";
   cout << "layer = " << pl->lay <<" "<< "xmin = " << pl->xmin << "\n";  
   cout <<"SUPERLAYER "<< gede.index() <<" "<< gede->cmp <<" "<< gede->lay <<" " 
   << gede->sec <<" "<< gede->shp <<" "<< gede->xmin <<"\n";
#endif
  }
}









