// *****************************************************************************
//
// source:    L3_Geometry.hh
//
// type:      include file
//
// created:   06 Dec 1999
//
// authors:   Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// changes:
//
// *****************************************************************************
#ifndef _L3_GEOMETRY_HH_
#define _L3_GEOMETRY_HH_
#include "L3_Defines.hh"
#include <vector>
#include "L3_Plane.hh"
#include "L3_Target.hh"
#include "L3_Window.hh"
#include "L3_RFBand.hh"

class L3_TargetStation {  //just a helper class, still very primitive
public:
  L3_TargetStation(){}
  L3_TargetStation(float z,int nr) : t_z(z),t_nr(nr) {}
  inline float z(){ return t_z; } //return mean z position of target station
  inline int nr(){ return t_nr; } //return number (1(@ larger z) or 2)
  inline L3_TargetStation& operator=(const L3_TargetStation& a){ t_z=a.t_z; t_nr=a.t_nr; return *this; }
private:
  float t_z;
  int t_nr;
};

class L3_Geometry {
public:
  static void initialize();
  static void update();
  static void updateTarget();

  inline static L3_Plane plane(const int&, const int&, const int&); //returns plane(SL,Sec,Lay)
  inline static const L3_Plane& getPlane(const int&, const int&, const int&); //returns plane(SL,Sec,Lay)
  inline static int getPlaneIndex(int sl,int sec,int firstORsecond,int i);
  inline static float SLz(const int&, const int&); //returns z-position of quadrant(SL,Sec)
  inline static const L3_Window& getSector(int sl,int sec); //returns VDS sector sec of superlayer sl 

  static const int nMaxTarget = 8;                        //maximal number of target wires in HERA-B 
  inline static void addTarget(const L3_Target&);  //add a target (for testing)  
  inline static int nTarget();                     //returns number of used targets
  inline static L3_Target target(const int n);     //returns n-th target
  inline static const L3_Target& getTarget(const int n);      //returns const reference to n-th target
  inline static const L3_Target& getTargetZ(const int n);      //returns const reference to n-th target

  inline static float xtarget();  //return (mean) x-position of target
  inline static float ytarget();  //return (mean) y-position of target
  inline static float ztarget();  //return (mean) z-position of target
  inline static Hep3Vector meantarget(); //returns mean target position
  inline static float minztarget(); //returns z-position of most downstream target 
  inline static float maxztarget(); //returns z-position of most upstream target 
  inline static float minytarget(); //returns smallest y-position of any target 
  inline static float maxytarget(); //returns largest y-position of any target
  inline static float minxtarget(); //returns smallest x-position of any target 
  inline static float maxxtarget(); //returns largest x-position of any target
  inline static float TargetBox_xmin(); // returns minimum x of target box constraint
  inline static float TargetBox_xmax(); // returns maximum x of target box constraint
  inline static float TargetBox_ymin(); // returns minimum y of target box constraint
  inline static float TargetBox_ymax(); // returns maximum y of target box constraint
  inline static float TargetBox_zmin(); // returns minimum z of target box constraint
  inline static float TargetBox_zmax(); // returns maximum z of target box constraint

  inline static int nStations();
  inline const L3_TargetStation& getStation(const int n);

  static void calculate_TargetBox(); //calculates target box around wires for tracking constraint

  static const L3_Target& ClosestTarget( Hep3Vector v );
  inline static int nPlanes(int sl0,int sec0);
  inline static int sectorAtSameZ(int sl0,int sec0);

  inline static int nRFBand(){ return rfbands.size(); }
  inline static const L3_RFBand& getRFBand(int index){ return rfbands[index]; }

private:
  static vector<L3_Target> targetList; //list with targets (i.e. target wires)
  static vector<L3_Target*> targetListZ;    //list with (pointers to) targets sorted in z
  static float the_maxztarget;    //z-position of most upstream target, is calculated in readTargets()
  static float the_minztarget;    //z-position of most downstream target, is calculated in readTargets()
  static float the_maxytarget;   
  static float the_minytarget;   
  static float the_maxxtarget;   
  static float the_minxtarget;   
  static float the_TargetBox_xmin; //minimum z of target box constraint
  static float the_TargetBox_xmax; //maximum z of target box constraint
  static float the_TargetBox_ymin; //minimum z of target box constraint
  static float the_TargetBox_ymax; //maximum z of target box constraint
  static float the_TargetBox_zmin; //minimum z of target box constraint
  static float the_TargetBox_zmax; //maximum z of target box constraint

  static L3_Plane SiVD_PlaneList[7][4][4]; //array of VDS planes
  static float SuperLayer_z[7][4];         //array of medium z-positions of VDS-quadrants
  static L3_Window secs[7][4];             //array of VDS sector boundaries 
  static Hep3Vector the_mean_target;
  static int LUT[7][4][4][3];  //[sl][sec][plane 0,1,2 or 3][stereo,other1,other2]
  static vector<L3_TargetStation> stationList;
  static int numPlanes[7][4];  
  static int sectorAtSameZPosition[7][4];
  static vector<L3_RFBand> rfbands;

  static void fillTargetGeometry();//fill target geometry
  static void fillVDSGeometry();   //fill vertex detector geometry    
  static void readTargets(); //read target info from GTAR-bank (ARTE) or from slp-frame (SLP)
  static int readPlanes();  //read plane  info from GEDE-bank (ARTE) or from SLT planes(SLP)

  static void calculate_TargetPosition(); //calculates target position
  static void calculate_mean_z();    //calculates mean z-position of quadrants 
  static void calculate_mean_z_old();    //calculates mean z-position of quadrants 
  static void calculate_sectors();   //calculates VDS sector boundaries
  static void showGeometry();
  static void calculateLUT();
  static int NPNPorNPPN(int sl,int sec);  
};
#include "L3_Geometry.icc"
#endif





