// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: an (incomplete) collection of distance functions
//
// changes:
// 21 Aug 2000 (TG) creation
// 04 Sep 2000 (TG) SGNdistance(const Track&, const Vertex&) function added
// 06 Sep 2000 (TG) SGNdistance(const Track&, double x, double y, double z) added
// 26 Okt 2000 (TG) #ifndef NO_ARTE added (for integration into CLUE)
// 17 Okt 2001 (TG) added/improved #defines for ARTE
//
// *****************************************************************************
#include <iostream>

#include "vt++/VtUtil.hh"
#include "vt++/VtVertex.hh"
#include "vt++/VtTrack.hh"
#include "vt++/VtDistance.hh"

#if defined NO_ARTE && defined USE_ROOT
#include "clue/cluearte.hh"
#include "interfaces/VertexIf.hh"
#include "clue/RecoVertex.hh"
#endif

namespace VERTEX {

#ifndef NO_ARTE
  //============================================================================
  // distance: spatial distance between RVER and MVER
  //============================================================================
  double distance(const ArtePointer<RVER>& rver ,
		  const ArtePointer<MVER>& mver) {
    return sqrt(sqr(rver->xv - mver->x) + sqr(rver->yv - mver->y) +
		sqr(rver->zv - mver->z));
  }

  //============================================================================
  // distance: spatial distance between Track and Wire
  //============================================================================
  double distance(const Track& t, const ArtePointer<GTAR>& gtar) {
    double mz1 = 1.;
    double rx2 = gtar->x;
    double ry2 = gtar->y;
    double rz2 = gtar->z;

    double drx = t.x() - rx2;
    double dry = t.y() - ry2;
    double drz = t.z() - rz2;
    double mx1 = t.tx();
    double my1 = t.ty();
    double mx2=0, my2=0, mz2=0;
    
    if(gtar->dx > gtar->dy) {
      mx2 = 1.;
      my2 = 0.;
      mz2 = 0.;
    } else {
      mx2 = 0.;
      my2 = 1.;
      mz2 = 0.;
    }

    double denom = sqr(mx1*my2-my1*mx2) + sqr(my1*mz2-mz1*my2) +
      sqr(mz1*mx2-mx1*mz2);

    double detx = drx*my1*mz2 - drx*mz1*my2;
    double dety = dry*mz1*mx2 - dry*mx1*mz2;
    double detz = drz*mx1*my2 - drz*my1*mx2;
    
    return fabs(detx+dety+detz)/denom;
  }

  //============================================================================
  // distance: spatial distance between RVER and GTAR
  //============================================================================
  double distance(const ArtePointer<RVER>& rver ,
		  const ArtePointer<GTAR>& gtar) {

    double x1 = gtar->x;
    double y1 = gtar->y;
    double z1 = gtar->z;

    double a, b;

    if(gtar->dx > gtar->dy) {
      a = 1.;
      b = 0.;
    } else {
      a = 0.;
      b = 1.;
    }
    
    double x3 = rver->xv;
    double y3 = rver->yv;
    double z3 = rver->zv;

    double nom = sqr((x3-x1)*b - (y3-y1)*a) + sqr((z3-z1)*b) + sqr((z3-z1)*a);
    double denom = sqr(a)+sqr(b);

    return sqrt(nom/denom);
  }

  //============================================================================
  // distance: spatial distance between Vertex and GTAR
  //============================================================================
  double distance(const Vertex& vtx ,
		  const ArtePointer<GTAR>& gtar) {

    double x1 = gtar->x;
    double y1 = gtar->y;
    double z1 = gtar->z;

    double a, b;

    if(gtar->dx > gtar->dy) {
      a = 1.;
      b = 0.;
    } else {
      a = 0.;
      b = 1.;
    }
    
    double x3 = vtx.vx();
    double y3 = vtx.vy();
    double z3 = vtx.vz();

    double nom = sqr((x3-x1)*b - (y3-y1)*a) + sqr((z3-z1)*b) + sqr((z3-z1)*a);
    double denom = sqr(a)+sqr(b);

    return sqrt(nom/denom);
  }
#endif // NO_ARTE

  //============================================================================
  // distance: Chi2 distance between tracks
  //============================================================================
  double distanceChi2(Track& t1, Track& t2) {
    Vertex tmp(t1,t2);
    tmp.findVertexVt();
    return tmp.chi2();
  }

  //============================================================================
  // distance: spatial distance between Track and Vertex
  //============================================================================
  double distance(const Track& t, const Vertex& v) {
    double dx = v.vx() - t.x();
    double dy = v.vy() - t.y();
    double dz = v.vz() - t.z();
    double tx = t.tx();
    double ty = t.ty();
    double nom = sqr(dx*ty - dy*tx) + sqr(dy - dz*ty) + sqr(dz*tx - dx);
    double denom = sqr(tx) + sqr(ty) + 1.;
    return sqrt(nom/denom);
  }
 
  //============================================================================
  // SGNdistance: signed spatial distance between Track and Vertex
  //============================================================================
  double SGNdistance(const Track& t, const Vertex& v) {
    return SGNdistance(t, v.vx(), v.vy(), v.vz());
  }

  //============================================================================
  // SGNdistance: signed spatial distance between Track and space point
  //============================================================================
  double SGNdistance(const Track& t, double xv, double yv, double zv) {

    double A  = t.tx();
    double B  = t.ty();
    double C  = 1.;
    double xt = t.x();
    double yt = t.y();
    double zt = t.z();

    double denom = A*A + B*B + C*C;
    double nom   = A*(xt-xv) + B*(yt-yv) + C*(zt-zv);
    double rho   = nom / denom;

    // point of track closest to Vertex
    double x = xt - A*rho;
    double y = yt - B*rho;
    double z = zt - C*rho;

    double dx = x - xv;
    double dy = y - yv;
    double dz = z - zv;

    // The sign tells if the point of track closest to 
    // vertex is upstream or downstream the vertex
    return sgn(dz) * sqrt(dx*dx + dy*dy + dz*dz);
  }

  //============================================================================
  // distance: Chi2 distance between two vertices
  //============================================================================
  double distanceChi2(const Vertex& v1, const Vertex& v2) {
    return v1.distance(v2);
  }

  //============================================================================
  // distance: Chi2 distance between vertex and space point
  //============================================================================
  double distanceChi2(const Vertex& v, double x, double y, double z) {
    return v.distance(x,y,z);
  }

#if defined NO_ARTE && defined USE_ROOT
  //============================================================================
  // SGNdistance: signed spatial distance between Track and RecoVertex
  //============================================================================
  double SGNdistance(const Track& t, const ArtePointer<RecoVertex>& ap) {
    return SGNdistance(t, ap->vx(), ap->vy(), ap->vz());
  }

  //============================================================================
  // SGNdistance: signed spatial distance between Track and RecoVertex
  //============================================================================
  double SGNdistance(const Track& t, const VertexIf& v) {
    return SGNdistance(t, v.vx(), v.vy(), v.vz());
  }
#endif

} // end of namespace VERTEX
